; ======================================================================
; OR90V3S7 - 2017-08-12
; it only uses 5 calls to the BASIC ROM. These are:
; [$A002] - CHROUT - Writes a character to a device
; [$A004] - CSRDON - Turns on the cassette motor
; [$A006] - BLKIN  - Reads a block from the cassette
; [$A008] - BLKOUT - Writes a block to the cassette
; [$A00C] - WRTLDR - Turns motor on and writes a Leader on the cassette
; These calls only appear once each in the Orch90 code.
; CHROUT is only used for the LIST to printer function.
; ======================================================================
											; Orchestra90 variables at Page 0
SONGBEG	EQU	$005E						; 2 bytes - Address of the beginning of the song buffer
DEVNUM	EQU	$006F						; 1 byte  - Device number -2 = printer
SONGEND	EQU	$0074						; 2 bytes - Address of the end of current buffer
FLGPLAY	EQU 	$005A						; 1 byte  - Flags that a file is to play
FLGFAST	EQU 	$005C						; 1 byte  - Flag for fast CPU play
FLGDSK	EQU 	$00A9						; 1 byte  - Cassette/disk flag, 0=cass 1=disk
FLGV5		EQU 	$00AA						; 1 byte  - 4 or 5 voice flag, 0=4, 1=5
											; Error numbers (ASCII)
ERR1		EQU 	'1'+$40					; Error 1 $71 : Memory overflow
ERR2		EQU 	'2'+$40					; Error 2 $72 : Symbol out of context
ERR3		EQU 	'3'+$40					; Error 3 $73 : Parameter error
ERR4		EQU 	'4'+$40					; Error 4 $74 : Invalid part number
ERR5		EQU 	'5'+$40					; Error 5 $75 : Measure overflow
ERR6		EQU 	'6'+$40					; Error 6 $76 : Note Overflow
ERR7		EQU 	'7'+$40					; Error 7 $77 : File I/O error; translate A using table at X
											; Other Constants
PLAYRAM	EQU 	$08B4						; $9C = 156 bytes - area where the needed player will be copied (4V or 5V)

	if BIN_OUT == 1 						; to create a BIN file that loads at $4000
			ORG	$C000
			PUT	$4000
	else										; to create a ROM file starting at $c000
			ORG	$C000
			PUT	$C000
	endif

Start		lbra	LDF21X					; initialise
;Start up message
Title		FCC	"ORCHESTRA90-DRAGON"
			FCB 	$0D
			FCC	/SOFTWARE AFFAIR, LTD/
			FCB 	$0D
			FCC	/COPYRIGHT 1984 JON BOKELMAN/
			FCB 	$0D
			FCB 	$0D
			FCC 	/ADAPTED FOR THE DRAGON/
			FCB 	$0D
			FCC	/VERSION 3S7 (2017)/
			FCB 	$0D
			FCC	/MPI ADAPTED BY STEWART AND PERE/
			FCB 	$0D
			FCC	/KEYBOARD MODDED BY TONY JEWELL/
			FCB 	$0D
			FCC	/SOUND ADAPTED BY STEWART ORCHARD/
			FCB 	$0D
			FCC	/DISK SYSTEM BY PERE SERRAT/

; music data: demo song
; All lines begin with character code plus 128
; This eliminates the need for CRs in the score
; Any char read +=128, is presumed to be a new line
DemoSong	FCB	$AF						; '/' = 47+128
			FCC	/ WILLIAM TELL OVERTURE/
			FCB 	$AF						; '/' = 47+128
			FCC	/ BY ROSSINI/
			FCB	$A0						; ' ' = 32+128
			FCB	$A0						; ' ' = 32+128
			FCB	$A0						; ' ' = 32+128
			FCB	$A0						; ' ' = 32+128
			FCB	$A0						; ' ' = 32+128
			FCB	$A0						; ' ' = 32+128
			FCB	$CE						; 'N' = 78+128
			FCC	/Q=66Z2K4#<3O1/
			FCB	$CD						; 'M' = 77+128
			FCB	$D9						; 'Y' = 89+128
			FCC	/A(Q.6,S6;6;)1I6;4;2;4;6;4;6;9;/
			FCB	$CD						; 'M' = 77+128
			FCB	$B6						; '6' = 54+128
			FCC	/;4;2;4;6;4;6;9;/
			FCB	$D6						; 'V' = 86+128
			FCC	/2YAQ.2,S2;2;Q.2,S4;4;/
			FCB	$D6						; 'V' = 86+128
			FCC	/3YEH..$S2;2;/
			FCB	$CD						; 'M' = 77+128
			FCB	$A8						; '(' = 40+128
			FCC	/Q.6,S6;6;)3/
			FCB	$D6						; 'V' = 86+128
			FCC	/2I4;3;2;3;4;6;5;4;3;4;3;5;4;3;2;4;/
			FCB	$D6						; 'V' = 86+128
			FCC	/32;@1;3;1;*2;4;3;2;-1;2;-1;3;2;-1;-3;2;/
			FCB	$D0						; 'P' = 80+128
			FCC	/03/
			FCB	$B6						; '6' = 54+128
			FCC	/;S(6;6;I)1/
			FCB	$D6						; 'V' = 86+128
			FCC	/23;S(3;3;I)1/
			FCB	$D6						; 'V' = 86+128
			FCC	/3(1;S1;1;I1;1;/
			FCB	$C0						; '@' = 64+128
			FCC	/V4YA)1/
			FCB	$A8						; '(' = 40+128
			FCC	/R03)2/
			FCB	$D0						; 'P' = 80+128
			FCC	/ /
			FCB	$AA						; '*' = 42+128
			FCC	/W6/
			FCB	$D6						; 'V' = 86+128
			FCC	/23/
			FCB	$D6						; 'V' = 86+128
			FCC	/31/
			FCB	$C0						; '@' = 64+128
			FCC	/V4FH$/
			FCB	$D0						; 'P' = 80+128
			FCC	/05/
			FCB	$D5						; 'U' = 85+128
			FCC	/+7S1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC	/3YB3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/45;5;/
			FCB	$D0						; 'P' = 80+128
			FCC	/06/
			FCB	$A8						; '(' = 40+128
			FCC	/M/
			FCB	$B1						; '1' = 49+128
			FCC	/1,1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC	/333,3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/455,5;5;)1/
			FCB	$D0						; 'P' = 80+128
			FCC	/07/
			FCB	$AA						; '*' = 42+128
			FCC	/I2,3,4,@S1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC	/2I(1,)2/
			FCB	$D6						; 'V' = 86+128
			FCC	/3(3,)2S3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/4I(5,)2S5;5;/
			FCB	$CD						; 'M' = 77+128
			FCB	$B1						; '1' = 49+128
			FCC	/1,1;1;*22,4;4;/
			FCB	$C0						; '@' = 64+128
			FCC	/V2Q$I(1;S)2/
			FCB	$A8						; '(' = 40+128
			FCC	/V333,3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/455,5;5;)1/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC	/I3,1,@1,S1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC	/2I(1,)2/
			FCB	$D6						; 'V' = 86+128
			FCC	/3(4,)2S3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/4I(6,)2S5;5;/
			FCB	$D2						; 'R' = 82+128
			FCC	/06/
			FCB	$D0						; 'P' = 80+128
			FCC	/08/
			FCB	$AA						; '*' = 42+128
			FCC	/I2,3,4,S24/
			FCB	$C0						; '@' = 64+128
			FCC	/V2I(1,)2S1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC	/3I(3,)2S3;3;/
			FCB	$D6						; 'V' = 86+128
			FCC	/4I(5,)2S5;5;/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC	/Q6S6543/
			FCB	$C0						; '@' = 64+128
			FCC	/(V222,2;2;/
			FCB	$D6						; 'V' = 86+128
			FCC	/366,6;6;/
			FCB	$D6						; 'V' = 86+128
			FCC	/488,8;8;)1/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC	/I2;4;2;/
			FCB	$C0						; '@' = 64+128
			FCC	/V2$1;1;/
			FCB	$A8						; '(' = 40+128
			FCC	/V33;/
			FCB	$D6						; 'V' = 86+128
			FCC	/45;)2/
			FCB	$D2						; 'R' = 82+128
			FCC	/05YEV3YER06R07R06R08/
			FCB	$D0						; 'P' = 80+128
			FCC	/ /
			FCB	$AA						; '*' = 42+128
			FCC	/U0YASB;B;/
			FCB	$D6						; 'V' = 86+128
			FCC	/29;9;/
			FCB	$C0						; '@' = 64+128
			FCC	/V3YA0;0;/
			FCB	$D6						; 'V' = 86+128
			FCC	/47;7;/
			FCB	$D0						; 'P' = 80+128
			FCC	/13/
			FCB	$A8						; '(' = 40+128
			FCC	/M/
			FCB	$AA						; '*' = 42+128
			FCC	/BB,B;B;/
			FCB	$D6						; 'V' = 86+128
			FCC	/299,9;9;/
			FCB	$D6						; 'V' = 86+128
			FCC	/300,0;0;/
			FCB	$C0						; '@' = 64+128
			FCC	/V477,7;7;)1/
			FCB	$A8						; '(' = 40+128
			FCC	/M/
			FCB	$AA						; '*' = 42+128
			FCC	/IB;E;/
			FCB	$D6						; 'V' = 86+128
			FCC	/29;7;/
			FCB	$D6						; 'V' = 86+128
			FCC	/30;$/
			FCB	$D6						; 'V' = 86+128
			FCC	/4-7;$)1/
			FCB	$D0						; 'P' = 80+128
			FCC	/14/
			FCB	$C2						; 'B' = 66+128
			FCC	/;E;B;A;9;8;7;SB;B;/
			FCB	$D6						; 'V' = 86+128
			FCC	/2I9;7;9;8;7;7;4;S9;9;/
			FCB	$C0						; '@' = 64+128
			FCC	/V3I0;$3;$3;4;5;S0;0;/
			FCB	$D6						; 'V' = 86+128
			FCC	/4I7;$A;$(7;)2S7;7;/
			FCB	$D2						; 'R' = 82+128
			FCC	/13/
			FCB	$D0						; 'P' = 80+128
			FCC	/15/
			FCB	$AA						; '*' = 42+128
			FCC 	/IB;E;D;C#;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/29;7;A;9;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/30;$1;0;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V47;$4;4;/
			FCB	$D0						; 'P' = 80+128
			FCC 	/ /
			FCB	$AA						; '*' = 42+128
			FCC 	/D;C#;D;SB#;B;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2I8;7;8;S9;9;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V3I1;2#;1;S0;0;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4I(8;)2S7;7;/
			FCB	$D2						; 'R' = 82+128
			FCC 	/13R14R13R15/
			FCB	$D0						; 'P' = 80+128
			FCC 	/ /
			FCB	$AA						; '*' = 42+128
			FCC 	/(D;C#;E;C;)3/
			FCB	$D6						; 'V' = 86+128
			FCC 	/28;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3YCI-1;$$(S3;3;I3;)14;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V48;$$(S1;1;11,)1*2;/
			FCB	$CD						; 'M' = 77+128
			FCB	$A8						; '(' = 40+128
			FCC 	/D;C#;E;C;)3/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3I5;33"5;4;22"4;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/43;-1-1"3;2;-3-3"2;/
			FCB	$CD						; 'M' = 77+128
			FCB	$D3						; 'D' = 83+128
			FCC 	/D;C#;E;C;ID,/
			FCB	$C0						; '@' = 64+128
			FCC 	/V2$1;1;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3$4;4;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4+3;6;6;/
			FCB	$D0						; 'P' = 80+128
			FCC 	/ /
			FCB	$A8						; '(' = 40+128
			FCC 	/M/
			FCB	$AA						; '*' = 42+128
			FCC 	/YBSA;A;AA,/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2YC6;6;66,/
			FCB	$D6						; 'V' = 86+128
			FCC 	/33;3;33,/
			FCB	$C0						; '@' = 64+128
			FCC 	/V41;1;I1;)1/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/B;C,AA"C;B,99"B;A,S6;6;I6;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/29;A,66"A;9,44"9;8,/
			FCB	$D6						; 'V' = 86+128
			FCC 	/42;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/34;@(+5,8,7,6,5,4,3,2,1,/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4)1/
			FCB	$D2						; 'R' = 82+128
			FCC 	/05YAV2YAV3YAR06R07R06R08/
			FCB	$D2						; 'R' = 82+128
			FCC 	/05YEV3YER06R07R06R08/
			FCB	$D0						; 'P' = 80+128
			FCC 	/ Z1/
			FCB	$AA						; '*' = 42+128
			FCC 	/YAS(6;6;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2)1/
			FCB	$C0						; '@' = 64+128
			FCC 	/V41;1;/
			FCB	$A8						; '(' = 40+128
			FCC 	/M/
			FCB	$AA						; '*' = 42+128
			FCC 	/66,6;6;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/266,6;6;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V411,1;1;)1/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/I(B;C;D;$9;A;B;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2)1/
			FCB	$D6						; 'V' = 86+128
			FCC 	/44;5;6;$2;3;4;$/
			FCB	$CD						; 'M' = 77+128
			FCB	$C2						; 'B' = 66+128
			FCC 	/;C;D;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/24;5;6;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V43;2;1;H$/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/S66#78%8#99#A/
			FCB	$C0						; '@' = 64+128
			FCC 	/(V2S4;1;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3YA6;$/
			FCB	$D6						; 'V' = 86+128
			FCC 	/48;$)3/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/U+7S4%4#55#66#78/
			FCB	$C0						; '@' = 64+128
			FCC 	/(V2S4;1;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/36;$/
			FCB	$D6						; 'V' = 86+128
			FCC 	/48;$)3/
			FCB	$A8						; '(' = 40+128
			FCC 	/M/
			FCB	$AA						; '*' = 42+128
			FCC 	/YCS98A8/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2YD98A8/
			FCB	$C0						; '@' = 64+128
			FCC 	/V3I3;2;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/45;8;)7/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/U0G;S(G9)2/
			FCB	$C0						; '@' = 64+128
			FCC 	/V2YAQ$S(1;+2;)1/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3YEI3;$S3;$3;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4I5;$S5;$5/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/IG;S(D6)2ID;S(B4)2IB;S(92)2/
			FCB	$C0						; '@' = 64+128
			FCC 	/V2(I1;S1;1;I1;+2;)1/
			FCB	$A8						; '(' = 40+128
			FCC 	/V33;S3;3;I3;1;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/45;S5;5;I5;3;)2/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/(9;S9;9;I9;9;9;S6;6;I6;6;6;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2)1/
			FCB	$C0						; '@' = 64+128
			FCC 	/V33;(+2;)3(1;)3/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4(5;)4(8;)3/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/(S4;4;I4;4;4;S2;2;I2;2;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2)1/
			FCB	$C0						; '@' = 64+128
			FCC 	/V3(3;)3(5;)2/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4(A;)3(C;)2/
			FCB	$CD						; 'M' = 77+128
			FCB	$A8						; '(' = 40+128
			FCC 	/*V12;-1;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/22;-1;/
			FCB	$C0						; '@' = 64+128
			FCC 	/V35;8;)1/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4U7C;8;5;8;/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/(4;2;6;4;9;6;B;9;D;B;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2)1/
			FCB	$C0						; '@' = 64+128
			FCC 	/V3(3;5;1;3;5;8;3;5;1;3;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4)1/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/Q.G,SG;G;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2Q.B,SB;B;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3U7Q.9,S9;9;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4Q.2,S2;2;/
			FCB	$CD						; 'M' = 77+128
			FCB	$A8						; '(' = 40+128
			FCC 	/V1IG;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2B;/
			FCB	$D6						; 'V' = 86+128
			FCC 	/39;)3/
			FCB	$D6						; 'V' = 86+128
			FCC 	/42@131/
			FCB	$CD						; 'M' = 77+128
			FCB	$AA						; '*' = 42+128
			FCC 	/Q.G,S$G;Q.G,S$9;W9/
			FCB	$D6						; 'V' = 86+128
			FCC 	/2Q.B,S$B;Q.B,S$-3;W-3/
			FCB	$D6						; 'V' = 86+128
			FCC 	/3Q.9,S$6;Q.6,@S$1;W1/
			FCB	$D6						; 'V' = 86+128
			FCC 	/4Q.5,S$5;Q.5,S$C;WC/
Return	RTS

; "command mode" command L : List to printer
CmdList
LC78C    LDA   #$FE						; device number for printer
	    	STA   <DEVNUM					; save value

	if INT_DAC == 1					; mod to avoid sending garbage to the printer (by Stewart Orchard)
			lda 	#$30    					; bit2=0 - select direction register
			sta 	>$ff21  					;
			lda 	#$fe    					; set strobe as ouput ... 1111 1110 allowing output data
			sta 	>$ff20   				;
			lda 	#$34    					; bit2=1 - change back to data register
			STA   >$ff21					; $ff21=%b0011 0100 disable CA1 (printer busy)
	endif

	    	LDX   <$60						; point to beginning of score
LC882    TST   ,X							; is byte negative (1st in a row)?
	    	BPL   LC88A						; no, skip next
	    	LDA   #$0D						; get carriage return
	    	BSR   LC89A						; output character to DEVN
LC88A    LDA   ,X+						; get bayte, advance pointer
	    	ANDA  #$7F						; reset bit7

	if INT_DAC == 1
	    	BEQ   endList					; if end of text, go configure strobe as input (printer disabled)
	else
			BEQ	Return    				; if end of text return
	endif

	    	CMPA  #$60   					; is it a sign >= $60?
	    	BCS   LC896						; no, skip next
	    	SUBA  #$40   					; reset bit5
LC896    BSR   LC89A						; output character to DEVN
	    	BRA   LC882						; loopback
LC89A    JMP  [$A002]					; BASIC vector - output character to DEVN

	if INT_DAC == 1					; mod to avoid sending garbage to the printer (by Stewart Orchard)
endList	lda 	#$30    					; bit2=0 - select direction register
			sta 	>$ff21  					;
			lda 	#$fc    					; set strobe as input ... 1111 1100 disabling printer
			sta 	>$ff20  					;
			lda 	#$34    					; bit2=1 - change back to data register
			STA   >$ff21					; $ff21=%b0011 0100 disable CA1 (printer busy)
			rts
	endif

; "EDIT mode" command shift up: Deletes an end of line
; Joins two rows, works only at the begining of the row
; joining that line with the previous one
CmdSup	LBSR  LCADF
			LDX   <$62
 			ROL   ,X
			CLRA
			ROR   ,X
			LBSR  LCAC2
			BRA   LC8E3

; display ?? on 2nd line
DspErr	LBRA  PrntWhat					; display ?? on 2nd line

; "command mode" command '/' (6F): Goto string (down in score)
CmdGoDwn	LDB   #$02

; "command mode" command '-' (6D): Goto string (up in score)
CmdGoUp	DECB
			STB   <$58
			LDX   <$62
LC8B7		LDA   <$58
			LEAX  A,X
			CMPX  <SONGEND					; end of song data
			BHI   DspErr					; display ?? on 2nd line
			CMPX  <SONGBEG					; start of song data
			BCS   DspErr					; display ?? on 2nd line
			PSHS  X
			LEAU  1,Y
LC8C7		LDA   ,U+
			BMI   LC8E1
			EORA  ,X+
			ASLA
			BEQ   LC8C7
			PULS  X
			BRA   LC8B7

; "??? mode" command <shift> <ENT>: Double Speed play
CmdFast	TST   <$B9						; is fast mode allowed?
			BNE   CmdEnter					; no, go "command mode" command ENT
			INC   <FLGFAST					; yes, set fast mpu rate flag

; "command mode" command ENT: send entered command
CmdEnter	STY   ,S
			BSR   Clr2Ln					; green fill 2nd line
			PULS  X,PC

LC8E1		PULS  X
LC8E3		STX   <$62
			LEAX  1,X
LC8E7		TST   ,-X
			BPL   LC8E7
LC8EB		LBSR  LCABB
			LEAU  30,X
			CMPU  <$62
			BCC   LC936
			LDA   ,X
			ASLA
			BEQ   LC934
			LDB   #$1F
LC8FD		LEAX  1,X
			TST   ,X
			BMI   LC8EB
			DECB
			BNE   LC8FD
			BRA   LC8EB

; green fill 2nd line
Clr2Ln	LDX   #$0420					; point to 1st char second line
			BRA   Clr2End					; green fill to end of line starting from X

LC90D		LDX   <SONGEND					; end of song data
LC90F		STX   <$5D
			STX   <SONGBEG					; start of song data
LC913		LBRA  LCAB9

; clear top line to green
ClrTop	LDX   #$0400

; green fill to end of line starting from X
Clr2End	LDA   #$01

; green fill A lines starting from X
clrALns	LDB   #$8F						; green semigraphic full bloc
LC91D		STB   ,X+						; put a green bloc
			EXG   X,D						; save D, get X
			BITB  #$1F						; reached end of line?
			EXG   X,D						; restore D
			BNE   LC91D						; keep going until end of line
			DECA								; decrement line counter
			BNE   LC91D						; do another line
			RTS								; return

; "command mode" command N : New
CmdNew	LDX   <SONGEND					; end of song data
LC92D		BSR   LC90F

; "command mode" command T : Top
CmdTop	LDX   <SONGBEG					; start of song data
			FCB   $8C 						; CMPX #	- to skip next one

; "command mode" command B : Bottom
CmdBot	LDX   <SONGEND					; end of song data
LC934		BSR   LC913
LC936		LDA   #$06
			LDU   <$60
LC93A		LDB   #$1F
			CMPU  <SONGBEG					; start of song data
			BEQ   LC94B
LC941		TST   ,-U
			BMI   LC948
			DECB
			BNE   LC941
LC948		DECA
			BNE   LC93A
LC94B		LDX   #$0440
			LDB   -1,X
			STB   <$6B
			TSTA
			BEQ   LC957
			BSR   clrALns					; green fill A lines starting from X
LC957		LEAX  -1,X
LC959		LDB   <$6B
			STB   $043F
			TFR   X,D
			ORB   #$E0
			INCB
			BEQ   LC969
			TST   ,U
			BPL   LC974
LC969		ROL   ,U
			COMB
			ROR   ,U
			BSR   Clr2End					; green fill to end of line starting from X
			CMPX  <$AD						; top of screen
			BEQ   LC99A
LC974		LDB   ,U
			ANDB  #$7F
			BNE   LC97C
			LDB   #$8F
LC97C		CMPU  <$62
			BNE   LC98D
			STX   <$88						; cursor position
			TST   <$55
			BNE   LC98D
			EORB  #$40
			BPL   LC98D
			EORB  #$4F
LC98D		STB   ,X+
			LDB   ,U+
			ASLB
			BNE   LC959
LC994		BSR   Clr2End					; green fill to end of line starting from X
			CMPX  <$AD						; top of screen
			BNE   LC994
LC99A		RTS

; "EDIT mode" command lookup table
LC99B 	FCB   $03
			FDB	LCBE1						; BRK
	 		FCB   $0A
	 		FDB	LCA84						; down
	  		FCB   $0B
	  		FDB	LCA70						; up
	 		FCB   $0D
	 		FDB	LCA80						; ENT
	 		FCB   $1A
	 		FDB	LCAA7						; shift down
	 		FCB   $1B
	 		FDB	CmdSup					; shift up
	 		FCB   $1D
	 		FDB	LCA80						; shift ENT
	 		FCB   $21
	 		FDB	LCAD1						; !
	 		FCB   $3F
	 		FDB	LCA7C						; ?
	 		FCB   $FF

LC9B7		LDA   1,X
			STA   ,X+

; "command mode" command shift left
LC9BB		TST   ,X
			BPL   LC9B7
			RTS

; "EDIT mode" command shift right: Add a space (before current char)
LC9C0		LEAX  31,Y
LC9C3		CMPX  <$88						; cursor position
			BEQ   LC9CD
			LDA   ,-X
			STA   1,X
			BRA   LC9C3
LC9CD		LDA   #$60
			STA   ,X
			RTS

; "EDIT mode" command lookup table
LC9D2 	FCB   $03
			FDB	CmdBreak					; BRK
	 		FCB   $08
	 		FDB	LCA62						; left
	 		FCB   $09
	 		FDB	LCA5B						; right
	 		FCB   $0C
	 		FDB	LCB2C						; CLR
	 		FCB   $0D
	 		FDB	CmdEnter					; ENT
	 		FCB   $13
	 		FDB	LCA12						; shift BRK
	 		FCB   $18
	 		FDB	LC9BB						; shift left
	 		FCB   $19
	 		FDB	LC9C0						; shift right
	 		FCB   $1C
	 		FDB	LCB3C						; shift CLR
	 		FCB   $1D
	 		FDB	CmdFast					; shift ENT (Dragon)
	 		FCB   $5F
	 		FDB	CmdFast					; shift ENT (XRoar)
	 		FCB   $FF

; clear 10 bytes starting at $0053
LC9F1		LDX   #$0053
			LDB   #$0A

; clear B bytes starting at X
LC9F6		CLR   ,X+
			DECB
			BNE   LC9F6
			RTS

LC9FC		LDY   #$0400
	    	TST   <$58
	    	BNE   LCA07
	    	LBSR  ClrTop					; clear top line to green
LCA07		STY   <$88						; cursor position
	    	BSR   LC9F1						; clear 10 bytes starting at $0053
	    	BRA   LCA22

; "command mode" command E : Edit
CmdEdit	LBSR  ClrTop					; clear top line to green
			DECB

; "EDIT mode" command shift BRK: UNDO mods in that row (only last row)
LCA12		INCB
			STB   <$58
			LBSR  Clr2Ln					; green fill 2nd line
			LDY   #$0500
			LDB   #$08
			STB   <$55
			STB   <$54
LCA22    STB   $FF22
LCA25    TST   <$54
			BEQ   LCA2C
			LBSR  LC936
LCA2C    LDB   #$8F
			STB   31,Y
			LBSR  LCC09
			BSR   LCA38
	    	BRA   LCA25
LCA38    STA   <$54
	    	TST   <$55
	    	BEQ   LCA44
	    	LDX   #LC99B
	    	LBSR  GetCmd					; command lookup A in table at X
LCA44    CLR   <$54
	    	LDX   #LC9D2
	    	LBSR  GetCmd					; command lookup A in table at X
	    	CMPA  #$20
	    	BCS   LCA5A
	    	ORA   #$40
	    	STA   ,X+
	    	TFR   X,D
	    	BITB  #$20
	    	BEQ   LCA6D
LCA5A    RTS

; "EDIT mode" command right: Move cursor to next char in the row
;										at the end of the row goes to beginning again (wraps)
LCA5B		TST   ,X+
			BPL   LCA6D
			CLR   <$89
LCA61		RTS

; "EDIT mode" command left: Move cursor to previous char in the row
;									at the beginning of the row goes to the end again (wraps)
LCA62		DEC   <$89
			BPL   LCA61
			FCB	$8C						; CMPX  - to skip next one
LCA67		LEAX  1,X
			TST   ,X
			BPL   LCA67
LCA6D		STX   <$88						; cursor position
			RTS

; "EDIT mode" command up: Goes to prior row (nothing on first row)
LCA70		BSR   LCADF
LCA72		CMPX  <SONGBEG					; start of song data
			BEQ   LCA94
			TST   ,-X
			BPL   LCA72
			BRA   LCA94

; "command mode" command '?' (7F): Display voicing at current position
LCA7C		LDX   <SONGEND					; end of song data
			BSR   LCAB9

; "EDIT mode" command ENT: accepts current line  / shift-ENT  does the same as Enter
LCA80		TFR   Y,X
			STX   <$88						; cursor position

; "EDIT mode" command down: Goes to next row (nothing on last row)
LCA84		BSR   LCADF
LCA86		CMPX  <SONGEND					; end of song data
			BEQ   LCAB9
			LEAX  1,X
			TST   ,X
			BPL   LCA86
			CMPX  <SONGEND					; end of song data
			BEQ   LCAB9
LCA94		BSR   LCABB
			CLRB
LCA97		CMPB  <$89
			BEQ   LCAA1
			INCB
			TST   B,X
			BPL   LCA97
			DECB
LCAA1		ABX
			STX   <$62
			LDX   <$60
			RTS

; "EDIT mode" command shift down: Adds a newline at actual position
;											 so it creates an empty row at the beginning of the row
;											 and splits a row in any other position
LCAA7		BSR   LCADF
LCAA9		LDX   <$62
			LDB   <$89
			BNE   LCABB
			LDB   #$01
			BSR   LCB10
			LDX   <$62
			LDB   #$60
			STB   ,-X
LCAB9		STX   <$62
LCABB		STX   <$60
			ROL   ,X
			COMB
			ROR   ,X
LCAC2		LDB   #$80
			STB  [$0074]
			ORB  [$005E]
			STB  [$005E]
			RTS

; "EDIT mode" command ! (61): Move one line UP (full line goes one row up)
LCAD1		CLR   <$89
			LDX   #$0800
			LBSR  Clr2End					; green fill to end of line starting from X
			BSR   LCA70						; "??? mode" command up
			BSR   LCAA9
			TFR   Y,X
LCADF		TFR   X,D
			CLRB
			TFR   D,X
			DECB
LCAE5		INCB
			TST   B,X
			BPL   LCAE5
			PSHS  A,B
			CLRA
			LDX   <$60
			CMPX  <SONGEND					; end of song data
			BEQ   LCAFA
LCAF3		INCA
			TST   A,X
			BPL   LCAF3
			LEAX  A,X
LCAFA		PSHS  A,B,X
			SUBB  ,S+
			BEQ   LCB02
			BSR   LCB10
LCB02		PULS  B,X,U
			TSTB
			BEQ   LCA94
LCB07		LDA   ,-U
			STA   ,-X
			DECB
			BNE   LCB07
			BRA   LCA94
LCB10		STB   <$5D
			NEGB
			BPL   LCB4B
			LDX   <SONGBEG					; start of song data
			LEAU  B,X
			CMPU  #$1064
			LBLS  Xerror1					; Error 1 : Memory overflow
			STU   <SONGBEG					; start of song data
LCB23		LDA   ,X+
			STA   ,U+
			CMPX  <$60
			BCS   LCB23
			RTS

; "EDIT mode" command CLR: clears (blanks) till the end of actual editing row
LCB2C		LBSR  Clr2End					; green fill to end of line starting from X
			TST   <$55
			BEQ   LCB3B
			TST   <$89
			BNE   LCB3B
			LDB   #$60
			STB   ,Y
LCB3B		RTS

; "EDIT mode" command shift CLR: deletes actual editing row
LCB3C		TFR   Y,X
			STX   <$88						; cursor position
			LBSR  Clr2End					; green fill to end of line starting from X
			TST   <$55
			BEQ   LCB3B
			INC   <$54
			BRA   LCADF

LCB4B		LDX   <$60
			LEAU  B,X
LCB4F		LDA   ,-X
			STA   ,-U
			CMPX  <SONGBEG					; start of song data
			BHI   LCB4F
			BEQ   LCB5B
 			LEAU  1,U
LCB5B		STU   <SONGBEG					; start of song data
			RTS

; "command mode" command lookup table
CmdTab 	FCB   '@'
			FDB	CmdLast					; $40 @ move to last played note after <BREAK>
	 		FCB   'A'
	 		FDB	CmdApend					; $41 A <filename> Append another file to the beginning of the current
	 		FCB   'B'
	 		FDB	CmdBot					; $42 B Bottom go to end of file
	 		FCB   'C'
	 		FDB	CmdCass					; $43 C Cassette Mode, puts orch90 in cassette mode
	 		FCB   'D'
	 		FDB	CmdDir					; $44 D Dir <:num> list directory of disk
	 		FCB   'E'
	 		FDB	CmdEdit					; $45 E Enter Edit mode
	 		FCB   'F'
	 		FDB	CmdPlayFast				; $45 F Plays at fast mpu speed
	 		FCB   'G'
	 		FDB	CmdGet					; $47 G Get<fname1><fname2><etc> load,score,play fnames
	 		FCB   'H'
	 		FDB	CmdHelp					; $48 H HELP. Shows allowed commands
	 		FCB   'K'
	 		FDB	CmdKill					; $4B K Kill <fname> (disk only)
	 		FCB   'L'
	 		FDB	CmdList					; $4C L List file to printer
	 		FCB   'M'
	 		FDB	CmdMulti					; $4D M Same as G, but starts over after last song
	 		FCB   'N'
	 		FDB	CmdNew					; $4E N New clears the music buffer
	 		FCB   'O'
	 		FDB	CmdOptmz					; $4F O Optimize the current file
	 		FCB   'P'
	 		FDB	CmdPlay					; $50 P Play [PartNum] P Play alone will play complete file
	 											;                        or from the partNumber to the end
	 		FCB   'R'
	 		FDB	CmdRead					; $52 R Read (Load) file
	 		FCB   'S'
	 		FDB	CmdScore					; $53 S Score file for playing
	 		FCB   'T'
	 		FDB	CmdTop					; $54 T Top go to beginning of file
	 		FCB   'V'
	 		FDB	CmdVerfy					; $56 V Verify <fname> Verifies cassette file
	 		FCB   'W'
	 		FDB	CmdWrite					; $57 W Write <fname> to cassette or disk
			fcb	'X'
			fdb	CmdRename				; $58 X Rename <old-fname><new-fname> on disk
			fcb	'Y'
			fdb	CmdMulti					; $59 Y 'ALL-ONCE'
	 		FCB   '!'+$40
	 		FDB	CmdPlyC					; $61 ! Play file from curren Pos
	 		FCB   '-'+$40
	 		FDB	CmdGoUp					; $6D - <string> finds string up
	 		FCB   '/'+$40
	 		FDB	CmdGoDwn					; $6F / <string> finds string down
	 		FCB   '0'+$40
	 		FDB	CmdSlot					; $70 0 Sets slot 1 on the MPI
	 		FCB   '1'+$40
	 		FDB	CmdSlot					; $71 1 Sets slot 2 on the MPI
	 		FCB   '2'+$40
	 		FDB	CmdSlot					; $72 2 Sets slot 3 on the MPI
	 		FCB   '3'+$40
	 		FDB	CmdSlot					; $73 3 Sets slot 4 on the MPI
	 		FCB   '4'+$40
	 		FDB	Cmd4V						; $74 4 Sets & scores in 4 voice mode
	 		FCB   '5'+$40
	 		FDB	Cmd4V						; $75 5 Sets & scores in 5 voice mode
	 		FCB   '<'+$40
	 		FDB	CmdProt					; $7C < UnProtects a disk file
	 		FCB   '>'+$40
	 		FDB	CmdProt					; $7E > Protects a disk file
	 		FCB   '?'+$40
	 		FDB	CmdVocD					; $7F ? Display voicing at current pos
	 		FCB   $FF

; "command mode" command M : Multi
CmdMulti	LBSR 	GetParm					; find command parameter
			BMI 	LCBE0						; no parameter, returns
			STA   <$58						; save last read char
LCBB1		LDA   <$58						; gets it back
			BEQ   ReInit					; Reinit the system if NULL
			LDY   #$0400					; point to entered command (1st char)

; "command mode" command G : Get
CmdGet	STY   <$64						; save pointer
			STA   <$57						;
			LEAS  2,S						; get rid of return address
			ldy	#0							; get 16 bits zero
			sty	>oldPointer				; set old pointer to null
LCBC0		LDY   <$64						; get pointer
			ldx	<SONGBEG					; get beginning of song
			stx	>oldBeginning			; save as old value
			LBSR  CmdRead					; command R : Read
			LBSR  InitHw					; initialise hardware
			LBSR  CmdScore					; command S : Score
			LBSR  JmpPlay					; play song
			BRA   LCBC0						; loopback to next song

; command lookup A in table at X
; doesn't return if command found
GetCmd	LEAX  3,X						; point to next entry
			CMPA  -3,X						; is A = code of previous entry?
			BCS   LCBDD						; if code greater than A, exit
			BNE   GetCmd					; if not equal loopback
			LDX   -2,X						; found! get address
			STX   ,S							; put onto stack (to jump there)
LCBDD		LDX   <$88						; cursor position
			CLRB								; clear B
LCBE0		RTS								; return if not found, else jump to found address

; "EDIT mode" command BRK: exits edit mode and goes to Command Mode
LCBE1		LBSR  LCADF
			CLR   <$55
LCBE6		LBSR  LC936
ReInit	LDS   #$00FA					; Reset stack back to $FA
			LBSR  InitHw					; initialise hardware
ReInit1	BSR   CmdLoop					; Start the command loop again
			BRA   ReInit					; Reinit the system after cmd loop returns error
												; or after cmd has processed. Main loop.

; Command Loop - polls keyboard until input,
; if cmd is found, it never returns from GetCmd.
; If it returns, there was an error and falls through to PrntWhat
CmdLoop	BSR   PollKey					; scan keyboard
			LBSR  LC9FC
			LDA   ,Y
			BMI   ReInit					; Reinit the system
			sta	>orcComand				; save received command
			LDX   #CmdTab					; "command mode" lookup table
			BSR   GetCmd					; command lookup A in table at X

; display ?? on 2nd line
PrntWhat	LDD   #$7F7F					; Put "??" in D
			STD   $0420						; to Print it on the beginning of the 2nd line
			RTS								; return

LCC09		CLR   <$6A
			LDX   <$88						; cursor position
			LDB   ,X
			STB   <$AF						; saved character under cursor
LCC11		BSR   LCC1F
			TST   <$B0
			BEQ   LCC11
			LDA   <$87
			ASL   <$94
			BCS   LCC11
			BRA   LCC44

LCC1F		LDX   <$88						; cursor position
			LDB   ,X
			TST   <$6A
			BNE   LCC2D
			EORB  #$40
			BPL   LCC2D
			EORB  #$4F
LCC2D		STB   ,X
			BSR   PollKey					; scan keyboard
			BNE   LCC3C
			DEC   <$6A
			LDB   <$6A
			BITB  #$3F
			BNE   LCC1F
			RTS

LCC3C		LEAS  2,S
			STA   <$87
			LDB   #$FC
			STB   <$94
LCC44		LDB   <$AF						; saved character under cursor
			STB  [$0088]
			RTS

LCC4B		LDX   #$045E					; set up for approx 10ms delay
			TST   <$70						; transfer mode flag
			LBEQ  LD408						; delay X*8 cycles
			LDB   #$9D
LCC56		SYNC
			DECB
			BNE   LCC56
			RTS

; scan keyboard
; first gets scan code of pressed key
; then converts to character code modified by shift
PollKey	LDU   #$FF00
			LDX   #$00B0
			CLR   ,X+
			CLRA
			DECA
			PSHS  A,X
			STA   2,U						; initialise $ff02 (keyboard)
LCC69		ROL   2,U						; select next keyboard col
			BCC   LCCAB						; all cols done. go to exit.
			INC   ,S
			LBSR  LCCEC
			STA   1,S
			TFR   A,B
			EORA  ,X
			STB   ,X
			ORB   <$B0
			STB   <$B0
			ANDA  ,X+
			BEQ   LCC69
			LDB   2,U
			STB   2,S
			LDB   #$F8

; Patch to Dragonize the keyboard routine (by Tony Jewell)
; then once we know what the PA readings
; are, we'll change the order so it feels like a Coco matrix
pa1		addb	#$08
			lsra
			bcc	pa1
			addb	,s
			cmpb	#$0F						; PA0 - PA1
			bhi	pa2
			addb	#$20						; add 4 rows
			bra   LCC8E   					; to save one byte. jmp	$LCC8E	; and back to the ROM
pa2		cmpb	#$2F						; PA2 - PA5
			bhi	LCC8E
			subb	#$10						; remove 2 rows
; end of keyboard patch

LCC8E		CMPB  #$1A
			BHI   LCCAF						; not @-Z
			TST   <$70						; transfer mode flag
			BEQ   LCC9A
			BSR   LCCDC						; check for down arrow pressed
			BNE   LCC9C
LCC9A		ADDB  #$40						; convert @-Z to character code
LCC9C		STB   ,S
			BSR   LCC4B
			LDA   2,S
			BSR   LCCEA
			CMPA  1,S
			BNE   LCCAB
			LDB   ,S
			FCB   $8C						; CMPX  # - to skip next one
LCCAB	   CLR   ,S
			PULS  A,X,PC
LCCAF		LDX   #LCCDE
			CMPB  #$21
			BLO   LCCCC						; up dn lft rgt spc 0
			LDX   #LCCBF+1
			CMPB  #$30
			BHS   LCCCC						; ENT CLR BRK . . . . SHFT
			BSR   LCCE3						; check for shift key pressed
LCCBF		CMPB  #$2B
			BLS   LCCC5						; 1 2 3 4 5 6 7 8 9 : ; (not , - . /)
			EORA  #$40						; invert shift flag
LCCC5		TSTA
			BNE   LCC9C
			ADDB  #$10
			BRA   LCC9C
LCCCC		ASLB
			TST   <$70						; transfer mode flag
			BEQ   LCCD3
			ADDB  #$12
LCCD3		BSR   LCCE3						; check for shift key pressed
			BEQ   LCCD8						; no shift
			INCB								; adjust for shifted key
LCCD8		LDB   B,X
			BRA   LCC9C

; check for down arrow pressed
; requires U = $ff00
LCCDC		LDA   #$EF
LCCDE		BSR   LCD04						; read keyboard col
			ANDA  #$08
			RTS

; check for shift pressed
; requires U = $ff00
LCCE3		LDA   #$7F
			BSR   LCD04						; read keyboard col
			ANDA  #$40
			RTS

LCCEA		STA   2,U
LCCEC		BSR   LCD06
			TST   2,U
			BMI   LCCF5
			ANDA  #$3F
			RTS
LCCF5		TST   <$70						; transfer mode flag
			BEQ   LCD01
			LDB   2,U
			CMPB  #$EF
			BNE   LCD01
			ANDA  #$77
LCD01		ORCC  #$01
			RTS

; read keyboard col
; value to write in A
; returns inverted keyboard output in A
; requires U = $ff00
LCD04		STA   2,U
LCD06		TST   <$70						; transfer mode flag
			BEQ   LCD0E
			SYNC
			LDA   <$6A
			FCB   $8C						; CMPX  # - to skip next one
LCD0E 	LDA   ,U
			COMA
			ANDA  #$7F						; mask off comparator bit
			RTS

; scan code to character code lookup table
; normal code followed by shifted code
	 		FCB   $0B,$1B					; up
	 		FCB   $0A,$1A					; down
	 		FCB   $08,$18					; left
	 		FCB   $09,$19					; right
	 		FCB   $20,$20					; space
	 		FCB   $30,$5F					; 0

; scan code to character code lookup table
; normal code followed by shifted code
	 		FCB   $0D,$1D					; ENT
	 		FCB   $0C,$1C					; CLR
	 		FCB   $03,$13					; BRK

; alternate scan code lookup table (transfer mode)
; normal code followed by shifted code
	 		FCB   $5E,$5C					; up
	 		FCB   $00,$00					; down
	 		FCB   $08,$5B					; left
	 		FCB   $09,$5D					; right
	 		FCB   $20,$20					; space
	 		FCB   $30,$5F					; 0

; alternate scan code lookup table (transfer mode)
; normal code followed by shifted code
	 		FCB   $0D,$0D					; ENT
	 		FCB   $1B,$80					; CLR
	 		FCB   $FF,$8F					; BRK

; data
	 		FCB   $20,$3F,$54,$50    	;    ?  T  P
	 		FCB   $4D,$4F,$52,$50    	; M  O  R  P

MsgOvlap FCB   $4F,$56,$45,$52,$4C,$41,$50,$3F 	; "OVERLAP?"

; munge high nybbles at X from 4 bytes into 2
LCD48		BSR   LCD4C
			TFR   B,A
LCD4C		LDB   1,X
			LSRB
			LSRB
			LSRB
			LSRB
			ORB   ,X++
			RTS

; Division: (,X) = D/(1,S) rem D ?
LCD55		CLR   ,X
			FCB	$8C						; CMPX - to avoid next one
LCD58		INC   ,X
			SUBD  3,S
			BCC   LCD58
			ADDD  3,S
			RTS
LCD61		LDA   9,U
			ADDA  #$0F
			LDB   <$66
			MUL
			STA   <$1C
			LEAX  1,U
			BSR   LCD48						; munge high nybbles at X from 4 bytes into 2
			STD   <$18
			BSR   LCD48						; munge high nybbles at X from 4 bytes into 2
			STD   <$1A
LCD74		LDD   <$18
			BNE   LCD7C
			LDD   <$1A
			STD   <$18
LCD7C		ADDD  <$18
			BCC   LCD84
			EORA  <$1A
			EORB  <$1B
LCD84		STD   <$18
			ANDA  #$80
			BEQ   LCD8C
			LDA   <$1C
LCD8C		STA  [$0016]
			INC   <$17
			BNE   LCD74
			LBRA  LCE5A

; Division: D = D / X ?
LCD97		PSHS  B,X
			LDX   #$006B
			TFR   A,B
			CLRA
			BSR   LCD55						; Division: (,X) = D/(1,S) rem D ?
			TFR   B,A
			LDB   ,S
			LEAX  1,X
			BSR   LCD55						; Division: (,X) = D/(1,S) rem D ?
			PULS  B,X
			LDD   <$6B
			RTS

LCDAE		LBSR  LDAE4
			LDX   #LCEE2
			LDU   #$0800
			LBSR  LD4AA						; copy bytes from X to U. (Length at -1,X)
			PULS  Y
			LDB   #$80						; construct -ve half of sine wave
			LEAX  B,X
LCDC0		LDA   ,X+
			NEGA
			STA   ,U+
			DECB
			BNE   LCDC0
			CLR   <$56
			LDY   #$0F00					; Point to wave data
			LDA   #$0A						; number of bytes to test
LCDD0		STA   <$16						; save at own variable
			LDD   #$0A04					; A=numbytes - B=control value (Zero flag in CCR set to 1)
			STA   <$6C						; save numbytes at counter
LCDD7		LDA   ,Y+						; get a byte from wave data
			CMPA  49,Y						; is it the same as 'old' data copy'?
			PSHS  CC							; save CCR onto stack
			ANDB  ,S+						; AND with stacked CCR. If stacked flag was 0, B value is zero too
			STA   49,Y						; update 'copy of data'
			DEC   <$6C						; decrement counter
			BNE   LCDD7						; if not zero, compare next byte
			TSTB								; is B zero (different data)?
			BNE   LCE5A						; NO, skip section (don't recalculate waves)
			CLR   <$17
			LEAU  -10,Y
			TST   ,U
			LBEQ  LCD61
			LDA   #$08
			LDX   <$8A						; zero
			STX   <$14
LCDFA		LDB   A,U
			ABX
			DECA
			BNE   LCDFA
			LDA   #$08
LCE02		PSHS  A
			CLRB
			LDA   A,U
			BEQ   LCE10
			BSR   LCD97						; Division: D = D / X ?
			DECA
			BMI   LCE10
			LDB   #$FF
LCE10		PULS  A
			STB   A,U
			DECA
			BNE   LCE02
LCE17		BSR   LCE83
			TSTA
			BPL   LCE1E
			BSR   LCE75
LCE1E		CMPD  <$14
			BCS   LCE25
			STD   <$14
LCE25		INC   <$17
			BPL   LCE17
			LDD   <$14
			BSR   LCE7C						; signed divide D by 16
			PSHS  A,B
			LDX   ,S++
			BEQ   LCE39
			LDA   <$67
			CLRB
			LBSR  LCD97						; Division: D = D / X ?
LCE39		STB   <$14
			CLR   <$17
LCE3D		BSR   LCE83
			LDB   <$14
			BSR   LCE70						; signed multiply D = A x B
			BSR   LCE7C						; signed divide D by 16
			TFR   B,A
			LDB   9,U
			ADDB  #$0F
			BSR   LCE70						; signed multiply D = A x B
			ADDA  <$67
			BPL   LCE52
			CLRA
LCE52		STA  [$0016]
			INC   <$17
			BNE   LCE3D
LCE5A		LDA   <$16
			INCA
			CMPA  #$0F

	if INT_DAC == 1					; mod to send sound to internal DAC (by Stewart Orchard)
    		LBCS  LCDD0
			ldx 	#$0a00					; make all wave data half amplitude
1			lsr 	,x+
			cmpx 	#$f00
			blo 	1b
	else
			LBLO  LCDD0
  	endif

			TST   <FLGPLAY					; is file to be played?
			BNE   LCEAF
			TST   <$59
			BEQ   LCEAC
			LBSR  LC90D
			BRA   LCEAF

; signed multiply D = A x B
LCE70		TSTA
			BPL   LCEDF						; multiply and return
			NEGA
			MUL
LCE75		COMB
			COMA
			ADCB  #$00
			ADCA  #$00
			RTS

; signed divide D by 16
LCE7C		BSR   LCE7E
LCE7E		BSR   LCE80
LCE80		ASRA
			RORB
			RTS

LCE83		CLR   ,-S
			CLR   ,-S
			LDA   #$08
LCE89		PSHS  A
			LDB   A,U
			BEQ   LCEA5
			PSHS  B
			CLRB
LCE92		ADDB  <$17
			DECA
			BNE   LCE92
			LDX   #$0800
			ABX
			LDA   ,X
			PULS  B
			BSR   LCE70						; signed multiply D = A x B
			ADDD  1,S
			STD   1,S
LCEA5		PULS  A
			DECA
			BNE   LCE89
			PULS  A,B,PC

LCEAC		LBSR  CmdTop
LCEAF		CLR   <$5D
			CLR   <$59
LCEB3		LDS   #$00F8

; Initialise hardware
InitHw	ORCC  #$FF
			TFR   CC,DP						; DP points to page $ff
			CLR   <$D8						; set slow MPU rate
			CLR   <$D6						; set slow MPU rate
			CLR   <$C6						; set screen base address to $400
			LDD   #$343C					; set up PIAs
			STA   <$01						; $FF01=%b0011 0100 disable HS irq
												; and make LSB from MUX select lines = %b0
	if INT_DAC == 1					; mod to allow sending sound to internal DAC (by Stewart Orchard)
			sta 	<$03						; $ff03=%b0011 0100 disable FS irq
												; and make MSB from MUX select lines = %b0, so selecting %b00 = 6 bits D/A converter
	else
			STB   <$03						; $ff03=%b0011 1100 disable FS irq
												; and make MSB from MUX select lines = %b1, so selecting %b10 = Cartridge sound
	endif

	if INT_DAC == 1					; mod to avoid sending garbage to the printer (by Stewart Orchard)
			lda 	#$30    					; bit2=0 - select direction register
			sta 	<$21  					;
			lda 	#$fc    					; reset bits 0-1 - casin & strobe as inputs (resistor will keep strobe high)
			sta 	<$20    					;
			lda 	#$34    					; bit2=1 - change back to data register
	endif

			STA   <$21						; $ff21=%b0011 0100 disable CA1 (printer busy)
			STB   <$23						; $ff23=%b0011 1100 disable cartridge irq
												; and enable 6 bit sound (DAC)
			TST   <$02						; clear FS irq flag
			CLR   <$22						; set VDG to flat char screen: 512 bytes
			clra
			CLR   <$48						; disk output latch $ff40
			TFR   A,DP						; DP points to Page0
			CLR   <$A0						; copy of disk output latch value
			RTS

; called by 8x8 signed multiply
LCEDF		MUL
			RTS

; +ve half of sine wave. Amplitude 2A
; length of data
			FCB   $80
; data copied to $0800
LCEE2 	FCB   $00,$01,$02,$03,$04,$05,$06,$07
			FCB	$08,$09,$0A,$0B,$0C,$0D,$0E,$0F
			fcb	$10,$11,$12,$13,$14,$15,$16,$17
			fcb	$18,$18,$19,$1A,$1B,$1C,$1C,$1D
			fcb	$1E,$1F,$1F,$20,$21,$21,$22,$23
			fcb	$23,$24,$24,$25,$25,$26,$26,$27
			fcb	$27,$27,$28,$28,$29,$29,$29,$29
			fcb	$2A,$2A,$2A,$2A,$2A,$2A,$2A,$2A
			FCB	$2A,$2A,$2A,$2A,$2A,$2A,$2A,$2A
			FCB	$2A,$29,$29,$29,$29,$28,$28,$27
			FCB	$27,$27,$26,$26,$25,$25,$24,$24
			FCB	$23,$23,$22,$21,$21,$20,$1F,$1F
			FCB	$1E,$1D,$1C,$1C,$1B,$1A,$19,$18
			FCB	$18,$17,$16,$15,$14,$13,$12,$11
			FCB	$10,$0F,$0E,$0D,$0C,$0B,$0A,$09
			FCB	$08,$07,$06,$05,$04,$03,$02,$01

; +ve half of sine wave. Amplitude 3F
; data copied to $0800
	 		FCB   $00,$02,$03,$05,$06,$08,$09,$0B
	 		FCB	$0C,$0E,$0F,$11,$12,$14,$15,$17
	 		FCB	$18,$1A,$1B,$1D,$1E,$1F,$21,$22
	 		FCB	$23,$25,$26,$27,$28,$29,$2B,$2C
	 		FCB	$2D,$2E,$2F,$30,$31,$32,$33,$34
	 		FCB	$35,$36,$36,$37,$38,$39,$39,$3A
	 		FCB	$3B,$3B,$3C,$3C,$3D,$3D,$3E,$3E
	 		FCB	$3E,$3F,$3F,$3F,$3F,$3F,$3F,$3F
	 		FCB	$3F,$3F,$3F,$3F,$3F,$3F,$3F,$3F
	 		FCB	$3E,$3E,$3E,$3D,$3D,$3C,$3C,$3B
	 		FCB	$3B,$3A,$39,$39,$38,$37,$36,$36
	 		FCB	$35,$34,$33,$32,$31,$30,$2F,$2E
	 		FCB	$2D,$2C,$2B,$29,$28,$27,$26,$25
	 		FCB	$23,$22,$21,$1F,$1E,$1D,$1B,$1A
	 		FCB	$18,$17,$15,$14,$12,$11,$0F,$0E
	 		FCB	$0C,$0B,$09,$08,$06,$05,$03,$02

; data copied to $f00
LCFEB 	FCB   $01,$E0,$F0,$A0,$50
	 		FCB   $40,$40,$00,$00,$E0
	 		FCB   $01,$40,$80,$F0,$80
	 		FCB   $F0,$20,$00,$00,$F0
	 		FCB   $01,$E0,$00,$50,$00
	 		FCB   $F0,$00,$00,$00,$A0
	 		FCB   $01,$F0,$40,$00,$80
	 		FCB   $00,$00,$00,$00,$B0
	 		FCB   $01,$40,$F0,$20,$80
	 		FCB   $10,$40,$00,$00,$D0

; "command mode" command C : Cassette
CmdCass 	CLR 	<FLGDSK					; cassette/disk mode flag
			RTS

; "command mode" command W : Write
CmdWrite	TST   <FLGDSK					; disk mode flag
			BEQ   WriTape					; write file to tape
	 		lbsr	diskCmd					; execute mini-DOS command
			lbra	CmdTop					; comand TOP
WriTape	LDX   <SONGBEG					; start of song data
			BSR  	NewGFN					; get filename
			LBMI  LCBB1
			LDD   <SONGEND					; end of song data
			SUBD  <SONGBEG					; start of song data
			STD   <$0E
			ldx	#$0080					; set default header length (128 bytes)
			stx	<$90						; in system variables
			JSR  [$A00C]					; BASIC vector - Write leader
LD0BE		BSR   LD0ED						; Write block
			LDD   #$50FF					; delay between blocks
LD0C3		DECA								;
			BNE   LD0C3						;
			INCA								; D = $01ff  (data block, length = $ff)
			STD   <$7C						; cassette block type/length
			LDX   <$12						; start of data to write
			LDU   #$0802					; write buffer address
			STU   <$7E						; tape/disk data pointer
			LBSR  LD4B1						; copy B bytes from X to U
			STX   <$12						; start of data to write
			LDD   #$FF01					; -255
			ADDD  <$0E						; subtract 255 from length of data
			STD   <$0E						; length of data to write
			BPL   LD0BE						; more data to write
			ADDD  #$00FF					; correction
			STB   <$7D						; tape block length
			BEQ   LD0E8						; no more data
			TSTA
			BPL   LD0BE						; write last data block
LD0E8		LDD   #$FF00					; zero length EOF block
			STD   <$7C						; Block type & length
LD0ED		JMP  [$A008]					; BASIC vector - Write block

; NEW getFileName using mini-DOS parser (reduced copy of std routine for tape names)
NewGFN	stx	<$12						; save song beginning
			lbsr	getFName					; get entered filename
			lbcs	errorSN					; if error detected, show it
			ldx   <$8a						; point to address $0000
			stx	<$a1						; to clear wildcard flag and drive number
			ldd   #$000f					; filename block, length 15
			std   <$7c						; cassette block type & length
			lda	#$20						; get space
nD00		sta	,x+						; fill one char in filename buffer
			decb								; decrement counter
			bpl	nD00						; not done, loop
			lda	#1							; to form D=#$01ff
			std	<$0b						; put at $0b
			clr	<$0d						; clear $0d (header end)
			ldx	#$0003					; point to first char of filename for tapes
			stx	<$7e						; save pointer for tape
			stb	,x							; put an $ff
			ldb   #$08						; 8 chars to copy
			ldu	#fileN1					; point to parsed name
nD01		lda	,u+						; get a char
			beq	nD02						; if null, skip next
			sta   ,x							; put at destination
nD02		leax	1,x						; advance pointer
			cmpa	#$2a						; was a wildcard?
			bne	nD03						; no, skip next
			sta	<$a1						; save it at $a1
nD03		decb								; decrement counter
			bne   nD01						; not done, loop
			rts								; return

; "command mode" command 0-1-2-3 : Set HDW active MIP slot
; must change *ONLY* LOW nibble because high one is CTS (ROM access)
; that way $ff40-ff5f can be in a different slot
; regA come with $70,$71,$72,$73
CmdSlot	suba	#$70						; convert into 0-1-2-3 (low nibble)
			ldb	>$ff7f					; get actual MPI configuration
			andb	#$f0						; save high nibble
			pshs	b							; onto stack
			adda	,s+						; add high nibble and clean stack
			sta	>$ff7f					; update MPI variable
			bra	cmP01						; exit

; "command mode" command X : Rename
CmdRename
; "command mode" command K : Kill
CmdKill
; "command mode" command > and command < : PROTECT ON - OFF
CmdProt	tst   <FLGDSK					; disk mode flag?
			lbeq 	PrntWhat					; no, display ?? on 2nd line
			lbsr	diskCmd					; execute mini-DOS command
cmP01		lbra	CmdTop					; comand TOP

; "command mode" command V : Verify
CmdVerfy	TST   <FLGDSK					; disk mode flag
	    	LBNE  PrntWhat					; display ?? on 2nd line
			FCB   $21						; BRN	 hides $CC and exposes LDD <$8A

; "command mode" command R : Read
CmdRead	FCB   $CC						; creates LDD #$DC8A with next one
			LDD   <$8A	   				; zero
			LDX   <SONGEND					; end of song data
			BRA   LD174

; "command mode" command A : Append
CmdApend	LDX   <SONGBEG					; start of song data
			CLRA
			INCB
LD174		STD   <$6B						; read & verify flags
			tst	<FLGDSK					; are we loading from disk?
			bne	LD174a					; yes, skip TAPE code
			BSR 	NewGFN					; get filename
			LBMI  LCBB1
LD174a	LDX   #$0802
			STX   <$7E						; tape/disk data pointer
			TST   <$6B						; read flag
			BEQ   LD189						; not reading - skip New

			LBSR  CmdNew					; command N : New
LD189		TST   <FLGDSK					; disk mode flag
			LBNE  LD254						; disk mode read or append
			LDX   #$0420
			LDD   #$0F13
			STB   ,X+
			LBSR  clrALns					; green fill A lines starting from X
LD19A		JSR  [$A004]					; BASIC vector - Read leader
			BSR   LD20A						; Flash cursor, read block, point X to data
			TST   <$7C						; cassette block type
			BMI   LD189						; EOF
			BNE   LD19A						; loop until filename block found
			LDU   #$0422
			LBSR  LD308						; copy 8 characters from X to display at U
			LEAX  -8,X
			LDU   #$0003					; point to specified filename
			LDB   #$08						; 8 characters in filename
LD1B3		LDA   ,U+
			CMPA  #$2A
			BEQ   LD1C0						; no filename was specified. load this file.
			CMPA  ,X+
			BNE   LD19A						; filename doesn't match. get next file.
			DECB
			BNE   LD1B3						; loop for 8 characters in filename
LD1C0		ABX								; move X to 1st byte after filename
			LDU   ,X							; check file type
			CMPU  <$0B						; <$0b = $01ff (data, ASCII)
			BNE   LD19A						; file not correct type, get next file
			TST   <$81						; cassette error flag
			BNE   Xerror7					; Error 7 : File I/O error
			LDU   3,X						; get length of file
			STU   <$0E						; length of data
			LDD   <$12						; load end address
			SUBD  <$0E						; calculate load start address
			BCS   Xerror1					; Error 1 : Memory overflow
			CMPD  #$1064					; lowest allowed address
			BLS   Xerror1					; Error 1 : Memory overflow
			TFR   D,U
			TST   <$6C						; verify
			BEQ   LD1E4						; verify: don't store new start address
			STU   <SONGBEG					; start of song data
LD1E4		CMPU  <SONGBEG					; start of song data
			BNE   Xerror7					; Error 7 : File I/O error
			LDA   #$46   					; 'F'
			STA   $0420						; Put it on the 2nd screen line
LD1EE		STU   <$12						; current load address
			BSR   LD20A						; Flash cursor, read block, point X to data
			TSTA
			BNE   Xerror7					; Error 7 : File I/O error
			TST   <$7C						; cassette block type
			BEQ   Xerror7					; filename block... Error 7 : I/O error
			BMI   LD240						; EOF
			LDU   <$12
			LDB   <$7D						; copy or compare data just loaded
			BEQ   LD1EE
			LBSR  LD495
			BEQ   LD1EE

; Error 7 : File I/O error
Xerror7	LDA   #ERR7						; Error 7 : File I/O error
			BRA   PrntErr					; Error message

; Flash cursor, read block, point X to data
LD20A		LDA   $0420
			EORA  #$40
			STA   $0420
			JSR  [$A006]					; BASIC vector - Read block
			LDX   <$7E						; tape/disk data pointer
			RTS

; Error 1 : Memory overflow
Xerror1	LDA   #ERR1						; Error 1 : Memory overflow
			TST   <$59
			BEQ   PrntErr					; Error message
			CLR   <$56
PrntErr	STA   $043E						; saves error code
			LDD   #$4552					; 'ER'
			STB   $043C						; put second 'R'
			STD   $043A						; put 'ER'
			TST   <$56
			BEQ   LD236
			LEAX  -1,U
LD233		LBSR  LC8E3
LD236		TST   <$59
			BEQ   LD23D
			LBSR  LC90D
LD23D		LBRA  CmdBreak					; "??? mode" command BRK

LD240		LDX   <SONGBEG					; start of song data
			LDA   #$8F
			STA   $0420
			LBRA  LC92D

; Disk mode read or append			; new procedure to work with drives using the mini-DOS
LD254		lbsr	diskCmd					; execute LOAD file
			ldx	<SONGBEG					; get beginning address
			lbsr	LC90F						; update system variables
			lbra	CmdTop					; comand TOP

LD2AC		LDA   ,-U						; copies data from X (loaded) to score (U)
			ORA   #$40
			STA   ,-X
			CMPU  <SONGBEG					; start of song data
			BHI   LD2AC
			RTS

; copy 8 characters from X to display at U
LD308		LDB   #$08
LD30A		LDA   ,X+
			ORA   #$40
			STA   ,U+
			DECB
			BNE   LD30A
LD313		RTS

; "command mode" command D : Dir
CmdDir	INCB
			STB   <FLGDSK					; disk mode flag
			CLR   <$A2						; drive number
			lbsr	diskCmd					; execute DIR
			lda	>pressedKey				; get pressed key
			cmpa	#3							; was it Break?
			lbne	CmdTop					; no, goto command TOP
			rts								; yes, return so that user can read the files

; delay X*8 cycles
LD408		LEAX  -1,X
			BNE   LD408
			RTS

; "command mode" command @ : Pinpoint passage just played
CmdLast	LEAS  2,S
			CLRA
			BSR   LD4DC
			LBRA  LCBE6

; copy or compare data just loaded
LD495		TST   <$6C						; verify flag
			BNE   LD4B1						; copy B bytes from X to U
LD499		LDA   ,U+
			CMPA  ,X+
			BNE   LD4A2
			DECB
			BNE   LD499
LD4A2		RTS

; This routine actually scans the current score line until
; it finds ANY 'first character' in a line, and returns.
; I.E beginning of next line. ALL lines start with the first
; character + 128 ($80)
; though, this does allow it to 'jump' over comment lines
; by the calling routine.
LD4A3		TST   ,U+						; is byte negative?
			BPL   LD4A3						; no, read next
			LEAU  -1,U						; set pointer to that negative byte
			RTS								; return

; copy from X to U.
; length at -1,X. If FLGV5 set then copy from X+length
LD4AA		LDB   -1,X						; get length of 4 voices player
			TST   <FLGV5					; voices flag: 0=4V, 1=5V
			BNE   LD4B1						; copy B bytes from X to U
			ABX								; to skip the 5 voices player and point to the 4 voices player

; copy B bytes from X to U
LD4B1		LDA   ,X+
			STA   ,U+
			DECB
			BNE   LD4B1
			RTS

; "command mode" command S : Score
; This loop only loops through the entire score until "V5"
; is found. If not found, 4 voices is assumed. If found, it jumps
; from the loop to Cmd5V and marks 5 voices.
CmdScore	LDU   <SONGBEG					; start of song data
LD4BB		BSR   LD4A3						; search for 1st 'negative' byte
												; which is the 1st byte in each score line
LD4BD		LDA   ,U+						; get byte
			ANDA  #$7F						; reset bit 7
			BEQ   Cmd5V						; if zero jump to 4 voices
			CMPA  #$6F   					; equals $6F (Comment line)?
			BEQ   LD4BB						; yes, back for another negative byte
			CMPA  #$56   					; equals $56 (Voice)?
			BNE   LD4BD						; no, get another byte
			LDA   ,U							; get next byte
			ANDA  #$7F						; reset bit 7
			CMPA  #$75   					; equals $75 (5 voices)?
			BNE   LD4BD						; no, get another byte

; "command mode" command (74): score 4 voices only
; "command mode" command (75): score all 5 voices
; If 'fall through from 'S' above, will set voice count to
; the highest voice found while scanning (4 or 5)
Cmd4V		SUBA  #$74   					; subtract $74
Cmd5V		STA   <FLGV5					; voices flag: 0=4V, 1=5V
; At this point, 4 or 5 voice has been decided, now the scoring
; is continued to set up the ram buffers and proper play routine.
			CLRA								; A=0
LD4D8		CLR   <$78
			DEC   <$78						; $78 = $FF
LD4DC		STD   <$76						; $76 = $00
			INC   <$56
			LDX   #LD522
			LDU   #$0066
			BSR   LD4AA						; copy bytes from X to U. (Length at -1,X)
			LDX   #LCFEB
			LDU   #$0F00
			LDB   #$64   					; to copy 100 bytes. Changed to avoid losing volume (half) each time we score a song. Was #$32
			LDA   <FLGV5					; get voices flag: 0=4V, 1=5V
			STA   <$AB						; save actual voices flag to previous voices flag: 0=4V, 1=5V
			BSR   LD4B1						; copy B bytes from X to U
			LBSR  LDC88
			EXG   U,X
			LDB   #$39
			LBSR  LC9F6						; clear B bytes starting at X
			LDU   #LD52A
			EXG   X,U
			LDB   #$0A
			BSR   LD4B1						; copy B bytes from X to U
			LDX   #$0F64
			STX   <$51
			STX   2,U
			LEAX  12,X
			STX   <$7A
			STX   ,U
			LDU   <SONGBEG					; start of song data
LD51D		BSR   LD543
			BRA   LD51D

; data copied to $0066
	 		FCB	$04						; number of bytes to copy
LD522		FCB	$55,$2A,$0A,$05		; bytes to copy (4 voices)
			FCB	$7f,$40,$06,$04		; bytes to copy (5 voices)

; 10 bytes copied to ram as part of scoring
LD52A		FCB	$00,$00,$00,$0D,$0D,$0D,$0D,$0D,$02,$60

; translation table
LD534 	FCB   $48,$60					; 'H'
	  		FCB   $49,$18					; 'I'
	 		FCB   $51,$30					; 'Q'
	 		FCB   $53,$0C					; 'S'
	 		FCB   $54,$06					; 'T'
	 		FCB   $57,$C0					; 'W'
	 		FCB   $58,$03					; 'X'
	 		FCB   $FF

LD543		BSR   LD5A5
			TSTA
			LBEQ  LCDAE
			LDX   #LD669					; "programming mode" command lookup table
			LBSR  GetCmd					; command lookup A in table at X
			TST   <$53
			BNE   LD543
			LDX   #LD534
			LBSR  ReadTbl					; translate A using table at X
			BCS   LD58A
			TFR   A,B
			STB   <$0E
			BSR   LD5A5
		   CMPA  #$7A
			BNE   LD576
			ASLB
			BCS   LD580
			CLRA
LD56A		SUBB  #$03
			INCA
			BCC   LD56A
			DECA
			TFR   A,B
			STB   <$0E
LD574		BSR   LD5A5
LD576		CMPA  #$6E
			BNE   LD585
			LSR   <$0E
			ADDB  <$0E
			BCC   LD574
LD580		LDA   #ERR6						; Error 6 : Note Overflow
			LBRA  PrntErr					; Error message
LD585		STB   <$0E
			LEAU  -1,U
			RTS

LD58A		LDB   <$12
			LBSR  LD6B4
			STB   <$6C
			BSR   LD5F1
			BSR   LD5F1
			EORB  #$10
			PSHS  B
			ORB   #$70
			CMPB  ,S
			PULS  B
			BNE   LD5C1
			LDB   <$6C
			BRA   LD5C1
LD5A5		STU   <$62
			LDA   ,U+
			BPL   LD5BB
			LEAU  -1,U
			STU   <$60
			CMPU  <$76
			BNE   LD5B7
			LBSR  LDA11
LD5B7		LDA   ,U+
			ANDA  #$7F
LD5BB		RTS

LD5BC		LEAS  2,S
			FCB   $8C						; CMPX  # - to skip next one

; "programming mode" command (64): '$' could represent a REST (?)
LD5BF 	LDB   #$70
LD5C1		LDA   #$FF
			STA   <$6C
			LDX   <$8A						; zero
			LDA   <$10
			LEAY  A,X
			LDX   ,Y
			STB   <$6B
			LDA   -1,X
			CMPA  <$6B
			BEQ   LD5E3
			ANDA  #$8F
			CMPA  <$6B
			BNE   LD5FB
			LDA   #$60
			ANDA  -1,X
			CMPA  #$60
			BEQ   LD5FB
LD5E3		LDA   -2,X
			ADDA  <$0E
			BCS   LD5FB
			CMPA  #$FF
			BEQ   LD5FB
			STA   -2,X
			BRA   LD605
LD5F1		LDX   #LD659
			BSR   LD646
			BCS   LD5BC
			ORB   -1,X
LD5FA		RTS
LD5FB		LDA   <$0E
			STA   ,X+
			LDA   ,X
			STA   <$6C
			STB   ,X+
LD605		PSHS  X
			LDX   #LD660
			BSR   LD646
			PULS  X
			BCS   LD63B
			LSRA
			BCC   LD61B
			TFR   A,B
			LDA   <$0E
LD617		LSRA
			DECB
			BNE   LD617
LD61B		TFR   A,B
			NEGA
			BEQ   LD63B
			ADDA  -2,X
			STA   -2,X
			LDA   <$12
			ASLA
			BEQ   LD62F
			LDA   -2,X
			STB   -2,X
			TFR   A,B
LD62F		STB   ,X+
			LDA   ,X
			LDB   #$70
			STB   ,X+
			ANDA  <$6C
			STA   <$6C
LD63B		STX   ,Y
			INC   <$6C
			BEQ   LD5FA
			LDA   #ERR5						; Error 5 : Measure overflow
			LBRA  PrntErr					; Error message
LD646		LBSR  LD5A5
			BSR   ReadTbl					; translate A using table at X
			BCC   LD658
			LEAU  -1,U
			RTS

; translate A using table at X
; If match is found, next byte is load into A and returned
ReadTbl	CMPA  ,X++						; compare A with table
			BCS   LD658						; exit if carry set
			BNE   ReadTbl					; no match, loop
			LDA   -1,X						; match found, get previous byte
LD658		RTS								; return

; translation table
LD659 	FCB   $63,$20
	 		FCB   $65,$10
	 		FCB   $66,$40
	 		FCB   $FF

; translation table
LD660 	FCB   $62,$04
	 		FCB   $67,$02
	 		FCB   $6C,$03
	 		FCB   $7B,$05
	 		FCB   $FF

; "programming mode" command lookup table
LD669 	FCB   $40
			FDB	LD6DB						; @
	 		FCB   $4A
	 		FDB	LD7D4						; J
	 		FCB   $4B
	 		FDB	LD72B						; K
	 		FCB   $4D
	 		FDB	LD6F4						; M
	 		FCB   $4E
	 		FDB	LD787						; N
	 		FCB   $4F
	 		FDB	LD79D						; O
	 		FCB   $50
	 		FDB	LD83D						; P
	 		FCB   $52
	 		FDB	LD82D						; R
	 		FCB   $55
	 		FDB	LD6E2						; U
	 		FCB   $56
	 		FDB	LD7C7						; V
	 		FCB   $59
	 		FDB	LD759						; Y
	 		FCB   $5A
	 		FDB	LD7B3						; Z
	 		FCB   $5F
	 		FDB	LD6D8						; shift 0
	 		FCB   $60
	 		FDB	LD786						; space
	 		FCB   $64
	 		FDB	LD5BF						; $
	 		FCB   $68
	 		FDB	LD771						; (
	 		FCB   $69
	 		FDB	LD778						; )
	 		FCB   $6A
	 		FDB	LD6DE						; *
	 		FCB   $6F
	 		FDB	LD4A3						; /
	 		FCB   $7C
	 		FDB	LD7AE						; <
	 		FCB   $7D
	 		FDB	LD7BF						; =
	 		FCB   $7E
	 		FDB	LD798						; >
	 		FCB   $FF

; translation table
LD6AC 	FCB   $6B,$00					; + (treble clef)
	 		FCB   $6D,$80					; - (bass clef)
	 		FCB   $FF

LD6B1		LBSR  LD5A5
LD6B4		PSHS  B
			LDX   #LD6AC
			BSR   ReadTbl					; translate A using table at X
			BCS   LD6C2
			STA   ,S
			LBSR  LD5A5
LD6C2		CMPA  #$47
			BNE   LD6CA
			TST   ,S
			BPL   LD6D6
LD6CA		LBSR  LD7FB
			TSTA
			BNE   LD6D2
			ORA   #$80
LD6D2		ORA   ,S
			STA   ,S
LD6D6		PULS  B,PC

; "programming mode" command shift 0 = next notes are in percussion clef
LD6D8		LDA   #$E0
			FCB   $8C						; CMPX  # - skip the next one

; "programming mode" command @: next notes bass clef
LD6DB 	LDA 	#$80
			FCB	$21						; BRN	 - to skip the next one

; "programming mode" command (6A): '*' = notes play in treble clef
LD6DE 	CLRA
			STA   <$12
			RTS

; "programming mode" command 'U': transpose a signed hex number (up or down)
LD6E2		LDB   #$80
			BSR   LD6B1
			LDX   #$000D
			LDA   <$10
			TSTB
			BPL   LD6F1
			EORB  #$7F
			INCB
LD6F1		STB   A,X
			RTS

; "programming mode" command 'M': define beginning of a measure
LD6F4		PSHS  U
			LBSR  LDB25
			TST   <$17
			BNE   LD708
			LDX   #$0019
			LDU   #$0020
			LDB   #$23
			LBSR  LD4B1						; copy B bytes from X to U
LD708		PULS  U
			CLR   <$17
			FCB	$8C						; CMPX  # - to skip the next one

; find command parameter
; Get's one (A) parameter from Y and returns
; U pointing to the next, or end of line
GetParm 	LEAU  1,Y
LD70F		LDA   ,U+						; find end of command
			BMI   LD71D
			CMPA  #$60
			BNE   LD70F
LD717		LDA   ,U+						; skip over spaces
			CMPA  #$60						; until start of parameter
			BEQ   LD717						; or end of line
LD71D		TST   ,-U
			RTS

; translation table
LD720 	FCB   $48,$01
	 		FCB   $49,$04
	 		FCB   $51,$02
	 		FCB   $53,$08
	 		FCB   $54,$10
	 		FCB   $FF

; "programming mode" command 'K': define key signature
LD72B		LDX   #$0019
			LDB   #$07
			LBSR  LC9F6						; clear B bytes starting at X
			LBSR  LD5A5
			SUBA  #$70
			CMPA  #$08
			BCC   LD793
			TFR   A,B
			LBSR  LD5A5
			CMPA  #$66
			BEQ   LD752
			LEAX  -7,X
			CMPA  #$63
			BNE   LD793
LD74B		DECB
			BMI   LD765
			INC   ,X+
			BRA   LD74B
LD752		DECB
			BMI   LD765
			DEC   ,-X
			BRA   LD752

; "programming mode" command 'Y': set part voice to register
LD759		BSR   LD766
			ADDA  #$0A
			LDB   <$10
			LSRB
			LDX   #$0046
			STA   B,X
LD765		RTS
LD766		LBSR  LD5A5
			SUBA  #$41
			CMPA  #$05
			BCS   LD786
			BRA   LD793

; "programming mode" command (68): '(' = repetition definition beginning
LD771		STU   <$0A
			LDA   #$FF
			STA   <$0C
			RTS

; "programming mode" command (69): ')' = repetition definition end
LD778		LBSR  LD7F9
			TST   <$0C
			BEQ   LD786
			BPL   LD782
			STA   <$0C
LD782		DEC   <$0C
			LDU   <$0A

; "programming mode" command (60) ' ': end of programming sequence
LD786		RTS

; "programming mode" command 'N': define time signature
LD787		LDX   #LD720
			LBSR  LD646
			STA   <$4B
			BCC   LD7AD
			LEAU  1,U
LD793		LDA   #ERR3						; Error 3 : Parameter error
			LBRA  PrntErr					; Error message

; "programming mode" command (7E): '>' = indicates notes to be transposed up
LD798		BSR   LD7F9
			STA   <$14
			RTS

; "programming mode" command 'O': set options
LD79D		BSR   LD7F9
			TFR   A,B
			BEQ   LD7AB
			ORA   <$16
			ANDA  #$01
			ORB   <$17
			ANDB  #$02
LD7AB		STD   <$16
LD7AD		RTS

; "programming mode" command (7C): '<' = indicates notes to be transposed down
LD7AE		BSR   LD798
			NEG   <$14
			RTS

; "programming mode" command 'Z': map stereo channels
LD7B3		BSR   LD7F9
			CMPA  <$68
			BCC   LD793
			LDB   #$05
			MUL
			STB   <$18
			RTS

; "programming mode" command (7D): '=' set tempo
LD7BF		BSR   LD80C
			LSRA
			BEQ   LD793
			STA   <$4C
			RTS

; "programming mode" command 'V': add notes to a voice in a measure
LD7C7		LBSR  LD5A5
			SUBA  #$71
			CMPA  #$05
LD7CE		BCC   LD793
			ASLA
			STA   <$10
			RTS

; "programming mode" command 'J': define named register
LD7D4		BSR   LD766
			LDB   #$0A
			MUL
			LDX   #$0F00
			ABX
			LBSR  LD5A5
			SUBA  #$52
			CMPA  #$02
			BCC   LD7CE
			STA   ,X+
			LDB   #$09
LD7EA		BSR   LD7F9
			ASLA
			ASLA
			ASLA
			ASLA
			STA   ,X+
			DECB
			BNE   LD7EA
			RTS

LD7F6		ANDCC #$FB
			RTS

LD7F9		LDA   ,U+
LD7FB		ANDA  #$3F
			SUBA  #$30
			CMPA  #$0A
			BCS   LD80B
			ADDA  #$2F
			CMPA  #$06
			BCC   LD856
			ADDA  #$0A
LD80B		RTS

LD80C		LBSR  LD5A5
LD80F		LEAU  -1,U
LD811		BSR   LD7F9
			ASLA
			ASLA
			ASLA
			ASLA
			PSHS  A
			BSR   LD7F9
			ORA   ,S+
			RTS

LD81E		LDX   #$0F64
			FCB	$8C						; CMPX   - to skip the next one
LD822		LDX   ,X
			BEQ   LD7F6
			STX   <$7C						; cassette block type/length
			CMPA  2,X
			BNE   LD822
			RTS

; "programming mode" command 'R': repeat named part
LD82D		LBSR  LDAE4
			BSR   LD80C
			BEQ   LD853
			BSR   LD81E
			BNE   LD853
			LDX   10,X
			STX   <$4D
LD83C		RTS

; "programming mode" command 'P': define beginning of named part
LD83D		LBSR  LDAE4
			LBSR  LD5A5
			CLR   <$45
			CMPA  #$60
			BEQ   LD83C
			BSR   LD80F
			STA   <$45
			BEQ   LD83C
			BSR   LD81E
			BNE   LD83C
LD853		LDA   #ERR4						; Error 4 : Invalid part number
			FCB	$8C						; CMPX - to skip the next one
LD856		LDA 	#ERR2  					; Error 2 : Symbol out of context
			LBRA  PrntErr					; Error message

; "command mode" command F : Fast Play
CmdPlayFast
			clr	<$B9						; allow fast mode
			inc	<FLGFAST					; set fast mpu rate flag (PSR 31-03-2016)

; "command mode" command P : Play
; Can be alone or followed by <num> for the part number to start at
CmdPlay	CLR   <$56
			TST   <$5D
			BNE   LD853
			LBSR  GetParm					; find command parameter
			BMI   JmpPlay					; if no more parms, play
			BSR   LD811
			FCB	$21						; BRN   - to skip the next one
JmpPlay	CLRA								; cmds jump here to play without parms
			BSR   LD81E
			BNE   LD853
			LDX   #LDA7A					; frequency table
			LDU   #$089A					; frequency table copied to RAM
			TST   <FLGFAST					; fast mpu rate flag
			BEQ   LD87B
			LEAU  24,U
LD87B		TFR   U,Y
			LBSR  LD4AA						; copy bytes from X to U. (Length at -1,X)
LD880		LDD   22,Y
			LSRA
			RORB
			STD   ,--Y
			CMPY  #$0822
			BNE   LD880
			CLR   ,-Y
			CLR   ,-Y
			LDD   -2,X
			TST   <FLGFAST					; fast mpu rate flag
			BEQ   LD899
			ASRA
			RORB
LD899		STD   ,--Y
			ASRA
			RORB
			ADDD  ,Y
			CMPY  #$0800
			BNE   LD899
			LDD   #$7FEC
			STD   ,Y
			LDU   #PLAYRAM					; player core in RAM
			LDX   #Play5					; player core
			LBSR  LD4AA						; copy bytes from X to U. (Length at -1,X)
			TFR   U,X						; point X to next address after player
			LDU   #LDAAE
LD8B8		LDA   <FLGV5					; voices flag: 0=4V, 1=5V
			ASLA
			LDA   A,U
			PSHS  B
			MUL
			PULS  B
			TST   <FLGFAST					; fast mpu rate flag
			BEQ   LD8C7
			ASLA
LD8C7		STA   B,X
			BNE   LD8CD						; ensure non-zero
			INC   B,X						;
LD8CD		INCB
			BPL   LD8B8						; for B = 0 to 127
			LDY   #$007C
			LDD   #$08FB
			STB   $FF02						; select key col containing BJRZ2:<BRK>
			TST   <FLGFAST					; fast mpu rate flag
			TFR   A,DP						; DP now starts at $0800
			BEQ   LD8E6
			INC   $FFD7						; fast MPU rate
			INC   $FFD9						; fast MPU rate
LD8E6		JMP  	<$B4						; player core in RAM

; 5 voice player core routine copied to $08b4
; number of bytes to copy
PlyLen5	FCB   $9C
; code to copy
Play5		BRA   LD94C
LD8EB		LDA   #$00
LD8ED		LDA   A,X
			STA   <$B2
LD8F1		LDD   <$20
			ADDD  #$0000
			STD   <$BF
			STA   <$EB
			LDD   <$20
			ADDD  #$0000
			STD   <$C8
			STA   <$EE
			LDD   <$20
			ADDD  #$0000
			STD   <$D1
			STA   <$F1
			LDD   <$20
			ADDD  #$0000
			STD   <$DA
			STA   <$F4
			LDD   <$20
			ADDD  #$0000
			STD   <$E3
			STA   <$F7
			LDA   $0A00						; get data for voice 1
			ADDA  $0A00						; add data for voice 2

  if INT_DAC == 1						; mod to send sound to internal DAC (by Stewart Orchard)
			adda 	$0a00						; add data for voice 3
			adda 	$0a00						; add data for voice 4
			adda 	$0a00						; add data for voice 5
			sta 	$ff20						; send to Dragon internal DAC (6 lower bits)
  else
			LDB   $0A00						; get data for voice 3
			ADDB  $0A00						; add data for voice 4
			ADDB  $0A00						; add data for voice 5
			STD   $FF7A						; send to Orchestra L+R DAC registers
  endif

			DEC   <$B2
			BNE   LD8F1
			DEC   <$B3
			BNE   LD940
			DEC   <$B4
			BEQ   LD96A
LD93C		LDA   <$B5
			STA   <$B3
LD940		LDA   $FF00						; read keyboard
			COMA
			ASLA								; lose comparator bit
			BEQ   LD8EB						; none of BJRZ2:<BRK> pressed
			BPL   LD8ED						; one of BJRZ2: pressed
			JMP 	LDD87						; break key handler
LD94C		LDY   ,Y
			BNE   LD954
			JMP  	LCEB3
LD954		LDD   3,Y
			STA   <$EA
			STB   <$ED
			LDD   5,Y
			STA   <$F0
			STB   <$F3
			LDA   7,Y
			STA   <$F6
			LDA   9,Y
			STA   <$B7
			LDU   10,Y
LD96A		PULU  A,B
			STA   <$B5
			INCA
			BEQ   LD94C
			STB   <$BD
			PULU  A,B
			STA   <$C6
			STB   <$CF
			PULU  A,B
			STA   <$D8
			STB   <$E1
			LDA   8,Y
			STA   <$B4
			BRA   LD93C
; end of $9c bytes copied to $08b4

; 4 voice player core routine copied to $08b4
Play4		BRA   LD9F5
LD987		LDA   #$00
			STA   <$B3
LD98B		LDA   $FF00						; read keyboard
			COMA
			ASLA								; lose comparator bit
			BEQ   LD997						; none of BJRZ2:<BRK> pressed
			BPL   LD999						; one of BJRZ2: pressed
			JMP  	LDD87						; break key handler
LD997		LDA   #$00
LD999		LDA   A,X
			STA   <$B2
LD99D		LDD   <$20
			ADDD  #$0000
			STD   <$CF
			STA   <$F2
			LDD   <$20
			ADDD  #$0000
			STD   <$D8
			STA   <$F5
			LDD   <$20
			ADDD  #$0000
			STD   <$E1
			STA   <$F8
			LDD   <$20
			ADDD  #$0000
			STD   <$EA
			STA   <$FB
			LDA   $0A00
			ADDA  $0A00

  if INT_DAC == 1						; mod to send sound to internal DAC (by Stewart Orchard)
			adda 	$0a00
			adda 	$0a00
			sta 	$ff20
  else
			LDB   $0A00
			ADDB  $0A00
			STD   $FF7A						; L+R DAC registers
  endif

			DEC   <$B2
			BNE   LD99D
			DEC   <$B3
			BNE   LD98B
			DEC   <$B4
			BNE   LD987
LD9DC		LDA   8,Y
			STA   <$B4
			PULU  A,B
			STA   <$B7
			INCA
			BEQ   LD9F5
			STB   <$CD
			PULU  A,B
			STA   <$D6
			STB   <$DF
			PULU  A
			STA   <$E8
			BRA   LD987
LD9F5		LDY   ,Y
			BNE   LD9FD
			JMP  	LCEB3
LD9FD		LDD   3,Y
			STA   <$F1
			STB   <$F4
			LDD   5,Y
			STA   <$F7
			STB   <$FA
			LDA   9,Y
			STA   <$C7
			LDU   10,Y
			BRA   LD9DC
; play4 ends here
; extra bytes are copied to ram to match 9C count,
; but not called or used from ram play routine

LDA11		PSHS  A,B,X,Y,U
			LBSR  LDB25
			PULS  A,B,X,Y,U
			CLR   <$53
			TST   <FLGPLAY					; is file to be played?
			BEQ   LDA3D
			RTS

; "command mode" command (61) '!': plays from the cursor position forward
CmdPlyC	INC   <FLGPLAY					; set 'playing' flag
			CLR   <FLGFAST					; clear fast mpu rate flag

; "command mode" command (7F) '?': displays voicing at cursor
CmdVocD	INC   <$53
			LEAS  4,S
			LDD   <$60
			PSHS  A,B
			LBSR  LD4D8
			LDX   ,S
			LBSR  LC934
			TST   <FLGPLAY					; is file to be played?
			BEQ   LDA3A						; no, the skip playing
			LBSR  JmpPlay					; yes, then go play it
LDA3A		LBRA  CmdBreak					; "??? mode" command BRK
LDA3D		LDU   #$000D
			LDY   #$0046
			LDX   #$0420
			LDB   <$69
			PSHS  B
			LDD   #$4374
			ADDD  <FLGDSK					; disk mode flag & ???
			STD   ,X++
LDA52		LDD   #$6059
			STD   ,X++
			LDD   #$3755
			ADDA  ,Y+
			STD   ,X++
			LDB   #$6B
			LDA   ,U++
			BPL   LDA67
			LDB   #$6D
			NEGA
LDA67		STB   ,X+
			ADDA  #$90
			DAA
			ADCA  #$40
			DAA
			ORA   #$40
			STA   ,X+
			DEC   ,S
			BNE   LDA52
			BRA   LDA3A

; frequency lookup data copied to $089a or $08b2
; length of data
			FCB   $1A
; data to be copied
LDA7A 	FDB   $26EA
	 		FDB   $293B
	 		FDB   $2BAF
	 		FDB   $2E48
	 		FDB   $3108
	 		FDB   $33F3
	 		FDB   $370A
	 		FDB   $3A50
	 		FDB   $3DC8
	 		FDB   $4174
	 		FDB   $4558
	 		FDB   $4978
	 		FDB   $004E

; alternate frequency table
	 		FDB   $2008
	 		FDB   $21EF
	 		FDB   $23F4
	 		FDB   $2618
	 		FDB   $285C
	 		FDB   $2AC2
	 		FDB   $2D4D
	 		FDB   $2FFF
	 		FDB   $32D9
	 		FDB   $35DF
	 		FDB   $3914
	 		FDB   $3C78
	 		FDB   $0040

LDAAE 	FDB   $A54A
	 		FDB   $8542

; Stereo mapping table
LDAB2 	FCB   $00,$01,$02,$03,$04
	 		FCB   $00,$02,$01,$03,$04
	 		FCB   $02,$00,$01,$03,$04
	 		FCB   $00,$02,$03,$01,$04
	 		FCB   $02,$00,$03,$01,$04
	 		FCB   $02,$03,$00,$01,$04
	 		FCB   $00,$02,$03,$04,$01
	 		FCB   $02,$00,$03,$04,$01
	 		FCB   $02,$03,$00,$04,$01
	 		FCB   $02,$03,$04,$00,$01

LDAE4		PSHS  U
			BSR   LDB25
			LDU   <$4F
			STU  [$0051]
			STU   <$51
			LDX   #$0043
			LDB   #$03
			LBSR  LD4B1						; copy B bytes from X to U
			LDA   <$18
			LDY   #LDAB2					; Stereo mapping table
			LEAY  A,Y
			LDA   #$05
			STA   <$AC
LDB04		LDA   ,X+
			LDB   ,Y+
			STA   B,U
			DEC   <$AC
			BNE   LDB04
			LEAU  5,U
			LDB   #$04
			LBSR  LD4B1						; copy B bytes from X to U
			LDX   <$7A
			LDD   #$FF0C
			STA   ,X+
			STX   <$4F
			ABX
			STX   <$4D
			STX   <$7A
			PULS  U,PC

LDB25		LDX   <$8A						; zero
			LDU   #$0802
			LDB   #$05
LDB2C		STU   ,X++
			LEAU  66,U
			DECB
			BNE   LDB2C
			TST   <$53
			LBNE  LDC88
			LDY   <$7A
LDB3D		CMPY  <$78
			LBHI  LCEB3
			LEAX  31,Y
			CMPX  <SONGBEG					; start of song data
			BLS   LDB72
			TST   <$59
			BNE   LDB72
			LDX   #MsgOvlap				; "OVERLAP?"
			LDU   #$042E
			LBSR  LD308						; copy 8 characters from X to display at U
			LEAU  1,U
			STU   <$88						; cursor position
			LDA   #$59
			TST   <$57
			BNE   LDB65
			LBSR  LCC09
LDB65		STA   $0437
			pshs	d,x
			lbsr	Clr2Ln					; clear second screen line
			puls	d,x
			LDX   <$60
			CMPA  #$59
			LBNE  LD233
			INC   <$59
LDB72		LEAX  $00A0,Y
			CMPX  <$60
			LBCC  Xerror1					; Error 1 : Memory overflow
			LDD   #$FF05
			PSHS  A,B
			LDX   <$8A						; zero
LDB83		LDA  [,X++]
			CMPA  ,S
			BCC   LDB8B
			STA   ,S
LDB8B		DECB
			BNE   LDB83
			PULS  A,B
			CMPA  #$FF
			LBEQ  LDC85
			STA   ,Y+
LDB98		PSHS  A,B
			LDB   #$05
			SUBB  1,S
			ASLB
			STB   <$10
			LDX   <$8A						; zero
			ABX
			LDU   ,X
			LDD   ,U
			CMPA  #$FF
			BEQ   LDBB8
			SUBA  ,S
			STA   ,U
			BNE   LDBB6
			LEAU  2,U
			STU   ,X
LDBB6		CMPB  #$70
LDBB8		BEQ   LDC31
			PSHS  B
			ANDB  #$70
			CMPB  #$60
			PULS  B
			BEQ   LDC2B
			STB   <$6B
			ANDB  #$8F
			BNE   LDBCC
			ORB   #$10
LDBCC		BPL   LDBD1
			EORB  #$7F
			INCB
LDBD1		ADDB  #$0E
			LDX   #$000D
			LDA   <$10
			ADDB  A,X
			FCB	$8C						; CMPX - to skip the next one
LDBDB		SUBB 	#$0E
LDBDD		ADDB  #$07
			BMI   LDBDD
			CMPB  #$2A
			BCC   LDBDB
			LDX   #LDC5B
			ABX
			PSHS  X
LDBEB		SUBB  #$07
			BPL   LDBEB
			LDX   #LDC5B
			LDB   B,X
			PSHS  B
			LDB   <$16
			BEQ   LDC00
			LDA   <$10
			ASRA
			LDB   #$07
			MUL
LDC00		ADDB  ,S+
			LDX   #$0020
			ABX
			LDA   #$70
			ANDA  <$6B
			BEQ   LDC16
			LBSR  LCE7C						; signed divide D by 16
			LDU   #LDC4F-1
			LDB   A,U
			STB   ,X
LDC16		LDA   <$14
			ADDA  ,X
			PULS  X
			ADDA  ,X
			FCB	$8C						; CMPX  - to skip the next one
LDC1F		SUBA  #$18
LDC21		ADDA  #$0C
			BLE   LDC21
			CMPA  #$48
			BCC   LDC1F
			BRA   LDC32
LDC2B		TFR   B,A
			ANDA  #$0F
			COMA
			FCB	$21						; BRN  - to skip the next one
LDC31		CLRA
LDC32		ASLA
			ADDA  #$20
			LDB   <$10
			ASRB
			ADDB  <$18
			LDX   #LDAB2					; Stereo mapping table
			LDB   B,X
			STA   B,Y
			PULS  A,B
			DECB
			LBNE  LDB98
			LDA   <$69
			LEAY  A,Y
			LBRA  LDB3D

; data
LDC4F		FCB	$00,$01,$02,$FF,$FE,$01
			FCB	$03,$05,$00,$02,$04,$06
LDC5B		FCB 	$F5,$F7,$F9,$FA,$FC,$FE
			FCB 	$00,$01,$03,$05,$06,$08
			FCB	$0A,$0C,$0D,$0F,$11,$12
			FCB	$14,$16,$18,$19,$1B,$1D
			FCB	$1E,$20,$22,$24,$25,$27
			FCB	$29,$2A,$2C,$2E,$30,$31
			FCB 	$33,$35,$36,$38,$3A,$3C

LDC85		STY   <$7A
LDC88		LDA   #$06
LDC8A		STA   <$AC
			CLR   <$10
			LDX   #$0800
			LDU   <$8A						; zero
LDC93		LDD   #$FF60
			STD   ,X++
			DEC   <$AC
			BEQ	LDCAB
			STX   ,U++
			LDB   #$40
LDCA0		STA   ,X+
			DECB
			BNE   LDCA0
			BRA   LDC93
LDCA7		LDA   ,X+
			STA   ,U+
LDCAB		RTS

; "command mode" command O : Optimise
CmdOptmz	LDX   <SONGBEG					; start of song data
			TFR   X,U
			FCB	$8C						; CMPX  - to skip the next one
LDCB1		BSR 	LDCA7
LDCB3		BSR   LDCA7
			BPL   LDCCE
			ANDA  #$7F
			LBEQ  LDD77
			CMPA  #$60
			BNE   LDCCE
			LEAU  -1,U
			TST   -1,X
			BPL   LDCB3
			ROL   ,X
			COMA
			ROR   ,X
			BRA   LDCB3
LDCCE		CMPA  #$6F
			BEQ   LDCE2
			CMPA  #$4D
			BEQ   LDCEA
			CMPA  #$50
			BEQ   LDCB1
			CMPA  #$60
			BNE   LDCB3
			LEAU  -1,U
			BRA   LDCB3
LDCE2		LDA   ,X
			BMI   LDCB3
			BSR   LDCA7
			BRA   LDCE2
LDCEA		LDA   ,X
			BMI   LDCB3
			BSR   LDCA7
			CMPA  #$60
			BEQ   LDCB3
			LEAU  -1,U
			BRA   LDCEA
LDCF8		BSR   LDD08						; display character in B at X
LDCFA		LDB   ,-U
			BPL   LDCF8
LDCFE		TST   <$14
			BEQ   LDD4B
			BSR   LDD4D
			BEQ   LDCFE
			STB   <$18
LDD08		LBRA  LDF7C						; display character in B at X

LDD0B		TST   <$14
			BNE   LDD40						; screen colour set 0
			INC   <$14
			LDU   #MsgOvlap				; "OVERLAP?"
			BSR   LDCFA
			LDB   <$14
			BEQ   LDD40						; screen colour set 0
			LDX   <SONGBEG					; start of song data
			STX   <$04
			BRA   LDD3D						; screen colour set 1
			TST   <$14
			BMI   LDD40						; screen colour set 0
			RTS
LDD25		TST   <$14
			BNE   LDD40						; screen colour set 0
			DEC   <$14
			BRA   LDD31
			TST   <$14
			BNE   LDD4B
LDD31		LBSR  LC90D
			LDX   #$1064
			STX   <$02
			STX   <$12
			STB   ,X

; screen colour set 1
LDD3D		LDA   #$08
			FCB	$8C						; CMPX  - to skip the next one

; screen colour set 0
LDD40		LDD 	<$8A						;zero
			STA   $FF22
			STB   <$14
LDD47		LDB   #$0D
			BSR   LDD08						; display character in B at X
LDD4B		CLRB
			RTS

LDD4D		LBSR  PollKey					; scan keyboard
			BEQ   LDD4B
			TSTB
			BMI   LDD5C
			CMPB  #$60
			BCS   LDD5B
			SUBB  #$40
LDD5B		RTS
LDD5C		ASLB
			BEQ   LDD25
			BPL   LDD0B
LDD61		ORCC  #$FF
			BSR   LDD47
			STX   <$6D
			LDX   <SONGBEG					; start of song data
			LDU   #$1064
			CMPU  <$02
			BCC   LDD7E
			STU   <SONGBEG					; start of song data
			LDU   <$02
			LEAU  1,U
LDD77		LDX   <SONGEND					; end of song data
			LEAX  1,X
			LBSR  LD2AC						; copy score to RAM
LDD7E		LBSR  LC92D						; update variables
			TST   <$70						; transfer mode flag
			BEQ   CmdBreak					; "??? mode" command BRK
			CLR   <$70						; transfer mode flag

; break key handler (during playback)
LDD87		STU   $0078

; soft reset handler
LDD8A		NOP
			LDS   #$00FA
			LBSR  InitHw					; initialise hardware
			LBSR  Clr2Ln					; green fill 2nd line
			LDX   <SONGBEG					; start of song data
			TST   <$70						; transfer mode flag
			BNE   LDD61
			LDX   #LDD8A					; soft reset handler
			STX   <$72						; soft reset vector
			LDB   #$55
			STB   <$71						; cold boot flag

; "command mode" command BRK: while playing  = Stops the music
CmdBreak	CLR   <$58
			LBRA  ReInit					; reinit the system

; irq handler
IrqHndl	LDB   2,Y
			LSRB
			LDD   <$0E
			BMI   LDED5
			BNE   LDED9
			BCS   LDEF3
			LDD   #$0827
			BRA   LDEF1
LDED5		BCC   LDEF3
			BRA   LDEF0
LDED9		DECB
			BNE   LDEF1
			ROR  [$000A]
			LDB   #$1A
			DECA
			BNE   LDEF1
			LDX   <$0A
			ASL   ,X
			CLRB
			ROR   ,X
			INC   <$0B
			INC   <$00
LDEF0		COMA
LDEF1		STD   <$0E
LDEF3		LDD   <$10
			BEQ   LDF0B
			DECB
			BNE   LDF19
			COMB
			ROR  [$000C]
			ROLB
			ROLB
			ANDB  #$02
			STB   ,Y
			CLRB
			DECA
			BEQ   LDF19
			BRA   LDF17
LDF0B		TST   <$01
			BEQ   LDF1B
			INC   <$0D
			DEC   <$01
			LDA   #$0B
			CLR   ,Y
LDF17		LDB   #$1A
LDF19		STD   <$10
LDF1B		LDA   $FF00
			STA   <$6A
			RTI

; clear screen
ClrScrn	LDX   #$0600
			STX   <$88						; cursor position
			STX   <$6D
			STX   <$AD						; top of screen
			BRA   LDFF4						; clear from X to end of screen

; display character in B at X
LDF7C		BSR   LDF99						; invert cursor
			CMPB  #$20
			BLO   LDFC7						; handle control character
			CMPB  #$60
			BLO   LDF88
			SUBB  #$20
LDF88		ANDB  #$BF
			STB   ,X+
			ORB   #$40
LDF8E		CLR   <$17
			DEC   <$17
LDF92		STX   <$88						; cursor position
			CMPX  #$0800					; bottom of screen?
			BHS   LDFE3						; scroll (insert new line at bottom of screen)

; invert cursor
LDF99		LDX   <$88						; cursor position
			LDA   ,X
			EORA  #$40
			STA   ,X
			RTS

; handle TAB
LDFA2		EXG   X,D
			ANDB  #$F8
			EXG   X,D
			LEAX  8,X
			BRA   LDF8E

; handle LF
LDFAC		INC   <$16
			BEQ   LDF99						; invert cursor
LDFB0		INC   <$17
			LDA   <$89
			ANDA  #$1F
			ORA   <$17
			BEQ   LDF99						; invert cursor
			LEAX  33,X

; handle backspace
LDFBD		CMPX  <$AD						; top of screen
			BLS   LDF99						; invert cursor
			LDA   #$20
			STA   ,-X
			BRA   LDF92

; handle control character
LDFC7		CMPB  #$09						; TAB?
			BEQ   LDFA2						; handle tab
			CMPB  #$0A						; LF?
			BEQ   LDFAC						; handle LF
			CMPB  #$08						; backspace?
			BEQ   LDFBD						; handle backspace
			CMPB  #$0D						; CR?
			BNE   LDF99						; invert cursor
			EXG   X,D
			ANDB  #$E0
			EXG   X,D
			CLR   <$16
			DEC   <$16
			BRA   LDFB0

; scroll (insert new line at bottom of screen)
LDFE3		LEAX  -32,X
			STX   <$88						; cursor position
			LDX   <$AD						; top of screen
LDFEA		LDA   32,X
			STA   ,X+
			CMPX  #$07E0					; start of bottom line?
			BLO   LDFEA

; clear from X to end of screen
LDFF4		LDA   #$20
LDFF6		STA   ,X+
			CMPX  #$0800
			BCS   LDFF6
			BRA   LDF99						; invert cursor

LDFFF		FCB   $FF

; =========================================================================================================
; ***  E  N  D      O  F     O  R  C  H  E  S  T  R  A  9  0 - C C    adapted to Dragon Kbd & DAC  ***
; =========================================================================================================
; for fast positioning  ###  -  ADDED FOR DRAGON
; =========================================================================================================
													; fixed areas in use
table  					equ   $02f0          ; Table with all needed sectors for the requested file
readBuffer				equ	$0800				; the disk buffer will be: $0800-$08ff

dp_vars					equ 	$0378				; miniDOS direct page variables ($0378-$03e8)
fileN1					equ 	dp_vars+$00		; (8 bytes) first filename entered
fileE1					equ 	dp_vars+$08		; (3 bytes) extension for 1st filename
file1Null				equ 	dp_vars+$0b		; (1 bytes) null terminator
fileN2					equ 	dp_vars+$0c		; (8 bytes) second filename entered
fileE2					equ 	dp_vars+$14		; (3 bytes) extension for 2nd filename
file2Null				equ 	dp_vars+$17		; (1 bytes) null terminator
wildCard					equ 	dp_vars+$18		; (1 bytes) wildcard found flag
songBegin				equ 	dp_vars+$19		; (2 bytes) beginning of song in RAM
reqNumSector			equ 	dp_vars+$1b		; (1 bytes) Number of sectors needed
maxSect					equ 	dp_vars+$1c		; (2 bytes) highest number of sector in that disk
oldFile					equ 	dp_vars+$1e		; (1 bytes) flag for old file found
oldBak					equ 	dp_vars+$1f		; (1 bytes) flag for old backup file found
bytesLastSect			equ 	dp_vars+$20		; (1 bytes) number of bytes in last sector
sectorsToFree			equ 	dp_vars+$21		; (1 bytes) sectors of BAK file to be freed
neededSect				equ 	dp_vars+$22		; (1 bytes) number of sectors needed to save the file
numSecFound				equ 	dp_vars+$23		; (1 bytes) number of sectors found by now
firstDirSect			equ 	dp_vars+$24		; (2 bytes) LSN number of 1st sector of the directory (FAT)
numEntries				equ 	dp_vars+$26		; (1 bytes) number of needed directory entries
retryCounter			equ 	dp_vars+$27		; (1 bytes) number of retries for a disk command
fatReqSector			equ 	dp_vars+$28		; (1 bytes) FAT sector number to work with
firstFatByte			equ 	dp_vars+$29		; (1 bytes) byte number inside FAT sector
algorithm3				equ 	dp_vars+$2a		; (1 bytes) flag for search by algorithm number 3
rwndCounter				equ 	dp_vars+$2a		; (1 bytes) flag for already rewinded before (overloaded BYTE)
firstTime				equ 	dp_vars+$2b		; (1 bytes) flag to be used with LOAD to detect the very 1st sector
dataPointer				equ 	dp_vars+$2c		; (2 bytes) pointer to DATA in RAM
zero16bits				equ 	dp_vars+$2e		; (2 bytes) 16 bit zero value
dirPointer				equ 	dp_vars+$30		; (2 bytes) entry pointer inside a directory sector
numTabEnt				equ 	dp_vars+$32		; (1 bytes) counter of records in table tabEntries
tabEntries				equ 	dp_vars+$33		; (16 bytes) table 8 words
rwFlag					equ 	dp_vars+$43		; (1 bytes) read / write flag
latchValue				equ 	dp_vars+$44		; (1 bytes) disk output latch value $ff40
fatSector				equ 	dp_vars+$45		; (1 bytes) FAT sector being processed
driveNumber				equ 	dp_vars+$46		; (1 bytes) driver in use
diskTrack				equ 	dp_vars+$47		; (1 bytes) working track number
diskSector				equ 	dp_vars+$48		; (1 bytes) working sector number
diskErrorFlags			equ 	dp_vars+$49		; (1 bytes) error code in disk operations
dirSector				equ 	dp_vars+$4a		; (1 bytes) number of last directory sector accessed
secNameFlag				equ 	dp_vars+$4b		; (1 bytes) flag for 2nd filename control
nmiFlag					equ 	dp_vars+$4c		; (1 bytes) flag for NMI
diskModeFlag			equ 	dp_vars+$4d		; (1 bytes) operation mode (0=Tape, 1=Disk)
addedEntries			equ 	dp_vars+$4e		; (1 bytes) number of entries in the table
diskGeoData				equ 	dp_vars+$4f		; (2 bytes) here go NumTracks and SectorsPerTrack (disk geometry data)
screenBottom			equ 	dp_vars+$51		; (2 bytes) bottom of text screen
firstByte				equ 	dp_vars+$53		; (1 bytes) first byte to 'free' bits in the FAT
diskSide					equ 	dp_vars+$54		; (1 bytes) to save the side of the disk to access
ramDestPtr				equ 	dp_vars+$55		; (2 bytes) pointer to RAM destination
leadDriveFlag			equ 	dp_vars+$57		; (1 bytes) flag for leading drive number
secNumBytes				equ 	dp_vars+$58		; (1 bytes) number of bytes to use from read sector
colonFndFlag			equ 	dp_vars+$59		; (1 bytes) flag to remember a colon has been found
pressedKey				equ 	dp_vars+$5a		; (1 bytes) key pressed by user
sequenFlag				equ 	dp_vars+$5b		; (1 bytes) sequential processing flag
sequenDrive				equ 	dp_vars+$5c		; (1 bytes) drive with last read pointer
sequenSector			equ 	dp_vars+$5d		; (1 bytes) sector number when working sequentially
sequenEntry				equ 	dp_vars+$5e		; (1 bytes) entry number inside a sector when working sequentially
orcNMIhook				equ 	dp_vars+$5f		; (3 bytes) full NMI hook received when Orchestra calls mini-DOS
orcDataPia0				equ 	dp_vars+$62		; (2 bytes) config data for PIA0
orcDataPia1				equ 	dp_vars+$64		; (2 bytes) config data for PIA1
orcComand				equ 	dp_vars+$66		; (1 bytes) command received by Orchestra90
tablePointer			equ 	dp_vars+$67		; (2 bytes) pointer to table
oldPointer				equ 	dp_vars+$69		; (2 bytes) pointer to next parameter
oldBeginning			equ 	dp_vars+$6b		; (2 bytes) pointer to beginning of previous loaded song
drivesCurTracks		equ 	dp_vars+$6d		; (4 bytes) current tracks for drives 1-4 (fc-fd-fe-ff)
													; total USED bytes = 113 ($71) using 00-70
dp_tabs					equ	$0900
tabCopy					equ 	dp_tabs+$00		; (101 bytes) copy of table $2f0
tabEntCopy				equ 	dp_tabs+$65  	; (17 bytes: 1 + 16) copy of numTabEnt and table tabEntries
													; total extra USED = 118 ($76) using $900-$975
; ---------------------------------------------------------------------------------------------------------
; Customized Initialise
; ---------------------------------------------------------------------------------------------------------
LDF21X	ldx	#$0000					; get 16 bit zero
			stx	>sequenSector			; set last sector / entry read = 0-0
			stx	>sequenFlag				; set flag and drive to none
												; sector should range 3-18 and entry should be 1-10 (decimal numbers)
			lds	#$00FA					; new stack value
; initialise track table to force restore on first access
			lda	#-1						; set negative
			tfr	a,b						; values
			std	>drivesCurTracks		; to current track
			std	>drivesCurTracks+2	; for all 4 drives
; other initialization tasks
			lbsr  InitHw					; initialize hardware
			STD   <FLGV5					; (A = 0, B points to slot 4)
			CLRB								; get zero value
			STD   <$70						; transfer mode flag & ??
			STD   <$A8						; nmi flag & disk mode flag
			inc	<$a9						; set disk mode ON
			STD   <$8A						; zero
; set new IRQ Handler
			LDA   #$7E						; jmp opcode
			LDX   #IrqHndl					; point to new IRQ handler
			LDU   #$010F					; point to hook for IRQ
			PSHU  A,X						; setup new IRQ handler
			LBSR  ClrScrn					; clear screen
			INC   <$88						; increment cursor position
			LBSR  LDD3D						; set screen colour to set 1 (orange)
			ldu	#Title					; point to intro message
			ldx	#$06a0					; new value for the text
			stx	<$88						; to gain three screen text lines
			bra	LDF46XR					; only way to skip next three bytes!
LDF46X	LBSR	LDF7C						; display character in B at X
LDF46XR	LDB   ,U+						; get a char from message
			BPL   LDF46X					; if not negative show it
			LDX   <$AD						; get top of screen
LDF4EX	LDA   ,X+						; get char from screen
			ORA   #$40   					; set bit 6 (uppercase)
			STA   -513,X					; copy it one text screen before
			CMPX  #$0800					; go to end of 'second' text screen?
			BNE   LDF4EX					; no, loopback
											; create a delay to avoid the credits screen being skipped
			pshs	x							; save working register
			ldx	#$bfff					; set a delay
1			leax	-1,x						; decrement counter
			bne	1b							; if not zero, loop
			puls	x							; restore register
			LBSR  LCC09						; others and wait a keypress returned in regA
			clr	<$b9						; by default high speed
			cmpa	#$1d						; is it Shift+Enter (in a real Dragon)?
			beq	2f							; yes, skip controls
			cmpa	#$5f						; is it Shift+Enter (in XRoar)?
			beq	2f							; yes, skip next
			inc	<$b9						; set slow speed
2			LBSR  ClrScrn					; clear screen
			LDX   #DemoSong				; demo song
			LDU   #Return+1				; point to byte past end of song
			STX   <SONGBEG					; start of song data
			INC   <$70						; transfer mode flag
			LBRA  LDD77						; continue Orchestra90 initialization
; ---------------------------------------------------------------------------------------------------------
; show allowed commands
; -----------------------------------------------------------------------------------------------------------------
CmdHelp	pshs	a,b,x,y,u				; save registers
			ldy	#helpScr					; get text address
			ldx	#$0400					; point to destination screen
1			ldu	,y++						; get a word
			stu	,x++						; put on screen
			cmpx	#$0600					; go to the end
			bne	1b							; no, loop
											; wait for NO key pressed
relKey	clra								; to read the eight columns
			sta	$ff02						; see all columns
			lda	$ff00						; read all
			coma								; invert bits
			anda	#$7f						; get rid of Joystick bit
			bne	relKey					; any key pressed, wait for it to be released
											; wait for a key
noKey		sta	$ff02						; set PIA0 B-side
			lda	$ff00						; read PIA0 A-side
			coma								; invert bits
			anda	#$7f						; get rid of Joystick bit
			beq	noKey						; if none was zero (now '1') loop
											; set SAM back to the std text screen
			lbsr	Clr2Ln					; clear second screen line
			puls	a,b,x,y,u				; restore registers
			lbra	CmdTop					; go show screen with score

; ----------------------------------------------------------------------------------------------------------------
			if (*%512 > 0)
					rzb	 512-(*%512)	; make sure the text begins in a $200 boundary page (most probably $DE00)
			endif
; -----------------------------------------------------------------------------------------------------------------
; HELP SCREEN (16 lines x 32 chars)
; -----------------------------------------------------------------------------------------------------------------
helpScr	FCC	"@"-64,96,"LAST",96,"PLD",96,"NOTE",96,"R"-64,96,"READ",96,96,96,96,96,96,96,96,96,96
			FCC	"A"-64,96,"APPEND",96,96,96,96,96,96,96,96,"S"-64,96,"SCORE",96,96,96,96,96,96,96,96,96
			FCC	"B"-64,96,"GO",96,"BOTTOM",96,96,96,96,96,"T"-64,96,"GO",96,"TOP",96,96,96,96,96,96,96,96
			FCC	"C"-64,96,"CASSETTE",96,"MODE",96,"V"-64,96,"VERIFY",96,"("+64,"TAPE",")"+64,96
			FCC	"D"-64,96,"DIR",96,96,96,96,96,96,96,96,96,96,96,"W"-64,96,"WRITE",96,96,96,96,96,96,96,96,96
			FCC	"E"-64,96,"EDIT",96,96,96,96,96,96,96,96,96,96,"X"-64,96,"RENAME",96,96,96,96,96,96,96,96
			FCC	"F"-64,96,"PLAY",96,"2"+64,"XSPEED",96,96,"!",96,"PLAY",96,"FROM",96,"POS",96
			FCC	"G"-64,96,"GET",96,96,96,96,96,96,96,96,96,96,96,"-",96,"FIND",96,"UP",96,96,96,96,96,96,96
			FCC	"H"-64,96,"HELP",96,96,96,96,96,96,96,96,96,96,"/",96,"FIND",96,"DOWN",96,96,96,96,96
			FCC	"K"-64,96,"KILL",96,96,96,96,96,96,96,96,96,96,"0123",96,"SET",96,"SLOT","1"+64,"-"+64,"4"+64
			FCC	"L"-64,96,"LIST",96,"TO",96,"LPT",96,96,96,"4",96,"SCORE",96,"4"+64,96,"VOICES"
			FCC	"M"-64,96,"MULTI",96,96,96,96,96,96,96,96,96,"5",96,"SCORE",96,"5"+64,96,"VOICES"
			FCC	"N"-64,96,"NEW",96,96,96,96,96,96,96,96,96,96,96,"<",96,"PROTECT",96,"OFF",96,96,96
			FCC	"O"-64,96,"OPTIMIZE",96,96,96,96,96,96,">",96,"PROTECT",96,"ON",96,96,96,96
			FCC	"P"-64,96,"PLAY",96,"1"+64,"XSPEED",96,96,"?",96,"SHOW",96,"VOICING",96,96
			FCC	"Y"-64,96,"*"+64,96,"ALL","-"+64,"ONCE",96,96,96,96,96,96,"P"-64,"R"-64,"E"-64,"S"-64,"S"-64,32,"A"-64,"N"-64,"Y"-64,32,"K"-64,"E"-64,"Y"-64,96
; =========================================================================================================
; ===================  HERE COMES THE PART THAT DEALS WITH THE WD2797 FOR DISK COMMANDS ===================
; =========================================================================================================
; DragonDir3r - Routines to LOAD a file from a Dragon Disk
; using Low Level Access v 0.3r - Pere Serrat - 2015-11-14
; ---------------------------------------------------------------------------------------------------------
; Assuming 4 extents in the file entry and four additional entry extensions
; (7 extents each) = 4 + 4*7 = 32  -> 3 bytes-blocks
; this will need as much as 96 bytes plus two Header and three EndOfTable bytes at the end = 101 bytes ($65)
; The sectors table will be then from $2F0 to $354:
; $2F0 the number of records in the table
; $2F1 number of bytes in last sector
; $2F2-$2F3 the first LSN in an extent  -  $2F4 the number of sectors in this extent
;     ... the same repeated as many times as needed ... for instance the very last one could be:
; $34F-$350 the first LSN in last entry -  $351 the number of sectors in last extent
; After the last extent data, an EndOfTable record should be put:
; $352-$353 = $FFFF no more entries
; $354      = $00   no sectors
; ---------------------------------------------------------------------------------------------------------
; Normal DIR entry structure
; byte  0     - flags
; bytes 1-8   - filename padded right with nulls
; bytes 9-11  - extension padded right with nulls
; bytes 12-14 - 1st extent: LSN and number of sectors
; bytes 15-17 - 1st extent: LSN and number of sectors
; bytes 18-20 - 1st extent: LSN and number of sectors
; bytes 21-23 - 1st extent: LSN and number of sectors
; byte  24    - next extent or bytes in last sector
; =========================================================================================================
minRAM	equ	$1064						; this is the value used by Orchestra90

diskCmd	pshs 	cc,dp						; save flags and direct page
			orcc 	#$50						; disable interrupts
			lda 	#dp_vars/256			; calculate miniDOS direct page
			tfr 	a,dp						; set it
			setdp dp_vars/256				; notify the compiler
			ldx 	#$600						; first byte past end of text screen
			stx 	<screenBottom			; save value
; save NMI vector
			lda	$0109						; get 1st byte of hook
			sta	>orcNMIhook				; save at variable
			ldd	$010a						; get address
			std	>orcNMIhook+1			; save at variable
; save pia state
			ldx 	#$ff00					; point to beginning of PIAs area
			lda 	1,x						; get contents of $ff01
			ldb 	3,x						; get contents of $ff03
			std	<orcDataPia0			; save to page0
			lda 	$21,x						; get contents of $ff21
			ldb 	$23,x						; get contents of $ff23
			std	<orcDataPia1			; save to page0
; disable hsync, vsync & ack sources
			ldb 	#$34						; value %00110100
			stb 	1,x						; $ff01 <- #$34
			stb 	3,x						; $ff03 <- #$34
			stb 	$21,x						; $ff21 <- #$34
; clear flags just in case
			lda 	,x							; clear $ff00 - PIA0 side A
			lda 	2,x						; clear $ff02 - PIA0 side B
			lda 	$20,x						; clear $ff20 - PIA1 side A
			lda 	$22,x						; clear $ff22 - PIA1 side B
; initial values
			clra								; get 16 bits
			clrb								; zero value
			sta	<diskSide				; set disk side = 0
			std 	<zero16bits				; initialize 16 bits zero variable
			sta 	<nmiFlag					; clear NMI flag
			sta	<secNameFlag			; clear flag 2nd name
; decode command to be executed
			lda	<orcComand				; get received command
			ldx	#stopMtrY				; get desired return point after command completion
			pshs	x							; push it onto the stack
			ldx	#miniDOS					; point to lookup table
			lbsr	GetCmd					; make a hole for the command address and call Orc90 function
			leas	2,s						; if it gets here, command NOT found, clean stack (should NEVER happen)
stopMtrY	clr	$ff48						; stop spinning disk motor
; restore NMI vector
			lda	>orcNMIhook				; get save value
			sta	$0109						; return to 1st byte of NMI hook
			ldd	>orcNMIhook+1			; get saved address
			std	$010a						; restore into NMI hook
; restore pia state
			ldx 	#$ff00					; point to beginning of PIAS area
			ldd	<orcDataPia1			; get values from page0
			sta 	$21,x						; restore them
			stb 	$23,x						; into PIA1
			ldd	<orcDataPia0			; get values from page0
			sta 	1,x						; restore them
			stb 	3,x						; into PIA0
			puls 	cc,dp,pc					; restore interrupts as they were on entering
; ---------------------------------------------------------------------------------------------------------
; Lookup table for mini-DOS accepted commands and dispatch addresses
miniDOS	fcb	#'A'						; orchestra command for APPEND
			fdb	doRead
			fcb	#'D'						; orchestra command for DIR
			fdb	doDir
			fcb	#'G'						; orchestra command for GET (Read, Score and Play)
			fdb	doRead
			fcb	#'K'						; new added command for KILL
			fdb	doKill
			fcb	#'M'						; orchestra command for MULTI as GET but looping infinitely
			fdb	doRead
			fcb	#'R'						; orchestra command for READ (load)
			fdb	doRead
			fcb	#'W'						; new command for WRITE (save)
			fdb	doWrt
			fcb	#'X'						; new command for RENAME
			fdb	doRnm
			fcb	#'Y'						; new command for ALL-ONCE
			fdb	doRead
			fcb	#'<'+$40					; new command for PROTECT OFF
			fdb	doProt
			fcb	#'>'+$40					; new command for PROTECT ON
			fdb	doProt
			fcb	#$ff						; end of table
; ---------------------------------------------------------------------------------------------------------
getFileName
			ldb	<orcComand				; get received command
			cmpb	#'D'						; is a DIR command?
			beq	getParmY					; yes, get dir params
			lbsr	getFName					; use new 'own' file parser
			bcs	errorSN					; if error detected, show it
			ldb	<orcComand				; get received command
											; command Y must be used with '*'
			cmpb	#'Y'						; is command 'ALL-ONCE'?
			bne	2f							; no, skip section
			lda	fileN1					; get first char of filename
			cmpa	#'*'						; is it '*'?
			beq	1f							; yes, go on
			bra	errorSN					; no, show SN error
											; Kill command cannot work with widcards
2			cmpb	#'K'						; is a KILL command?
			bne	1f							; no, skip control
			tst	<wildCard				; wildcards found?
			bne	errorSN					; yes, error
1			ldb	<wildCard				; get wildcard flag
			stb	<sequenFlag				; save as sequential flag
			rts								; return
getParmY	lbsr	sDrvNum					; get drive number
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; mini-DOS error messages, Dragon-like style
; the stack is not a problem because Orchestra restores is to $FA
; but the mini-DOS should better be aware of these changes! = TO-DO task
; ---------------------------------------------------------------------------------------------------------
errorNE	ldd	#"NE"						; File not found
			bra	erShow					; show error code
errorFE	ldd	#"FE"						; File already exists
			bra	erShow					; show error code
errorSN	ldd	#"SN"						; Syntax Error
			bra	erShow					; show error code
errorOM	ldd	#"OM"						; Out of Memory
			bra	erShow					; show error code
errorDN	ldd	#"DN"						; Drive number
			bra	erShow					; show error code
errorIO	ldd	#"IO"						; input/output
			bra	erShow					; show error code
errorIV	ldd	#"IV"						; not ready
			bra	erShow					; show error code
errorDF	ldd	#"DF"						; disk full
			bra	erShow					; show error code
errorFD	ldd	#"FD"						; full directory
			bra	erShow					; show error code
errorPT	ldd	#"PT"						; File Protection
erShow	sta	$ffd6						; ensure normal speed
			ldx	#$0437					; point to second screen line, right part
			std	,x							; print error code
			leax	3,x						; jump 3 spaces
			ldd	#"ER"						; to put word ERR
			std	,x++						; show two letters
			stb	,x+						; put last one
			lda	#'O'						; to add OR
			std	,x							; show two letters
			clr	<sequenEntry			; clear sequential entry
			clr	<sequenDrive			; clear sequential drive to avoid problems
			clr	>$ff48					; stop spinning motor!
			lbra	CmdBreak					; back to Orchestra Loop
; ---------------------------------------------------------------------------------------------------------
; disk select table for drives 1-4 (value to write to $ff48) for Dragon
; ---------------------------------------------------------------------------------------------------------
driveTab	FCB   $4,$5,$6,$7 			; Values for Dragon (without NMI)
; ---------------------------------------------------------------------------------------------------------
; copy 8 characters from X to display at U
; ---------------------------------------------------------------------------------------------------------
cpy8X2U	LDB   #$08						; number of bytes to copy
cpybX2U	LDA   ,X+						; get char
			bne	2f							; if not zero, skip next
			lda	#$20						; get space
2			ORA   #$40 						; set bit 6
			STA   ,U+						; put on screen
			DECB								; decrement counter
			BNE   cpybX2U					; not ended? loopback
			RTS								; return
; ---------------------------------------------------------------------------------------------------------
; "command mode" commands > < : Protect ON and OFF
; ---------------------------------------------------------------------------------------------------------
doProt	bsr	doSearch					; get the correct entry (without filling table)
			bcs	eDProt					; if failed, exit
			ldb	<sequenFlag				; had the name wildcard?
			bne	errorSN					; yes, NOT accepted, show Syntax error
			lda	,x							; get flags of filename we've found
			ldb	<orcComand				; get received command
			cmpb	#'>'+$40					; is it protect ON command?
			beq	doProtON					; yes, do it
			cmpb	#'<'+$40					; is it protect OFF command?
			bne	eDProt					; no, do nothing and quit
doProtOF	anda	#%11111101				; reset protected bit
			bra	doPUpd					; skip next
doProtON ora	#2							; set protected bit
doPUpd	sta	,x							; update file entry flags
											; write the dir sector just modified
			lbsr	rewDIR					; write dir sector back to update the 'renamed' file
eDProt	rts
; ---------------------------------------------------------------------------------------------------------
; "command mode" command D : Dir
; ---------------------------------------------------------------------------------------------------------
doDir		lbsr	comPart					; get parameters, set variables
			lbsr	stdBhv					; set pointer to first directory entry and read sector
			clr	<pressedKey				; clean pressed key before beginning reading entries
dLoop		lbsr	nxtEntry					; point to next entry
			lbsr	cmpStr					; verify if this entry meets the requested conditions
			bcs	dLoop						; no, get next entry
			lda	,x+						; get file flags, advance pointer
			pshs	a							; save file flags
			bsr	cpy8X2U					; copy 8 characters from X to display at U
			lda	#'.'+$40					; add point between name and extension
			sta	,u+						; on screen
			ldb	#3							; extension number of chars
			bsr	cpybX2U					; copy them to screen
			leau	1,u						; skip 1 extra space
			puls	a							; get file flags
			ldb	#$8f						; get space
			anda	#2							; is file protected?
			beq	doDir01					; no, skip next
			ldb	#$10						; get char for Protected (inverted P)
doDir01	stb	,u							; put on screen
			leau	3,u						; and jump 3 spaces to next column
			CMPU  <screenBottom			; end of text screen?
			BLO   dLoop						; no, get next entry
			JSR   waitKey					; wait for a key
			lda	<pressedKey				; get pressed key
			cmpa	#3							; was it the Break key?
			beq	2f							; yes, end DIR command
			lbsr  dir2Scrn					; continue listing on new screen
			bra	dLoop						; get next entry for new page
2			clr	<sequenEntry			; else reset entry number. No sequence to be continued
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; "command mode" commands A,R,G,M : Append, Read, Get, Multi
; ---------------------------------------------------------------------------------------------------------
doRead	bsr	doSearch					; get the correct entry and create table
			lbcc	ldFile					; if not error, load file
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; search for the entry where the received filename is
; ---------------------------------------------------------------------------------------------------------
doSearch lbsr	comPart					; get parameters, set variables
doSWrite	lbsr	readSE					; set pointer to next one after last read entry and read sector
rLoop		lbsr	nxtEntry					; point to next entry
			lbsr	cmpStr					; compare found filename to requested name
			bcs	rLoop						; if not equal, read next entry
											; added to detect file protection and avoid to KILL the file
			lda	<orcComand				; get received command
			cmpa	#'K'						; is it Kill?
			bne	doS01						; no, skip control
			lda	,x							; get flags of filename we have found
			anda	#2							; is it protected?
			lbne	errorPT					; yes, show error
doS01		lda	<dirSector				; get last read sector
			sta	<sequenSector			; save in sequential sector variable
			lda	<orcComand				; get received command
			cmpa	#'>'+$40					; is it Protect?
			beq	eDo						; don't fill table
			cmpa	#'<'+$40					; is it Protect?
			beq	eDo						; don't fill table
			cmpa	#'W'						; is it Write?
			beq	eDo						; don't fill table
			clr	<numTabEnt				; entry counter to zero
			lbsr	crtTable					; create the Table of sectors to be loaded
												; deleting them if KILL was received
eDo		rts								; return
; ---------------------------------------------------------------------------------------------------------
; "command mode" command K: Kill
; ---------------------------------------------------------------------------------------------------------
doKill	bsr	doSearch					; get the correct entry and create table
			bcs	eDo						; if failed, exit
			ldb	<sequenFlag				; had the name wildcard?
			bne	eDo						; yes, NOT accepted, exit
doKWrite	ldb	#1							; first FAT sector required (most usual)
			stb	<fatSector				; 1st FAT sector to work with
			lbsr	openDir					; read FAT
											; read the table to free each sector in the FAT
			ldx	#table					; point to the sectors table
			ldb	,x++						; get number of sector-chunks to process, point to first chunk
			lbeq	6f							; no, exit (not found)
1			ldd	,x							; get first sector LSN in that chunk
			lsra								;
			rorb								; divide regD by 2
			lsra								;
			rorb								; by 4
			lsra								;
			rorb								; by 8 - now regD contains the 1st FAT byte to be modified
			cmpd	#179						; is byte to modify in FAT sector 1?
			pshs	x							; save pointer to table
			bhi	readFAT2					; no, sector 2 needed
readFAT1 stb	<firstByte				; save that byte
			lda	<fatSector				; get last read FAT sector
			cmpa	#1							; have we read sector 1?
			beq	allOK						; yes, continue
			lbsr	rewFAT					; update modified 2nd sector
			ldb	#1							; ask for FAT sector 1
			bra	readFAT					; go read it
readFAT2 subd	#180						; subtract 180 (2nd sector begins at 0)
			stb	<firstByte				; save that calculated byte
			lda	<fatSector				; get last read FAT sector
			cmpa	#2							; have we read sector 2?
			beq	allOK						; yes, continue
			bsr	rewFAT					; update modified 1st sector
			ldb	#2							; ask for FAT sector 2
readFAT	stb	<fatSector				; update variable last read FAT sector
			lbsr 	openDir					; read requested FAT sector
allOK		puls	x							; restore pointer to table
			ldb	<firstByte				; get 1st FAT byte to mod
			lda	#8							; to calculate sectors
			mul								; D = A*B whole sectors
			pshs	d							; pass onto stack
			ldd	,x							; get first sector
			cmpd	#1439						; is it in 2nd FAT sector? ($59f)
			bls	sameSect					; no, skip next correction
			subd	#1440						; subtract 1440 (in each sector we count 0-1439 bits=LSN)
sameSect	subd	,s++						; subtract calculated and clean stack
			incb								; for 0 beginning
			lda	<firstByte				; now regA=byte number (0-179) and regB=bit in byte (0-7)
											; now set to 1 the bits for the (x+2) number of consecutive sectors
			ldu	#readBuffer 			; point to beginning of FAT
			pshs	b							; save bit number
			tfr	a,b						; pass byte number to low order byte
			clra								; make a 16 bit offset
			leau	d,u						; point to needed FAT byte
			inca								; to set bit 0. A = %00000001
2			dec	,s							; decrement bit number (0-7)
			ble	setBits					; exact bit or negative, exit loop
			lsla								; change mask to next bit
			bra	2b							; loopback
setBits	leas	1,s						; clean stack
sameByte	pshs	a							; save bitmask
			ora	,u							; free that sector number
			sta	,u							; update FAT byte
			puls	a							; restore bitmask
			dec	2,x						; decrement number of consecutive sectors
			beq	4f							; no more, seek next chunk
			lsla								; point to next bit
			bcc	sameByte					; not yet bit #9
			leau	1,u						; point to next FAT byte
											; MUST control if U changes from sector 1 to 2 to read it and save previous
			cmpu	#readBuffer+180		; needs a byte from 2nd FAT sector?
			blo	3f							; no, skip section
			pshs	x							; preserve table pointer. Called SBRs use regX
			bsr	rewFAT					; update modified 1st sector
			ldb	#2							; ask for FAT sector 2
			stb	<fatSector				; update variable last read FAT sector
			lbsr 	openDir					; read requested FAT sector
			ldu	#readBuffer				; point to first byte of next sector
			puls	x							; restore table pointer
3			lda	#1							; mask for bit 0
			bra	sameByte					; goto Byte process
4			leax	3,x						; point to next chunk
			tst	2,x						; are there sectors?
			beq	5f							; no, end procedure
			dec	>table					; decrement chunk counter
			lbne	1b							; not yet done, loop
5			bsr	rewFAT					; update last FAT sector in use
6			rts
; ---------------------------------------------------------------------------------------------------------
; This has to rewrite the modified directory sector (FAT or entries)
; ---------------------------------------------------------------------------------------------------------
rewFAT	ldb	<fatSector				; get last modified FAT sector
			stb	<dirSector				; update variable
rewDIR	lbsr	writeSec					; write sector to disk
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; "command mode" command X: Rename
; ---------------------------------------------------------------------------------------------------------
doRnm		bsr	comPart					; get the user parameters
											; if second filename exists issue error
			lda	#1							; value to
			sta	<secNameFlag			; set flag ON
			bsr	readSE					; set pointer to next one after last read entry and read sector
doR01		lbsr	nxtEntry					; point to next entry
			tst	<secNameFlag			; have we got to the end of directory?
			bmi	1f							; yes, alright, now verify 1st name
			lbsr	cmpStr					; compare found filename to requested name
			bcs	doR01						; if not equal, read next entry
			lbra	errorFE					; file found, not allowed. Error
											; if first filename doesn't exist issue error
1			clr	<secNameFlag			; set flag OFF
			bsr	readSE					; set pointer to next one after last read entry and read sector
doR02		lbsr	nxtEntry					; point to next entry
			lbsr	cmpStr					; compare found filename to requested name
			bcs	doR02						; if not equal, read next entry
doRWrite	lda	,x							; get flags of filename we've found
			anda	#2							; is it protected
			lbne	errorPT					; yes, show error
											; else overwrite filename-extension with second one
			ldu	#fileN2					; point to second filename
			ldb	#11						; char counter
			leax	1,x						; skip flags
doR03		lda	,u+						; get a char
			sta	,x+						; put at DIR entry
			decb								; all done?
			bne	doR03						; no, loop
											; write the dir sector just modified
			bsr	rewDIR					; write dir sector back to update the 'renamed' file
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; common part for DIR and READ commands (get parameters, set variables)
; ---------------------------------------------------------------------------------------------------------
comPart	ldb	#1							; put value 1 in
			STB   <diskModeFlag			; mode flag
			stb	>FLGDSK					; update ORC variable too
			CLR   <driveNumber			; current drivenumber = 0 (for drive 1)
			clr	>table					; no entries = Not Found
			JSR   getFileName				; get filename from Orchestra
; ---------------------------------------------------------------------------------------------------------
getGeoData									; added to get the disk geometry data
			clr	<diskSide				; set side to zero
			ldd	#$0001					; to read sector - sector number is 1 (1st FAT sector)
			lbsr	readBsec					; read requested sector from directory track 20
			ldx	>readBuffer+$fc		; get geometry word
			stx	<diskGeoData			; save in system variables
			rts								; yes return
; ---------------------------------------------------------------------------------------------------------
; read sector and set buffer pointer
; ---------------------------------------------------------------------------------------------------------
readSE	lda	<sequenFlag				; is sequential load active?
			beq	stdBhv					; no, do std behaviour
			lda	<sequenDrive			; get last read drive for sequential purposes
			cmpa	<driveNumber			; is the same we want to read now?
			beq	readS1					; yes, go on
			clr	<sequenEntry			; no, clear that one to prevent sequential using data from another drive
readS1	lda	<sequenEntry			; is there a previous read entry?
			beq	stdBhv					; no, do std behaviour
											; sequential behaviour
seqBhv	ldb	<sequenSector			; get last read sector
			cmpa	#9							; was it the last one in its sector?
			ble	saveSeq					; no, so next one will be in same sector
			clra								; reset entry number to zero
			incb								; increment sector number
											; test to control big disks with directory full (720k with 160 files)
			cmpb	#$12						; past directory (sector 18 is last one)?
			bls	saveSeq					; no, save values to be used
												; else set 1st DIR sector 1st entry to go back to the beginning
			ldb	<orcComand				; get received Orchestra command
			cmpb	#'M'						; is a Multi command?
			lbne	errorNE					; no, show error message
			tst	<rwndCounter			; has it already rewound directory?
			lbne	errorNE					; yes, show not exists error
			inc	<rwndCounter			; increment counter
; ---------------------------------------------------------------------------------------------------------
											; default std behaviour
stdBhv	ldb	#3							; to read first directory sector
			clra								; to read first entry of that sector
											; B should arrive with Sector number and A with entry number (minus one)
saveSeq	stb	<sequenSector			; save sector
			sta	<sequenEntry			; save old entry
			lda	<driveNumber			; get drive to work with
			sta	<sequenDrive			; save as sequential drive
											; read a directory entry and set buffer pointer
			BSR 	openDir					; open disk directory (reading one sector)
			lda	<sequenEntry			; get last entry read
			beq	backOne					; if zero, do not skip entries
corr01	leax	25,x						; skip one entry
			deca								; decrement counter
			bne	corr01					; if not zero, loopback
backOne	leax	-25,x						; point one entry before desired one
			stx	<dirPointer				; update pointer
			bsr	dir2Scrn					; clean 14 lines of screen
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; loop that reads the needed sectors-entries and verifies extension
; ---------------------------------------------------------------------------------------------------------
nxtEntry	BSR   getNEnt					; point to next directory entry in buffer
			lbcs	outPgm					; if got to the end, exit
			LDA   ,X							; get entry flags
nxtE01	bita	#$08						; is end of directory?
			beq	nxtE02					; no, go to next control
tstMulti lda	<orcComand				; get received Orchestra command
			cmpa	#'M'						; was it Multi?
			lbne	outPgm					; no, quit
			tst	<rwndCounter			; has it already rewound the directory?
			lbne	errorNE					; yes, show not exists error
			inc	<rwndCounter			; increment counter
			bsr	stdBhv					; yes, set values to repeat from the beginning
			bra	nxtEntry					; go read next entry
nxtE02	bita  #$01						; is an extent entry?
			bne	nxtEntry					; yes, go to next directory entry
			bita	#$80						; is bit7 = 1 (deleted)?
			bne	nxtEntry					; get next one
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; point to next directory entry in the data buffer (if needed read next sector)
; ---------------------------------------------------------------------------------------------------------
getNEnt	LDX   <dirPointer				; current directory entry in sector buffer
			inc	<sequenEntry			; increment entry counter (accepted values 1-10)
			LEAX  25,X						; Dragon directory entry length
			CMPX  #readBuffer+$f9		; past the 10th entry?
			BLO   1f							; no, save pointer and go on
			INC   <dirSector				; increment current sector on directory track
			LDB   <dirSector				; get current sector on directory track
			CMPB  #$12						; got past the last dir sector?
			bhi   3f							; yes, signal end of directory
			lda	#1							; put 1 as
			sta	<sequenEntry			; entry number of next sector
			CLRA								; flag for READ
			BSR   readA6					; READ next directory sector
1			CLRA								; set Z flag, clear Carry
			STX   <dirPointer				; current directory entry in sector buffer
			RTS								; return
											; this part is added to mark an end of directory
											; when it is full for ALL commands
3			lda	#$89						; set value for END of directory
			leas	2,s						; get rid of last return address
			bra	nxtE01					; go test this condition
; ---------------------------------------------------------------------------------------------------------
; open disk directory					; regB must come with sector number
; ---------------------------------------------------------------------------------------------------------
openDir	CLRA								; A=0 command 'read' sector
			sta	<diskSide 				; the directory is always on side 1 (sectors 1-18)
			BSR  	readBsec					; read sector regB from directory track 20
			STX   <dirPointer				; save current directory entry in sector buffer
endDir	LDA   <diskSector				; sector number that has been read
			STA   <dirSector				; save at current sector on directory track
			RTS								; return
; ---------------------------------------------------------------------------------------------------------
; clean screen (14 lines from third one)
; ---------------------------------------------------------------------------------------------------------
dir2Scrn	LDX   #$0440					; point to 3rd line on screen
			TFR   X,U						; U pointer gets screen address
			LDA   #$0E						; number of lines to clear
			JSR   clrALns					; green fill A lines starting from X
			rts								; return
; ---------------------------------------------------------------------------------------------------------
; READ/write sector to directory track
; ---------------------------------------------------------------------------------------------------------
writeSec	LDA   #$20						; code for op write
; ---------------------------------------------------------------------------------------------------------
; receives in regA the command $00=read $20=write
; ---------------------------------------------------------------------------------------------------------
readA6	LDB   <dirSector				; current sector on directory track
; ---------------------------------------------------------------------------------------------------------
; receives in regB the sector to be read
; ---------------------------------------------------------------------------------------------------------
readBsec	LDX   #readBuffer				; point to data buffer (or data area)
			STA   <rwFlag					; disk r/w flag
			LDA   #$14						; 20 = directory track
			STD   <diskTrack				; disk track & sector numbers (a3 & a4)
			STX   <dataPointer			; tape/disk data pointer
; ---------------------------------------------------------------------------------------------------------
; here arrives with data pointer already set
; ---------------------------------------------------------------------------------------------------------
read2Buf	PSHS  B,X,Y,U					; save working registers
			LDU   #$FF40					; disk registers pointer
			LDB   #$04						; 4 retries
			STB   <retryCounter			; retry counter
retryOp	LDD   <driveNumber			; drive number & track number
			LDX   #driveTab				; point to disk select table (drives 1-4)
			LDA   A,X						; get register value for drive
			CMPB  #$16						; from that track number on, apply precompensation
			BLO   notPreC					; track < 22
			ORA   #$10						; add enable write precompensation
notPreC	STA   8,U						; $ff48 disk output latch DRAGON
			TST   <latchValue				; disk output latch value $ff40
			BNE   noDelay					; if not zero skip delays
			BSR   del8X						; delay X*8 cycles (approx 480ms)
			BSR   del8X						; delay X*8 cycles (approx 590ms)
noDelay	STA   <latchValue				; disk output latch value $ff40
			BSR   waitCplt					; wait for disk op complete
			BNE   tstFlag					; if error skip next one
			CLR   <diskErrorFlags		; no errors, clean error flag
			BSR   seekdoOp					; SEEK track & DO the actual READ/write disk sector
tstFlag	LDA   <diskErrorFlags		; get error flags
			BEQ   endOK						; no error, exit
			DEC   <retryCounter			; decrement retry counter
			LBEQ  errorIO					; if used all retries, error.
			BSR   rstDrive					; restore current disk drive
			BNE   tstFlag					; if error test flags
			BRA   retryOp					; if restored OK, retry operation again
endOK		PULS  B,X,Y,U,PC				; restore regsters and return
; ---------------------------------------------------------------------------------------------------------
; restore current disk drive
; ---------------------------------------------------------------------------------------------------------
rstDrive	LDX   #drivesCurTracks		; point to current tracks for drives 1-4 table
			LDB   <driveNumber			; get current drive num 0-3
			CLR   B,X						; set current track to zero
			LDA   #$03						; disk command: restore
			BSR   dskComd					; exec disk command with timeout
			ANDA  #$10						; save bit5 (as error bit)
			BRA   cancel2					; save error result and exit
; ---------------------------------------------------------------------------------------------------------
; disk command with timeout (approx 1.3s)
; ---------------------------------------------------------------------------------------------------------
dskComd	BSR   wrtCmd					; write command to disk controller
; ---------------------------------------------------------------------------------------------------------
; wait for disk op to complete (approx 1.3s)
; ---------------------------------------------------------------------------------------------------------
waitCplt	ldb	#$02						; 720k disks need more time!
1			ldx	<zero16bits				; counter to max
2			LDA   ,U							; $ff40 disk status DRAGON
			BITA  #$01						; test bit 0
			bne   3f							; if command not yet complete skip next one
			RTS								; return OK
3			LEAX  -1,X						; decrement counter
			bne   2b							; not yet zero, loopback to test again
			decb								; decrement high order counter
			bne   1b							; not yet zero, loopback re-loading X
			bra	cancelOp					; timeout, cancel disk operation ... IF NOT MOVED COULD BE DELETED
; ---------------------------------------------------------------------------------------------------------
; cancel disk operation
; ---------------------------------------------------------------------------------------------------------
cancelOp	LDA   #$D0						; disk command: terminate without interrupt. Force Interrupt
			BSR   wrtCmd					; write command to disk controller
			LDA   ,U							; $ff40 disk status DRAGON
			ORA   #$80						; set bit7 in result
cancel2	STA   <diskErrorFlags 		; save operation result
cancel3	RTS								; return
; ---------------------------------------------------------------------------------------------------------
; approx 81ms delay						; $2345=9029 decimal * 8 cycles = 72,232 cycles + 8 = 72,240
; 												  at 0.89488MHz it gives 80,725 ms  ($2363 -> 80.994ms)
; ---------------------------------------------------------------------------------------------------------
del81ms	LDX   #$2345					; needed value for a 81mseg delay - 3 cycles
; delay X*8 cycles (9 ms)
del8X		LEAX  -1,X						; decrement counter 		- 5 cycles
			BNE   del8X						; if not zero, loopback - 3 cycles
			RTS								; return						- 5 cycles
; ---------------------------------------------------------------------------------------------------------
; write command to disk controller
; ---------------------------------------------------------------------------------------------------------
wrtCmd	ora   <diskSide				; add selected disk side (bit1)
			STA   ,U							; $ff40 disk command DRAGON
			PSHS  A,B						; 7 cycles
			PULS  A,B,PC					; 9 cycles
; ---------------------------------------------------------------------------------------------------------
; seek track & r/w disk sector
; ---------------------------------------------------------------------------------------------------------
initRes	bsr	rstDrive					; restore current disk drive
			bne	cancel3					; if not OK, return

seekdoOp	LDX   #drivesCurTracks		; point to current track for drives 1-4 table
			LDB   <driveNumber			; get current drive num 0-3
			LDA   B,X						; get track for current drive
			bmi initRes						; first access so restore drive
			STA   1,U						; $ff41 disk track DRAGON
			CMPA  <diskTrack				; same as disk track number
			BEQ   1f							; yes, already on the right track
			LDA   <diskTrack				; no, get disk track number
			STA   3,U						; $ff43 disk data DRAGON
			STA   B,X						; update current track number
			LDA   #$17						; disk command: seek track with verify
			BSR   dskComd					; disk command with timeout
			BNE   cancel2 					; disk command failed
			BSR   del81ms					; approx 81ms delay
1			LDB   <diskSector				; disk sector number
			STB   2,U						; $ff42 disk sector DRAGON
			LDX   <dataPointer			; tape/disk data pointer
			LDY   <zero16bits				; Y = #$0000 (max. delay)
			TST   ,U							; $ff40 disk status DRAGON
			LDD   #$8805					; A=read sector command, B=timeout count
			ADDA  <rwFlag					; disk r/w flag (0=r $20=w)
			CMPA  #$A0						; sets Carry flag if read sector
			BSR   wrtCmd					; write command to disk controller
			LDA   #$02						; set bit1
			sta	$ffd7						; switch to double speed to smooth data transfers
			BLO   readSec					; read sector if command written OK
; ---------------------------------------------------------------------------------------------------------
; write sector
; ---------------------------------------------------------------------------------------------------------
wrtSec	ldb	#%00000010				; to verify bit2 (copy of DRQ)
			LDA   ,X+						; get a data byte
2			bitb	,u							; test bit2 from $ff40 (status register)
			bne	3f							; if it is ready, send a byte
			LEAY  -1,Y						; decrement counter
			BNE   2b							; delay not yet elapsed , loop
			sta	$ffd6						; back to normal speed
			BRA   cancelOp					; cancel disk operation
3			STA   3,U						; $ff43 put byte on data register
			lda	,x+						; get a data byte
2			ldb	,u							; get status register $ff40
			bitb	#2							; check for drq
			bne	3b							; if data request, store byte
			bitb	#1							; else check for busy
			bne	2b							; if still busy, wait for next drq
			bra	endWork					; all done, exit sector write
; ---------------------------------------------------------------------------------------------------------
; read sector
; ---------------------------------------------------------------------------------------------------------
readSec	lda	#%00000010				; to verify bit2 (copy of DRQ)
1			bita	,u							; test bit2 from $ff40 (status register)
			bne   2f		     				; if data is ready go read first byte
			LEAY  -1,Y						; decrement counter
			BNE   1b							; delay not yet elapsed, loop
			lda	,u							; get status
			bita	#$10						; check for record not found
			bne	endWork					; exit early to save time
			DECB								; decrement 2nd level timecount
			BNE	readSec		     		; loop again until timeout
			sta	$ffd6						; back to normal speed
			LBRA  errorIV     			; flag ?NR ERROR (using IV instead)
2			LDA   3,U						; get byte, $ff43 disk data DRAGON
			STA   ,X+       				; save byte into buffer
3			lda	,u							; read $ff40 status register
			bita	#2							; check for drq
			bne	2b							; if data request, get byte
			bita	#1							; else check for busy
			bne	3b							; if still busy, wait for next drq
endWork	sta	$ffd6						; back to normal speed
			LDA   ,U							; $ff40 disk status DRAGON
			ANDA  #$7C						; save bits 2-3-4-5-6
			LBRA  cancel2					; save result and return
; ---------------------------------------------------------------------------------------------------------
; wait for a while and then stop disk motor but keep waiting for a key
; ---------------------------------------------------------------------------------------------------------
anyKey	fcn	/press any key/		; message
waitKey  pshs	cc,dp,a,x,y,u			; save some registers
			clrb								; put 0
			stb	<pressedKey				; no pressed key
			tfr	b,dp						; as DP for Basic
												; show message 'Any Key'
			ldx	#anyKey					; point to "File Loaded" message
			ldu	#$0430					; point to 4th screen line
1			lda	,x+						; get a char
			beq	eLoad						; if null, end
			bita	#%01000000				; has bit6 set (uppercase)?
			beq	doNot						; no, it is space, do nothing
			anda	#%10011111				; to show inverted letters
doNot		sta	,u+						; put on screen
			bra	1b							; go for next one
eLoad		lda   #$60						; green semigraphic full bloc
			sta	$41f						; to guarantee always the same switching scheme
			ldy	#$003C					; to control 30 seconds (double blinking)
2			ldx	#$1183					; aprox. 1/2 second
3			pshs	y							; save counter
			jsr	$8006						; get a key
			puls	y							; restore counter
			bne	4f							; if any key pressed, exit
			leax	-1,x						; decrement 1/2 second counter
			bne	3b							; not zero, loopback
			com	$41f						; invert bits (makes flashing cursor)
			leay	-1,y						; decrement seconds counter
			bne	2b							; if not 0, loopback
			clr	$ff48						; stop motor
			ldd	#$0D13					; get "ms" for "Motor Stopped"
			std	$41b						; put on 1st row of screen
			bra	2b							; recharge 1/2 second counter. 'Y' doesn't matter from now on
4			sta	>pressedKey				; save pressed key
			lbsr	Clr2Ln					; clean second line (Press any Key)
			lda	>pressedKey				; get last pressed key
			cmpa	#3							; was it Break?
			beq	oWK						; yes, skip cleaning screen
			LDX	#$0400					; point to beginning of text screen
			lda	#16						; all 16 lines
			lbsr	clrALns					; make a CLS
oWK		puls	cc,dp,a,x,y,u,pc		; restore saved registers and return
; ---------------------------------------------------------------------------------------------------------
; compare filename of this entry to requested filename
; X points to First byte of that entry
; must set Carry if not equal
; ---------------------------------------------------------------------------------------------------------
cmpStr	pshs	d,x,u						; save working registers
			leax	1,x						; skip first byte (entry flags)
			ldu	#fileN1					; point to requested name
			tst	<secNameFlag			; are we dealing with 2nd filename?
			beq	cmpName					; no, skip correcting pointer
			leau	16,u						; point to the second filename
cmpName	ldb	#$08						; length of filename
1			lda	,u+						; get a char from requested name
			cmpa	#'*'						; is a wildcard (*)?
			beq	cmpExt					; return match
			cmpa	,x+						; equal to pointed entry?
			bne	strNotEq					; no, error, quit
			decb								; decrement counter
			bne	1b							; if not 0, compare next char
			bra	noAdj						; is zero, no adjustements needed
cmpExt	leax	b,x						; skip not compared zeros
			decb								; because U goes one char advanced
			leau	b,u						; skip not compared zeros
noAdj		ldb	#$03						; length of extension
2			lda	,u+						; get a char from requested extension
			cmpa	#'*'						; is a wildcard (*)?
			beq	3f							; return match
			cmpa	,x+						; equal to pointed entry?
			bne	strNotEq					; no, error, quit
			decb								; decrement counter
			bne	2b							; if not 0, compare next char
3			andcc	#$fe						; string match, reset carry (could be a simpler CLRA)
			bra	eCS						; exit
strNotEq orcc	#$01						; error, set carry
eCS		puls	d,x,u,pc					; restore registers and return
; ---------------------------------------------------------------------------------------------------------
; End of program for some COMMANDS
; ---------------------------------------------------------------------------------------------------------
outPgm	ldb	<orcComand				; get received command
			cmpb	#'X'						; is a RENAME command?
			bne	1f							; no, skip process
			tst	<secNameFlag			; is flag 2nd name ON?
			beq	1f							; no, skip process
			lda	#$ff						; yes, get $ff
			sta	<secNameFlag			; update the flag
			rts								; return to the caller
1			leas	2,s						; get rid of return address
											; added to let write detect NON existence of files in disk
			cmpb	#'W'						; if it is Write command
			beq	eEndKO					; return with carry set
			cmpb	#'w'						; if it is Write command (subcase)
			beq	eEndKO					; return with carry set
											; only DIR can get to the end of directory without issuing error
			cmpb	#'D'						; is a DIR command?
			lbne	errorNE					; no, show error message
			lda	<pressedKey				; see last key pressed
			cmpa	#3							; was it Break?
			beq	eEndP						; yes, return
			JSR   waitKey					; wait for a key
eEndP		clr	<sequenEntry			; coming from a DIR must reset entry number. No sequence to be continued
			rts								; return
eEndKO	orcc	#1							; set carry flag
			rts								; return error
; ---------------------------------------------------------------------------------------------------------
; Directory entries Process -> to fill the Sectors Table
; X must be pointing to the flags of a file entry
; ---------------------------------------------------------------------------------------------------------
crtTable	lbsr	adTabEnt					; add entryNumber and sector to table tabEntries
			clr	<addedEntries			; entry counter to zero
			ldu   #table					; get beginning of destination table in RAM
			clr	,u+						; no entries in table
			clr	,u+						; no bytes last sector
		 	stu   <tablePointer			; save at pointer
										   ; for each extent in that file entry (four)
		  	lda   14,x						; is num sectors for 1st extent equal 0?
											; this should only happen if the requested file IS NOT in the drive
		 	lbeq	errorIO					; no sectors used, stop and show error message
		            					; add used sectors by this file
       	ldb   #14						; offset to number of sectors in 1st extent
extlop 	subb  #2							; offset to sector LSN
		 	ldy   b,x						; get LSN
		 	sty   ,u++						; put LSN in table
		 	sta   ,u+						; put numSects in table
		 	inc	<addedEntries			; increment counter
		 	addb  #5							; get offset to number of sectors in next extent
		 	cmpb  #23						; is it out of limits (past 4th extent)?
		 	bhi   1f							; yes, exit loop
		 	lda   b,x         			; is number of sectors in next extent equal 0?
       	bne   extlop					; no, back to process next extent
       	bra   hTrail					; out, go add table Head and Trail
		 									; see if there are more extents to analyze
1		 	stu   <tablePointer			; save tabpointer
moreExt	lda	,x							; get flags of this entry
       	bita  #32						; if flags AND 32 = 32 there are more extents
       	beq   hTrail           		; no more extents, end table
	 										; if KILL command set flags = $81 (0,x)
			lda	<orcComand				; get ORC command
			cmpa	#'K'						; is it 'KILL'?
			bne	2f							; no, skip section
			lda	#$81						; get value for Deleted
			sta	,x							; update entry flags
											; Process extra extents. Called with X pointing to old entry
											; calculates the correct entry into the correct dir sector for next extent
2			lda   24,x						; get offset 24 of previous entry as input (next entry to use)
		 	ldb   #1   						; 1st dir sector to begin with
3		  	incb								; increment sector counter
			suba  #10						; subtract 10 entries (1 sector)
		 	beq	4f							; avoid correction
		 	bpl   3b							; if not negative value, subtract another
		 	decb								; decrement sector number
		 	adda  #10						; to get a positive entry number (1-10) where the extents begin
4			addb	#2							; add 2 sectors to skip FAT
			pshs  b							; save new sector to read
		 	ldb   #25						; entry length
		 	mul								; number of bytes to skip over ('A' was entry to find in the calculated sector)
		 	ldx	#readBuffer				; point to beginning of buffer
		 	leax  d,x						; X points to first byte of desired new extents entry (in buffer area)
											; now we need to read the sector number X into std buffer but only if different than actual sector
			puls	b							; get calculated sector
			pshs	x							; save new entry pointer
			cmpb	<dirSector				; is the same sector we are now?
			beq	6f							; yes, don't read sector again
			pshs	b							; save sector number
			lda	<orcComand				; get ORC command
			cmpa	#'K'						; is it 'KILL'?
			bne	5f							; no, skip section
			lbsr	rewDIR					; write old sector back to update the 'deleted' entries found in it
5			puls	b							; restore sector number
			stb	<dirSector				; update variable with sector we are going to read
			clra								; command read sector
			lbsr	readBsec					; read sector from directory track 20
6			puls	x							; get calculated entry data pointer
											; process extents entry from just read sector
		 	lda   ,x							; get entry flags (first byte of entry)
       	bita	#1							; Flags (offset 0) AND 1 must be equal to 1 (extent entry)
											; this should only happen if extended entry is not an extent entry. Disk Error
       	lbeq  errorIO					; else we have a problem and must quit program
		  	                    		; for each one of the 7 extents in this entry
			bsr	adTabEnt					; add entryNumber and sector to table tabEntries
       	ldb   #3							; offset to number of sectors for 1st extent
       	ldu   <tablePointer			; get Table pointer back
7		  	lda   b,x 						; is number of sectors = 0?
       	beq   hTrail					; yes, go add table Head and Trail
		 	subb  #2							; offset to 1st LSN
 		 	ldy   b,x						; get the 1st LSN of the extent
 		 	sty   ,u++						; put into table
 		 	sta   ,u+						; add to table
		 	inc	<addedEntries			; increment counter
		 	addb  #5							; point to next extent
		 	cmpb  #22						; out of range?
		 	blo   7b							; no, process next extent
		 	stu   <tablePointer			; save pointer to table
		 	bra	moreExt					; are there more extents?
											; fill table header and trail
hTrail   lda	<orcComand				; get ORC command
			cmpa	#'K'						; is it 'KILL'?
			bne	8f							; no, skip section
			lda	#$81						; get value for Deleted
			sta	,x							; update entry flags
			lbsr	rewDIR					; write sector back to update the 'deleted' entries found in it
8			ldd	#$ffff					; mark end of table
			std	,u++						; in table trailer
			clr	,u							; and put zero sectors
			lda	<addedEntries			; get	number of entries in table
			ldb	24,x     				; get bytes in last sector of last entry
			std	>table					; put at the table header
			rts								; return
; ---------------------------------------------------------------------------------------------------------
;	Calculate the directory entry number in a sector and save into table
;	input:	regX must point to the flags of that entry
;	returns:	adds to the table tabEntries a new record
;				increments counter <numTabEnt
; ---------------------------------------------------------------------------------------------------------
adTabEnt	pshs	d,x,y						; save working registers
			ldy	#tabEntries				; point to tabEntries
			lda	<numTabEnt				; get number of entries in tabEntries
			asla								; multiply by 2 (record length)
			leay	a,y						; point to next record
			clra								; entry number
1			leax	-25,x						; point to previous entry
			inca								; increment entry counter
			cmpx	#readBuffer				; got out of the buffer?
			bhs	1b							; no, loopback
			ldb	<diskSector				; get dir sector where it is
			std	,y						; add record to table
			inc	<numTabEnt				; increment record counter
			puls	d,x,y,pc					; restore registers and return
; ---------------------------------------------------------------------------------------------------------
; LOADFILE process. 	Read the Sectors Table entries
; ---------------------------------------------------------------------------------------------------------
ldFile	ldx	#table					; point to table
			lda	,x++						; get numEntries and skip table header
			lbeq	errorNE					; if 0 show Not Found
			stx	<tablePointer			; save table pointer in variable
			ldb	#1							; mark as
			stb	<firstTime	   		; 1st sector
											; Calculate number of bytes to be copied to RAM
											; it should be zero for all of the sectors but First and Last
											; First could be max equal to 247. Last can be any value (0=256)
1			ldy	,x							; get 1st LSN of this entry
2			ldd	>table					; A=numEntries and B=bytes last sector (0 means 256)
			cmpa  #1							; is numEntries == 1 (the last one)?
			bne   3f							; no, goto next control
			lda	2,x						; get numSectors of this entry
			cmpa	#1							; is numSectors == 1  (last file sector. Could be the first too!)?
			beq	4f							; yes, use bytes in last sector
3			ldb	#247						; get max bytes in 1st sector
			tst	<firstTime				; is the very 1st sector (=1)?
			bne	4f							; yes, use 247 bytes
			clrb								; else use 256 bytes
4			stb	<secNumBytes			; save calculated number of bytes to be copied to RAM
											; Calculate TRACK - SECTOR - SIDE out of LSN that comes in register Y (RealTrack)
			lbsr	Lsn2TS					; convert LSN in Y into track-sector bytes
											; Load that sector into readBuffer or RAM
			clr   <rwFlag					; disk r/w flag 0=r $20=w
			ldx	<ramDestPtr				; get last pointer value
			tst	<secNumBytes			; read full sector?
			beq	ldsect					; yes, use last pointer value -> right into RAM
			ldx	#readBuffer				; special buffer for incomplete sectors
ldsect	stx   <dataPointer			; tape/disk data pointer
			lbsr	read2Buf					; read sector -> preserves B,X,Y,U
											; Copy data to destination if needed
			ldu	<ramDestPtr				; default destination (last used)
			ldb	<secNumBytes			; have we read a full sector?
			bne	docopy					; no, move loaded data
			leau	256,u						; point to next RAM sector
			bra	endcpy					; skip move data
docopy	ldx	#readBuffer				; point to buffer
			tst	<firstTime				; is flag 1st sector ON?
			beq	7f							; no, so it is last one. Skip section
											; calculating where to load the file
			pshs	d							; save work register
			ldb	<orcComand				; get ORC command
			cmpb	#'A'						; is it append?
			bne	calcLda					; no, use file header load info
calcApnd	ldd	>SONGBEG					; get end of song (is beginning of last loaded one)
			subd	>readBuffer+4			; subtract file length (offset +04)
			tfr	d,u						; pass load address to U
			bra	endCalc					; skip next one
calcLda	ldu	>readBuffer+2			; get Load Address from file header
endCalc	puls	d							; restore work register
			cmpu	#minRAM					; out of allowed limits?
			lblo	errorOM					; yes, show message
			stu	<songBegin				; save beginning of the song (file)
			leax	9,x						; set the pointer to the 10th byte of the read buffer (skip 9 bytes of fileheader)
			clr	<firstTime				; set flag 1st sector OFF
7			lda	,x+						; get a byte, increment origin pointer
			sta	,u+						; put at destination, increment pointer
			decb								; decrement counter
			bne   7b							; not yet finished, loopback
endcpy	stu	<ramDestPtr				; save destination pointer for next sector
			leay	1,y						; next sector
			ldx	<tablePointer			; get pointer to table
			dec	2,x						; decrement numSectors of this entry
			bne	2b							; loop to process next sector
			leax	3,x						; point to next table entry
			stx	<tablePointer			; save pointer
			dec	>table					; decrement numEntries of table
			bne	1b							; loopback to process next table entry
  			clr	<diskSide				; restore side = 0
			ldx	<songBegin				; get file load address
			stx	>SONGBEG					; pass to Orc90 as song beginning
			rts								; return
; =========================================================================================================
; GETNAME v0C - Pere Serrat - 21-03-2016
; -----------------------------------------------------------------------------------------------------
; the idea of this group of routines is help read the filenames/drive  number from the first screen row
; as the command is at $400, we should read from $401
; as Orchestra doesn't require double quotes we will do the same, space as separator required anyway
; X WILLTELL.ORC:1 1:WILLTELL.BAK
; if units are received, they must be the same number, else show error
; the drive has to be accepted both ways 1: or :1
; the former in front of the name, the latter at the end
; -----------------------------------------------------------------------------------------------------
; Requirements:
; As we are pointing to the char after the command, there MUST be minimum ONE space and LESS than 28
; must check if there is a leading drive number in form N:
; The filename cannot be longer than EIGHT chars and the extension can have maximum THREE chars
; While checking the chars of the Name, these chars are special and imply special actions:
;		Space 	- padd with zeros and add Extension .ORC
;		Asterisk - save it in the filename, padd with zeros and go for Extension
;		Dot		- padd with zeros and go for Extension
;		Colon		- padd with zeros, check for trailing drive number in format :N, put extension .ORC
; While checking the chars of the Extension, these chars are special:
;		Space		- padd with zeros
;		Asterisc - save it in the extension, padd with zeros
;		Colon		- padd with zeros, chaeck for trailing drive number in format :N
; Only drive numbers 1-2-3-4 are accepted
; A filename only can have one kind of drive number (leading or trailing)
; If each filename has a drive number associated, they MUST be the same number
; =========================================================================================================
getFName	pshs	d,x,y,u					; to preserve register values
getFN01	ldx	#$0401					; point to space after command
			ldb	<orcComand				; get received command
											; multi commands have to control pointer to parameter list
			cmpb	#'M'						; is a Multi command?
			beq	4f							; yes, get pointer
			cmpb	#'Y'						; is a ALL-ONCE command?
			beq	4f							; yes, get pointer
			cmpb	#'G'						; is a GET command?
			bne	gFN01						; no, skip pointer control
4			ldu 	<oldPointer				; is oldPointer null?
			beq	gFN01						; yes, skip control
			ldx 	<oldPointer				; use last pointer
gFN01		clr	<rwndCounter			; set counter to zero
			clra								; get zero
			sta	<wildCard				; clear wild card flag
			deca								; get $ff to
			sta	<driveNumber			; set to NO drive
			lbsr	skpSpa					; skip leading spaces
			bcc	1f							; if no error, skip controls
											; process commands differences
			ldb	<orcComand				; get received command
			cmpb	#'Y'						; is a ALL-ONCE command?
			beq	4f							; yes, reinit
			cmpb	#'M'						; is a Multi command?
			bne	3f							; no, skip section
4			ldx	<oldPointer				; get old pointer
			cmpx	#$0401					; was it to read 1st param?
			beq	2f							; yes, reinit
			ldx	#$0401					; get position after command
			stx	<oldPointer				; update old pointer
			bra	getFN01					; try again from the 1st parameter
3			cmpb	#'G'						; is a GET command?
			bne	1f							; no, skip section
											; this part restores the last loaded song
2			ldx	<oldBeginning			; get old song beginning
			stx	<songBegin				; update mini-DOS variable
			stx	>SONGBEG					; and ORC variable
			stx	>$005d					; set control variables
			stx	>$005e					; to that
			stx	>$0060					; same
			stx	>$0062					; value
			LDS   #$00FA					; reset stack back to $FA
			LBSR  InitHw					; initialise hardware
			lbsr  CmdTop					; go to TOP of song
			lbra	ReInit1					; goto command loop
											; continue reading parameters
1			ldu	#file1Null				; point to the end of filename1
			clrb								; get zero value
cleanF1	stb	,-u						; move pointer backwards and clean a byte
			cmpu	#fileN1					; got to the beginning?
			bne	cleanF1					; no, loop
											; if skpSpa returned no error, check for DIR without parameters
			tsta								; is that the case?
			bmi	endGet					; yes, go set drive number to 1
			lbsr	cpyNam					; get filename
			bcs	readKO					; if error, exit
			ldb	<orcComand				; get received command
			cmpb	#'M'						; is a MULTI command?
			beq	1f							; yes, save pointer
			cmpb	#'Y'						; is a ALL-ONCE command?
			beq	1f							; yes, save pointer
			cmpb	#'G'						; is a GET command?
			bne	test02					; no, skip next
1			stx	<oldPointer				; save pointer to end of last used filename
test02	cmpb	#'X'						; is a RENAME command?
			andcc	#$fe						; clear carry flag without fooling next test!
			bne	endGet					; skip second filename check
											; this part is to be used only by the RENAME command ('X')
			tst	<wildCard				; found any wildcard?
			beq	1f							; no, go for second filename
			bra	2f							; exit Error
1			bsr	skpSpa					; skip leading spaces
			bcs	readKO					; if error, exit
			ldu	#file2Null				; point to the end of filename2
			clra								; get zero value
cleanE1	sta	,-u						; move pointer backwards and clean a byte
			cmpu	#fileN2					; got to the beginning?
			bne	cleanE1					; no, loop
			lbsr	cpyNam					; get it
			bcs	readKO					; if error, exit
			tst	<wildCard				; found any wildcard?
			beq	endGet					; no, continue
2			orcc	#1							; set carry flag
			bra	readKO					; exit Error
endGet	tst	<driveNumber			; is there a valid drive number?
			bpl	readKO					; yes, skip next
			clr	<driveNumber			; set to drive 1
readKO	puls	d,x,y,u,pc				; restore registers and return
; ---------------------------------------------------------------------------------------------------------
; skip all leading spaces (one mandatory) and verify that 1st char
; ---------------------------------------------------------------------------------------------------------
skpSpa	clr   <leadDriveFlag			; reset flag drive in front
			ldb	#28						; max spaces counter
skpSpa1	lda	,x+						; get one char incrementing pointer to next one
			lbsr	stdChar					; standardize char
			cmpa	#' '						; is it a space?
			bne	eskpSpa					; no, go for more tests
			decb								; decrement counter
			bne	skpSpa1					; not the limit, get next char
											; there were no parameters entered
											; error for any command but DIR
			ldb	<orcComand				; get received command
			cmpb	#'D'						; is a DIR command?
			bne	wrong						; no, return error
			clr	<fileN1					; clean first char of filename 1
			lda	#$ff						; set A negative (when OK it returns zero)
			bra	skpOK						; return OK
eskpSpa	cmpb	#28						; found any space?
			beq	wrong						; not, show error (one is mandatory)
			lbsr	chkChar					; verify if it is a correct char
			bcs	wrong						; error, exit
			leax	-1,x						; point to first non space char
			bsr	isDrive					; check if leading drive is included, save and skip it
			bcs	wrong						; error, exit
			clra								; reset carry flag
skpOK		rts								; return OK
wrong		orcc	#1							; set carry flag
			rts								; return ERROR
; ---------------------------------------------------------------------------------------------------------
; check if a leading drive is here. Format should be: number and colon
; ---------------------------------------------------------------------------------------------------------
											; to avoid issuing DN error if first char is not a number
isDrive	lda	,x							; get char just read
			lbsr	stdChar					; standardize char
											; test if it is a number
			cmpa	#'0'						; is it lower than 0?
			blo	noDrive					; yes, no drive received
			cmpa	#'9'						; is it higher than 9?
			bhi	noDrive					; yes, no drive received
											; yes, verify if colon comes after it
			lda	1,x						; get second char
			lbsr	stdChar					; standardize char
			cmpa	#':'						; is it a colon?
			bne	noDrive					; no, exit
			lda	,x							; yes, get drive number again
			lbsr	stdChar					; standardize char
			cmpa	#'1'						; is it lower than 1?
			lblo	errorDN					; yes, not a valid drive number
			cmpa	#'4'						; is it higher than 4?
			lbhi	errorDN					; yes, not a valid drive number
			anda	#$07						; only lower 3 bits identify the drive number (1-4)
			deca								; we use drive numbers 0-1-2-3
			tst	<driveNumber			; have we already read a drive number?
			bmi	newDrive					; no, save this one
			cmpa	<driveNumber			; yes, is the same we have received now?
			beq	oldDrive					; yes, just exit
			lbra	errorDN					; drive number not allowed
newDrive	sta	<driveNumber			; save found drive number
			inc	<leadDriveFlag			; set this flag to 1
oldDrive	leax	2,x						; skip chars for that drive number
noDrive	clra								; reset carry flag
			rts								; return OK
; ---------------------------------------------------------------------------------------------------------
; read  a filename  (8 chars or padd it with zeros)
; ---------------------------------------------------------------------------------------------------------
cpyNam	clr	<colonFndFlag			; no colon found
			ldb	#8							; max name length
cpyN01	lda	,x+						; get a char from screen and increment pointer
			lbsr	stdChar					; standardize char
			cmpa	#' '						; is it a space (end of name)?
			beq	zeroN						; yes, fill with zeros
			cmpa	#'*'						; is it a wildcard?
			bne	cpyNCol					; yes, fill with zeros
			sta	,u+						; put it at destination
			inc	<wildCard				; set flag wildcard found
			decb								; decrement chars to be read
			bra	zeroN						; fill them wil zeros
cpyNCol	cmpa	#':'						; is it a colon?
			bne	cpyN02					; goto next control
			inc	<colonFndFlag			; set flag
			leax	1,x						; advance pointer
			bra	zeroN						; fill with zeros the rest of the filename
cpyN02	cmpa	#'.'						; is it a dot (end of name)?
			beq	zeroN						; yes, fill with zeros
			lbsr	chkChar					; verify if it is a correct char
			lbcs	errorSN					; not, syntax error
			sta	,u+						; put char in destination area
			decb								; decrement counter
			bne	cpyN01					; not all done, loopback
zeroN		cmpb	#8							; any valid chars read?
			bne	zeroN00					; yes, skip section
			lda	<orcComand				; get received command
			cmpa	#'D'						; is it a DIR?
			lbne	errorSN					; no, must have one valid char, show error
												; else accept NO name for a DIR command
zeroN00	tstb								; copied all 8 chars?
			beq	vfyExt					; yes, do not padd name
zeroN01	clr	,u+						; clear next byte
			decb								; end of name space?
			bne	zeroN01					; no, clear another
			tst	<wildCard				; wildcard found?
			bne	vfyExt					; yes, don't move pointer
			leax	-1,x						; decrement pointer
; ---------------------------------------------------------------------------------------------------------
; control first char for extension
; ---------------------------------------------------------------------------------------------------------
vfyExt	tst	<colonFndFlag			; colon found?
			beq	vfyE01					; no, go on
vfyE00	bsr	vfyColon					; check for colon correctness if found
			bcs	cpyErr					; error, exit
			bra	addORC1					; add std termination
vfyE01	lda	,x+						; get char
			lbsr	stdChar					; standardize char
			cmpa	#' '						; is it a space?
			beq	addORC					; yes, add extension ORC
			cmpa	#':'						; is it a colon?
			bne	vfyE02					; goto next control
			inc	<colonFndFlag			; set flag
			bra	vfyE00					; verify and add ORC
vfyE02	cmpa	#'.'						; is it a dot?
			lbne	errorSN					; not, syntax error
; ---------------------------------------------------------------------------------------------------------
; read extension chars (3 chars or padd it with zeros)
; ---------------------------------------------------------------------------------------------------------
cpyExt	ldb	#3							; max extension length
cpyE01	lda	,x							; get a char from screen
			lbsr	stdChar					; standardize char
			cmpa	#' '						; is it a space (end of extension)?
			beq	zeroE						; yes, fill with zeros
			leax	1,x						; advance pointer in not space
			cmpa	#':'						; is it a colon?
			bne	cpyE02					; goto next control
			inc	<colonFndFlag			; set flag
			cmpb	#3							; have we got any extension char?
			beq	addORC					; no, add std extension
			bra	zeroE						; fill with zeros the rest of the extension
cpyE02	bsr	chkChar					; verify if it is a correct char
			lbcs	errorSN					; not, syntax error
			sta	,u+						; put char in destination area
			cmpa	#'*'						; is it a wildcard?
			bne	1f							; no, control pending chras to read
			inc	<wildCard				; signal wildcard found
			bra	zeroE0x					; go to fill with zeros
1			decb								; decrement counter
			bne	cpyE01					; not all done, loopback
zeroE	   tstb								; copied all 3 chars?
			beq	endExt					; yes, do not padd name
zeroE01	clr	,u+						; clear one extension char
zeroE0x	decb								; decrement counter
			bne	zeroE01					; not all done, loopback
			tst	<colonFndFlag			; colon detected?
			bne	zeroE02					; yes, verify drive number
endExt	lda	,x+						; get char
			bsr	stdChar					; standardize it
			cmpa	#':'						; is it a colon?
			bne	zeroE0y					; no, skip process
			inc	<colonFndFlag			; set flag
zeroE02	bsr	vfyColon					; is drive correct?
			bcs	cpyErr					; no, error
endRead	clra								; reset carry flag
			rts								; return	OK
zeroE0y	leax  -1,x						; no, put pointer back at last value
			bra	endRead					; exit
; ---------------------------------------------------------------------------------------------------------
; add default extension chars. By now "ORC"
; ---------------------------------------------------------------------------------------------------------
addORC	leax	-1,x						; to point again to space char
addORC1	ldd	#"OR"						; get two chars "OR"
			std	,u++						; put at destination
			lda	#'C'						; get last char "C"
			sta	,u							; put at destination
			bra	endExt					; exit controlling colon
cpyErr	orcc	#1							; set carry flag
			rts								; return ERROR
; ---------------------------------------------------------------------------------------------------------
; verify if colon is correct here (searching for a trailing drive number)
; ---------------------------------------------------------------------------------------------------------
vfyColon	tst	<colonFndFlag			; found a colon?
			beq	evfyClno					; no, go verify extension
			tst	<leadDriveFlag			; is there a leading drive?
			beq	vfyC01					; no, skip section
errCln	orcc	#1							; set carry flag
			rts								; return ERROR
vfyC01	lda	,x							; get next byte
			bsr	stdChar					; standardize it
			cmpa	#'1'						; is it lower than 1?
			lblo	errorDN					; yes, not a valid drive number
			cmpa	#'4'						; is it higher than 4?
			lbhi	errorDN					; yes, not a valid drive number
			anda	#$07						; only lower 3 bits identify the drive number (0-7)
			deca								; we use drive numbers 0-1-2-3
			tst	<driveNumber			; have we already a drive number?
			bmi	saveDrv					; no, save this one
			cmpa	<driveNumber			; yes, is the same we have received now?
			beq	evfyCln					; yes, just exit skiping 2 chars
			lbra	errorDN					; different, not allowed
saveDrv	sta	<driveNumber			; save this drive number
evfyCln	leax	1,x						; skip colon and drive number
evfyClno	clra								; reset carry flag
			rts								; return OK
; ---------------------------------------------------------------------------------------------------------
; reset bit 6 of not alphabetics chars
; ---------------------------------------------------------------------------------------------------------
stdChar	cmpa	#$8f						; is it the Orchestra special space?
			bne	stdCh01					; no, test other chars
			lda	#' '						; yes substitute by std space ($20)
stdCh01	cmpa	#$60						; is a letter sign?
			blo	noMod						; yes, skip mod
			anda	#%10111111				; to reset bit6
noMod		rts								; return
; ---------------------------------------------------------------------------------------------------------
; check if regA contains a correct char (0-9 or A-Z or '-','*') else set carry
; ---------------------------------------------------------------------------------------------------------
chkChar	pshs	b							; save working register
			cmpa	#'*'						; is it an asterisc?
			beq	goodChar					; yes, acceped
			cmpa	#'-'						; is it a hyphen?
			beq	goodChar					; yes, acceped
			cmpa	#'0'						; is it lower than 0?
			blo	badChar					; error, rejected
			cmpa	#'9'						; is it 0 to 9?
			bls	goodChar					; yes, accepted
			cmpa	#'A'						; is it lower than A?
			blo	badChar					; error, rejected
			cmpa	#'Z'						; is it A to Z?
			bls	goodChar					; yes, accepted
badChar	orcc	#1							; set carry flag
			puls	b,pc						; return ERROR
goodChar clrb								; clear carry flag
			puls	b,pc						; return OK
; ---------------------------------------------------------------------------------------------------------
; search for a valid drive number for command DIR
; ---------------------------------------------------------------------------------------------------------
sDrvNum	lbsr	getFName					; search name, extension and drive
			bcs	isLead					; if errors, verify leading drive number
			tst	<fileN1					; found any letter for the filename?
			beq	sDrv01					; no, verify drive and set wildcards
			rts								; yes, exit
											; see if there is a leading drive
isLead	lda	#$ff						; get value $ff to
			sta	<driveNumber			; set to no disk
			ldx	#$0401					; point to char after the command
			lbsr	skpSpa					; see if there is a leading drive number
			bcc	sDrv01					; if OK, verify and set wildcards
											; else see if there is a trailing drive
isTrail  lda	#$ff						; get value $ff to
			sta	<driveNumber			; set to no disk
			leax	-1,x						; point back one char (must point to the colon)
			clr	<colonFndFlag			; no colon found
			lbsr	endExt					; is there a colon and a number?
			tst	<driveNumber			; found a valid drive number?
			bpl	setWild					; yes, set wildcards
			lbra	errorSN					; else rise error
											; common validations
sDrv01	tst	<driveNumber			; found a valid drive number?
			bpl	setWild					; yes, set wildcards
			clr	<driveNumber			; else set drive 1 by default, clear carry
setWild	lda	#'*'						; set wildcard
			sta	<fileN1					; for name
			sta	<fileE1					; and extension
			rts								; return OK
; =================================================================================================================
; SAVE a song (score) to DISK
; W [DriveNum:]FileName[.Ext][:DriveNum]
; -----------------------------------------------------------------------------------------------------------------
;   Create an entry (or more) on the selected drive for that FileName.Ext
;   if no extension is entered it will use .ORC
;   should rename a pre-existing one as .BAK, deleting previously an already existent .BAK file
; -----------------------------------------------------------------------------------------------------------------
;  Create the FileHeader (9 bytes) just before the beginning of the song data ($5e-9):
;   $55, $02, $Bh, $Bl, $Lh, $Ll, $Bh, $Bl, $AA
;   here Bh:Bl is the beginning address ($5e)
;        Lh:ll is the FileLength ($74 - $5e)
; -----------------------------------------------------------------------------------------------------------------
doWrt		ldb	#1							; put value 1 in
			STB   <diskModeFlag			; mode flag
			stb	>FLGDSK					; update ORC variable too
			CLR   <driveNumber			; current drivenumber = 0 (for drive 1)
			clr	<fatSector				; means NO FAT sector read
											; process start
			ldx	>SONGBEG					; get Orchestra song beginning
			leau	-9,x						; point to header beginning
			ldd	#$5502					; fixed values for Dragon binaries
			std	,u++						; put in header
			stx	,u++						; add load address
			ldd	>SONGEND					; get Orchestra end of song
			subd	>SONGBEG					; subtract beginning
			std	,u++						; add file length
			stx	,u++						; add exec address
			lda	#$aa						; get header end byte
			sta	,u							; put in header
											; initialize variables
   		clra								; set to zero
			sta	<oldFile					; flag exists a file with that same name
			sta	<oldBak					; flag exists a BAK file for that same name
   		sta 	<bytesLastSect			; bytes in last sector (with header)
   		sta	<sectorsToFree			; sectors used by BAK file to be deleted
   		sta	<neededSect				; required number of sectors
   		sta	<numTabEnt				; number of reusable dirEntries
											; search for FileName.Ext
			lbsr	doSearch					; search for a file with the requested filename without filling table $2f0
			bcs	notExists				; if not found, skip controls
			lda	,x							; get flags of filename we have found
			anda	#2							; is it protected?
			lbne	errorPT					; yes, show error
         inc	<oldFile					; no, set flag to true
											; save requested extension to 3 chars area
			ldx	#fileE1					; point origin to requested extension
			ldu	#fileE2					; point destination to second filename buffer
			ldd	,x++						; copy
			std	,u++						; extension
			lda	,x							; from origin
			sta	,u							; to destination
											; change extension to 'BAK'
			ldd	#"BA"						; first two letters
			std	-2,x						; at the beginning
			lda	#'K'						; last char
			sta	,x							; at the end
											; this time it must fill table $2f0
			lda	#'w'						; get false command 'w'
			sta   <orcComand				; update it to make next routine fill table $2f0
			lbsr	doSWrite					; search for a BAK file with the requested filename (using already got filename)
			bcs	notBakFnd				; if not found, skip calculations and controls
			lda	,x							; get flags of filename we have found
			anda	#2							; is it protected?
			lbne	errorPT					; yes, show error
         inc	<oldBak					; no, set flag to true
              							; 	set sectorsToFree = (total number of sectors of these chunks)
			ldu	#table					; point to begining of table
			ldb	,u+						; get number of chunks and skip it
			clra								; reset counter
1			leau	3,u						; point to number of sectors of next record
			adda	,u							; add sectors from that record
			decb								; decrement counter
			bne	1b							; not yet done, loopback
			sta	<sectorsToFree			; save value
notBakFnd								; correct command and extension
			lda	#'W'						; get good command 'W'
			sta   <orcComand				; update it to restore situation
			ldx	#fileE2					; point origin to second filename buffer
			ldu	#fileE1					; point destination to requested extension
			ldd	,x++						; copy
			std	,u++						; extension
			lda	,x							; from the copy
			sta	,u							; to requested area
											; Calculate the number of needed sectors
												; set neededSect = ($74 - $5e + 9)/256 rounded up
notExists									; set bytesLastSect = ($74 - $5e + 9) - (neededSect - 1)*256
     											; if bytesLastSect = 256 then bytesLastSect=0
			ldx	>SONGBEG					; get Orchestra song beginning
			ldd	-5,x						; get file length
			addd	#9							; add the header length
			tstb								; is regB zero?
			beq	2f							; is there an incomplete sector?
			inca								; add 1 sector for incomplete last block
2			stb	<bytesLastSect			; save remainder bytes
			stb	>table+1					; save value in chunks table too
     										; if needed sectors <= sectorsToFree GOTO enough (no need to search for more sectors)
     		cmpa	<sectorsToFree			; compare the size of file to be saved against BAK file
     		beq	exactSame				; if equal size, don't look for more
     		bhi	needMore					; if new file is greater, look for more free sectors
excess									; BAK is greater than file to save, we do not need all of its sectors / chunks
											; reduce the number of records in $2f0 and numsectors as needed
			ldu	#table					; point to table
			ldb	,u++						; get number of records point to 1st record
exComp	cmpa	2,u						; has this record enough sectors?
			bls	gEnough					; yes, adapt it
			suba	2,u						; if need more, subtract 1st record
			decb								; decrement counter
			leau	3,u						; point to next record
			bra	exComp					; loopback for next record
gEnough	sta	2,u						; update number of sectors needed of that record
			negb								; deduct this counter
			addb	>table					; from total records in table
			incb								; to get actual used number of records
			stb	>table					; update this value in table
			leau	3,u						; point to 1st not used record
			lda	#$ff						; get end of table value
			sta	,u+						; set LSN
			sta	,u+						; to $ffff
			clr	,u							; and num sectors to zero
			bra	exactSame				; jump to kill file
											; this part searches for more free sectors
needMore	suba	<sectorsToFree			; subtract BAK sectors
			sta	<neededSect				; save disk sectors we will need to add to BAK ones
 			lbsr	manTable					; no, Search in the FAT for that amount of free sectors and ADD them to table $2f0
exactSame								; The chunks are used in these amounts: 4 + 7 + 7 + 7 ... (first, other entries)
											; numEntries = 1 + (NumRecords-4)/7 rounded UP
			ldb	#1							; default value (minimum)
			lda	>table					; get number of records
			cmpa	#4							; are there more that 4 records (chunks)?
			bls	4f							; no, no more calculations
			suba	#4							; deduct the ones in 1st entry
3			incb								; one more entry needed
			suba	#7							; deduct extents of next entries
			bhi	3b							; if positive not zero, loopback
4	   	stb	<numEntries				; set number of needed entries
	  		lbsr	srchEnt					; Search in the DIR for that amount of free entries
			lbcs	errorFD					; show FD error (Full Directory)
											; enough entries found
		  	tst 	<oldBak					; was there a BAK file?
		  	beq	noDel						; no, nothing to kill
											; delete FileName.BAK (updating DIR and FAT)
												; write BAK as extension in requested name
			ldx	#fileE1					; point origin to requested extension
			ldd	#"BA"						; first two letters
			std	,x++						; at the beginning
			lda	#'K'						; last char
			sta	,x							; at the end
											; change command code to 'K'
			sta   <orcComand				; update it
											; save table tabEntries 'cause it will be modified by Kill
			ldx	#numTabEnt				; point to counter
			ldy	#tabEntCopy				; point to destination
			ldb	,x+						; get number of records to copy
			stb	,y+						; put in destination
cpySeg	ldu	,x++						; get a record
			stu	,y++						; copy it
			decb								; decrement counter
			bne	cpySeg					; not finished, loop
											; save table $2f0  'cause it will be modified by Kill
			ldx	#table					; point to table $2f0
			ldy	#tabCopy					; point to destination
			bsr	copier					; copy table to destination
											; KILL old BAK file
			lbsr 	getGeoData				; get physic disk data
			lbsr 	doSWrite 				; equals doSearch ommiting "bsr comPart", done previously
			lbsr 	doKWrite					; equals doKill without doSearch, done previously
											; restore table tabEntries
			ldx	#tabEntCopy				; point to copy of table
			ldy	#numTabEnt				; point to counter
			ldb	,x+						; get number of records to copy
			stb	,y+						; put in destination
rstSeg	ldu	,x++						; get a record
			stu	,y++						; copy it
			decb								; decrement counter
			bne	rstSeg					; not finished, loop
											; restore table $2f0
			ldx	#tabCopy					; point to destination
			ldy	#table					; point to table $2f0
			bsr 	copier					; copy table from destination
											; copy extension from 2nd name to 1st one
			ldx	#fileE2					; point origin to second filename buffer
			ldu	#fileE1					; point destination to requested extension
			ldd	,x++						; copy
			std	,u++						; extension
			lda	,x							; from the copy
			sta	,u							; to requested area
											; change command code to 'W'
			lda	#'W'						; set command 'W'
			sta   <orcComand				; update it
noDel	  	tst 	<oldFile					; was there a file with same name?
			beq	noRen						; no, nothing to rename
		  									; rename it to FileName.BAK (updating DIR)
		  									; copy filename1 to filename2 (no extension)
			ldx	#fileN1					; point origin to requested extension
			ldu	#fileN2					; point destination to second filename buffer
			ldb	#8							; 8 chars to copy
5			lda	,x+						; copy
			sta	,u+						; filename
			decb								; decrement counter
			bne	5b							; not yet done, loop
											; put BAK in second extension
			ldx	#fileE2					; point origin to requested extension
			ldd	#"BA"						; first two letters
			std	,x++						; at the beginning
			lda	#'K'						; last char
			sta	,x							; at the end
											; change command code to 'R'
			lda	#'R'						; set command 'R'
			sta   <orcComand				; update it
											; call function RENAME
			lbsr 	getGeoData				; get physic disk data
			lda	#'>'+$40					; get false command '>'
			sta   <orcComand				; update it to tell next routine to NOT fill table $2f0
			lbsr 	doSWrite 				; equals doSearch ommiting "bsr comPart", done previously
			lbsr	doRWrite					; perform the rename
											; change command code to 'W'
			lda	#'W'						; set command 'W'
			sta   <orcComand				; update it
noRen	  	lbsr	saveFile					; Write RAM (in blocks of 256 bytes) to the sector chunks from table $2f0
		  	lbsr	fillEnt					; Create the needed directory entries and Update DIR sector(s)
		  	lbsr	tagUSect					; Mark sectors as used in FAT and Update FAT sector(s)
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
; copies from X to Y number bytes: 2+3*([X]+1)
; -----------------------------------------------------------------------------------------------------------------
copier	ldb	,x							; get records counter
			incb								; so that end record is copied too
			ldu	,x++						; copy numrecords and bytes last sector (2 bytes)
			stu	,y++						; to destination
cpyMore	ldu	,x++						; get LSN
			stu	,y++						; save it
			lda	,x+						; get counter
			sta	,y+						; save it (3 bytes = 1 record)
			decb								; decrement records counter
			bne	cpyMore					; not yet done, loopback
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;  Create/add data into table $2f0
;	input:    	none
;  returns:  	regU = pointer to next record in table $2f0
;					the result is table at $2f0
;	requires: 	neededSect = number of free sectors to be found
;					sectorsToFree = sectors of old BAK file to be deleted
;	uses:			regA, regB, regX, regU
; -----------------------------------------------------------------------------------------------------------------
manTable	ldx	#table 					; point to beginning of table (sectors chunk: #$2f0)
			tst 	<sectorsToFree			; was there a BAK file?
			bne	1f							; yes, skip section
			clr	,x							; set zero records
1			ldb	,x						; get number of chunk records
			aslb								; double (x2)
			addb	,x++						; triple (x3) and increment pointer by 2
			abx								; set tabPtr += 2 + table[0]*3 - Point to first record to add
			tfr	x,u						; regU is our table pointer from now on
			lda	<neededSect				; get number of sectors to be found
			bsr	getSecCh					; search for the total number of requested sectors
			lbcs	errorDF 					; new errortype (disk Full) - MUST BE CREATED
			ldd	#$ffff					; add end record
			std	,u++						; table[tabPtr(+=2)] = (16bits)$ffff
			clr	,u							; table[tabPtr]=0
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;  Search for the requested number of free sectors in as many chunks as needed
;	input:    	regA Requested number of sectors
;  returns:  	Carry set if not enough sectors, otherwise cleared
;	requires: 	diskGeoData must have been filled
; 					regU pointing to next record on table $2f0
;	uses:			regB, regX
; -----------------------------------------------------------------------------------------------------------------
getSecCh	clr	<algorithm3				; no algorithm3 enabled
			sta	<reqNumSector			; save number of requested sectors
			ldd	<diskGeoData			; get numTracks and sectors x track
			mul								; calculate total number of tracks
			std 	<maxSect					; save value
			ldb	<diskGeoData+1			; get sectors x track
			ldx	#360						; 1st dir sector for single sided disks
			cmpb	#$12						; is a single sided disk?
			beq	1f							; yes, skip next
			ldx	#720						; 1st dir sector for double sided disks
1			stx	<firstDirSect			; set variable firstDirSect = X
											; fisrt searching algorithm: look for 36 sector chunks (2 tracks) descending from directory 1st sector
Alg1		leax  -36,x						; to return a max chunk of 36 sectors
			bsr 	hmSFree					; calculate how many consecutive free sectors are from regX LSN
			bcs	2f							; if not found any free sectors, go to next block same algorithm
			lbsr	addChTab					; add a record into table $2f0 (regX,regB)
			tst	<reqNumSector			; still more needed?
			beq	5f							; no, return OK
2			cmpx	<zero16bits				; have we arrived at sector 0?
			bhi	Alg1						; no, while (x>0) do search with algorithm 1
											; second searching algorith: from end of directory but beginning of track
			ldx	<firstDirSect			; point to the 1st sector of directory
			leax	+18,x						; 1st sector past the directory - no limit to number of sectors found
Alg2		bsr	hmSFree					; calculate how many free consecutive sectors exist
			bcs	3f							; if not found any free sectors, go to next block same algorithm
			lbsr	addChTab					; add a record into table $2f0 (regX,regB)
			tst	<reqNumSector			; still more needed?
			beq	5f							; no, return OK
											; if used chunk is greater than 18 sectors, must avoid looking for those
											; already assigned tracks again
			pshs	d,u						; save working registers
			ldu	#table					; point to table 2f0
			lda	,u+						; get numrecords
			ldb	#3							; 3 bytes long
			mul								; calculate ofset
			ldb	d,u						; get num sectors in last chunk
7			cmpb	#19						; is greater than 18?
			bcs	8f							; no, point to next block
			leax	+18,x						; skip 18 sectors
			subb	#18						; decrement counter
			bra	7b							; loopback
8			puls	d,u						; restore registers
3			leax	+18,x						; point to next track past directory
			cmpx	<maxSect					; got to the end of disk?
			blo	Alg2						; no, while (x<maxSect) do search with algorithm 2
			pshs	d,x,y,u				; if any chunk had been used with previous algorithms
			jsr	tagUSect				; we must tag them in the FAT as used, so this third algorithm
			puls	d,x,y,u				; will NOT try to use them AGAIN!
											; third algorithm. From sector zero get any chunk of free sectors
preA3		ldb	#1							; get value 1 because
			stb	<fatReqSector			; first time we should open 1st FAT sector
			stb	<algorithm3				; algorithm3 enabled
			clr	<firstFatByte 			; and begin at byte #0
Alg3		lbsr	searchAnyChunk			; search for a chunk from sector 0 to the end (regU-pointer to $2f0)
			bcs	6f							; exit with Error
			bsr	addChTab					; add a record into table $2f0 (regX,regB) updating regU
			tst	<reqNumSector			; still more needed?
			bne	Alg3						; yes, do search with algorithm 3
5			clra								; clear carry = OK
6			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	Count number of consecutive free sectors from the one received
;	input:   	regX = first sector to analyze
;	returns: 	regB = number of free consecutive sectors
;					sets carry flag if no sectors found
;              if needed updates <firstFatByte
;	requires:   reqNumSector = number of sectors to be found
;	uses:			regA
;	preserves: 	regX
; -----------------------------------------------------------------------------------------------------------------
hmSFree	pshs	x							; save working register
			bsr	sect2ByteBit			; should return ByteNumber in X, bitNumber in A
			ldb	#1							; ask for FAT sector 1
			cmpx	#180						; is sector in 1st FAT sector?
			blo	Rs							; yes, skip corrections
			leax	-180,x					; deduct 180 bytes from 1st sector
			incb								; ask for FAT sector 2
Rs			cmpb	<fatSector				; is same last read sector?
			beq	dNrA						; yes, do not read again
			stb	<fatSector				; update variable last read FAT sector
			pshs	a,x						; save: bit number, Byte number
			lbsr 	openDir					; read requested FAT sector
			puls	a,x						; restore them
dNrA		leax  readBuffer,x			; point to FAT byte that contains the bit for requested LSN
			tst	,x							; has that first byte any free sector?
			beq	5f							; no, all used, exit with carry set!
			clr	,-s						; create a hole for free sectors
			sta	,-s						; and another for bit number (must be 0-7)
			nega								; negate
			adda	#8							; add 8 to calculate number of bits to count
			ldb	,s							; get bit number
Rb			sta	,s							; save number of usable bits (requested and next ones)
			cmpx	#readBuffer+180		; reads past valid FAT bytes?
			bhi	3f							; yes, stop searching, return what is found
			lda	,x+						; get byte from the FAT
			tstb								; any bit to be skipped?
			beq	1f							; no, exit loop
Ra			lsra								; shift right read byte
			decb								; decrement counter
			bne	Ra							; loop
											; now we are pointing to the desired bit (sector)
1			dec	,s							; decrement usable bits counter
			bmi	2f							; if done the 8 bits, exit loop
			lsra								; move lower bit to carry
			bcc	3f							; if sector is in use exit loop (0=used, 1=free)
			inc	1,s						; increment number of detected free sectors
			bra	1b							; loopback
2			ldb	<reqNumSector			; get requested number of sectors
			cmpb	1,s						; greater than found by now?
			bls	3f							; no, exit search loop
											; we need more sectors so will read another byte
			lda	#8							; next bytes, verified
			clrb								; from bit 0
			cmpx	#readBuffer+180		; pointer is past valid FAT bytes?
			bhi	3f							; yes, return what we have found
			inc	<firstFatByte			; no, increment variable
			bra 	Rb							; use next FAT byte
3			clra								; clear carry
			leas	1,s						; get rid of counter
			ldb	,s+						; get number of consecutive free sectors
			bne	4f							; if any found, return with carry clear
5			orcc	#1							; set carry (no sectors found)
4			puls	x,pc						; restore regX and return
; -----------------------------------------------------------------------------------------------------------------
;	Converts a Sector number into FAT byte number and bitnumber inside that byte
;	input:   	regX = LSN number to convert
;	returns: 	regX = FAT byte number
;					regA = bitnumber inside that byte(0 lsb - 7 MSB)
;	uses:			regB
; -----------------------------------------------------------------------------------------------------------------
sect2ByteBit
			tfr	x,d						; D = LSN number
			pshs	b							; save lower LSN byte
			lsra								;
			rorb								; divide regD by 2
			lsra								;
			rorb								; by 4
			lsra								;
			rorb								; by 8 - now regD contains the FAT byte (0-359)
			tfr	d,x						; pass FAT byte number to regX
			puls	a							; restore LSN low byte
			anda	#$07						; use only three lower bits (0-7)
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	Add a record into table $2f0 for the new chunk
;	input:   	regX = first LSN of this chunk
;	returns: 	regB = number of free consecutive sectors in this chunk
;	requires:   regU pointing to the 'next' record  in table $2f0
;	uses:			regA, regU
; -----------------------------------------------------------------------------------------------------------------
addChTab	inc	table						; table[0]+=1
			pshs	b							; save received number of sectors
			lda	<reqNumSector			; get still needed sectors number
			suba	,s+						; calculate if still more needed
			pshs	a							; save still 'needed'
			bhs	1f							; if we still need more, skip next
			ldb	<reqNumSector			; else use just the requested
1			stx	,u++						; table[tabPtr(+=2)] = (16bits)Fisrt LSN in Chunk
			stb	,u+						; table[tabPtr(+=1)] = (8bits)Number Of Sectors In Chunk
			lda	,s+						; get still needed sectors number
			bpl	2f							; if positive skip next
			clra								; if negative set 'needed' to zero
2			sta	<reqNumSector			; update still needed sectors number
			addb	<numSecFound			; add found sectors in this chunk to total found
			stb	<numSecFound			; update variable. numSecFound += Number Of Sectors In Chunk
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	Search for any number of consecutive free sectors from fatReqSector
;	   begins searching from firstFatByte and steps to FAT sector 2 if needed
;	returns: 	regB = number of free consecutive sectors (from hmsFree)
;	requires:   reqNumSector = number of sectors to be found (for hmsFree)
;					fatReqSector = FAT sector to work (1 first time)
;					firstFatByte = number of byte to start with from FAT sector buffer (0 first time)
;	uses:			regA, regX
;	preserves:	regU (should be pointing to 'next' record in table $2f0)
; -----------------------------------------------------------------------------------------------------------------
searchAnyChunk								; search for a byte greater than zero
			pshs	u							; save sector and pointer to table $2f0
			ldb	<fatReqSector			; get FAt sector to work with
			cmpb	<dirSector				; is requested sector same as last FAT sector we read?
			beq	getBufPt					; yes, skip next
readCFAT	lbsr	openDir					; read requested sector
getBufPt	ldu	#readBuffer				; point to buffer
			lda	<firstFatByte			; get variable
			pshs	b							; save regB
			tfr	a,b						; pass regA to regB
			clra								; makes regB positive
			leau	d,u						; add regB to regU unsigned
			puls	b							; restore regB
another	lda	,u+						; get a byte
			bne	freeFnd					; if not 0 exit loop. we can use a sector or more
			cmpu	#readBuffer+180		; got past end of FAT sector?
			blo	another					; no, check next byte
			incb								; increment sector number
			stb	<fatReqSector			; update working sector
			cmpb	#3							; got to end of FAT?
			bhs	srchFail					; yes, return with carry set
			clr	<firstFatByte			; from 1st byte of new sector
			bra	readCFAT					; now read sector
srchFail	orcc	#1							; yes, set carry flag
			leas	2,s						; get rid of pointer
			rts								; return disk full
freeFnd	pshs	a							; save contents of 1st LSN with free byte(s)
			leau	-readBuffer,u			; calculate byte number in FAT
			pshs	b							; save FAT sector
			tfr	u,d						; pass bytenumber to D
			stb	<firstFatByte			; save next FAT byte to be read next time
			puls	b							; restore bytenumber
			leau	-1,u						; to point to last read byte
			cmpb	#1							; is it in 1st sector?
			beq	2f							; yes, skip next
			leau	180,u						; add bytes from 1st sector to get real sector number on second FAT sector
2			tfr	u,d						; pass data to D
			aslb
			rola								; multipy by 2
			aslb
			rola								; multipy by 4
			aslb
			rola								; multipy by 8
			tfr	d,x						; put byte number into regX
			ldb	#$ff						; bit counter to -1
			puls	a							; restore 1st not zero FAT byte
1			incb								; increment bit counter
			lsra								; send low bit to carry
			bcc	1b							; if not set, loop
			abx								; add to X to get the right LSN (bit: 0-7)
			lbsr	hmSFree					; call hmsFree with this LSN (byte * 8 + bit) updating FAT sector and next byte
			puls	u,pc						; restore pointer and return
; -----------------------------------------------------------------------------------------------------------------
;  Search for the needed unused entries in the directory
;	input:   	regB = number of needed directory entries
;	returns: 	tabEntries = structure(numEntry byte; sector byte)
;					carry set if not enough free entries found
; 	uses:			regB, regX, regY, regU
;  sideEfect:	regX points to the LAST unused entry found
; -----------------------------------------------------------------------------------------------------------------
											; to take into account the entries of BAK file to be deleted
srchEnt	cmpb	<numTabEnt				; has deleted file enough entries for new file?
			bhi	srchMore					; no, search for more
			beq	bothSame					; they are equal so no corrections, exit
											; must reduce not needed Entries. Not mandatory because we will use numEntries
			stb	<numTabEnt				; update counter to really needed amount
bothSame	clrb								; clear carry
			rts								; return
											; look for more free directory entries ...
srchMore	subb	<numTabEnt				; deduct the ones we already have
			tfr	b,a						; save number of entries to look for
			ldy	#tabEntries				; point to beginning of table of entries
			ldb	<numTabEnt				; get number of entries in the table
			aslb								; multiply by 2
			leay	b,y						; point to next empty record
			ldb	#3							; to read sector 3 of track 20 (1st DIR sector)
rdNxtDS	pshs	a							; save parameter
			lbsr	openDir					; read directory sector
			lda	,s							; restore parameter
			clr	,s							; clear entry counter
			ldx	#readBuffer				; set pointer to beginning of buffer
1			inc	,s							; increment entry counter
			ldb	,x							; get flags
			bpl	2f							; if in use see next
			tfr	d,u						; save D
			lda	,s							; get entry number
			ldb	<diskSector				; get dir sector where it is
			std	,y++						; fill tabEntries record with numEntry and sectorNumber
			inc	<numTabEnt				; increment counter
			tfr	u,d						; restore D
			deca								; decrement needed entries
			beq	allFnd					; all found, return OK
2			leax	25,x						; point to next entry
			cmpx  #readBuffer+$f9		; past the 10th entry?
			blo	1b							; loop to test next entry
			leas	1,s						; clean stack
			ldb	<dirSector				; get last read directory sector
			incb								; increment number
			cmpb	#19						; is sector > 18 (out of directory)
			bne	rdNxtDS					; no, read next one
			orcc	#1							; set carry flag
			rts								; return
allFnd	clra								; to clear carry
			puls	a,pc						; clean stack and return
; -----------------------------------------------------------------------------------------------------------------
;	Save the RAM data to disk sectors. We already know the needed number of sectors to write
;	even the number of bytes in the last sector
;	input:   	table $2f0 should have been filled with valid data
; 	uses:			regA, regX, regY, regU
; Stack:			RetAdr, NumRecords, numSectors
; -----------------------------------------------------------------------------------------------------------------
saveFile lda	>table					; get num records
			pshs	a							; save value
			ldx	>$005e					; get beginning of song
			leax	-9,x						; to point to 1st header byte
			stx	<dataPointer			; save in variable
			ldu	#table+2					; point to 1st LSn of first record
											; get one record
sFile01	ldy	,u++						; get 1st LSN in this chunk
			ldb	,u+						; get number of sectors
			pshs	b							; save on stack
											; process one record
sSameR	dec	,s							; decrement sectors to read
			bne	sFile04					; write one sector
											; detect very last file sector
			dec	1,s						; decrement record counter
			bne	sFile04					; not last one, skip section
												; yes, save that last sector
sLastOne ldb	>table+1					; get	bytes last sector
			beq	sFile04					; if 256, skip section
											; is last sector and has less than 256 bytes
			ldx	#readBuffer				; point to buffer
			ldu	<dataPointer			; point to RAM data
			stx	<dataPointer			; put readBuffer as next block pointer
sFile02	lda	,u+						; get a RAM byte
			sta	,x+						; put into buffer
			decb								; decrement counter
			bne	sFile02					; not yet done, loopback
			clra								; get a zero
sFile03	sta	,x+						; clear buffer byte
			cmpx	#readBuffer+256		; got end of buffer?
			blo	sFile03					; no, loopback
											; save 256 RAM bytes into a disk sector
sFile04	bsr	Lsn2TS					; convert into Track-Sector-DiskSide
			LDA   #$20						; code for op write
			STA   <rwFlag					; disk r/w flag
			lbsr	read2Buf					; WRITE sector (saves B,X,Y,U)
			leax	256,x						; point to next RAM sector
			stx	<dataPointer			; update pointer
			leay	1,y						; next LSN in same chunk
			tst	,s							; have we read all the sectors of this record?
			bne	sSameR					; no, loop to process next sector in same record
			tst	1,s						; are there more records?
			beq	sFile05					; no, exit
			leas	1,s						; get rid if sector counter
			bra	sFile01					; loopback to next chunk
sFile05	leas	2,s						; get rid of records counter and sector counter
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	Calculate TRACK - SECTOR - SIDE
;	input:   	regY	  	  = LSN (RealTrack)
;	returns: 	<diskSide with the side to be used
;					<diskTrack with calculated track-sector
;  requires:	<diskGeoData should have been calculated previously
;	uses:			regA, regB
; -----------------------------------------------------------------------------------------------------------------
Lsn2TS	ldb	<diskGeoData+1			; get numSect x track
			clra								; high byte to 0
        	pshs  a,b						; pushes 16 bits value
        	clr   ,-s						; create a counter in stack, value 0
        	tfr   y,d						; copies Y (LSN) into D
1		  	inc   ,s							; increment track counter
        	subd  1,s						; subtract numSect x track from LSN at D
        	bpl   1b							; if not negative, loopback
        	lda   ,s++						; get track counter, move stack 2 bytes up
        	deca								; decrement by one (we got a negative count)
        	addb  ,s							; convert B to positive number
        	incb								; increment it = sector number
      	clr	<diskSide				; by default side = 0
											; avoid corrections if disk is SINGLE sided
        	pshs	a							; save track number
        	lda	<diskGeoData+1			; get numSect x track
        	cmpa	#$24						; is the disk doublesided?
 			puls	a							; restore track number
 			bne	2f							; not double sided, skip correction
											; Correct side only if sector is greater than half the number of sectors per track
        	asr	,s							; divide numSect x track by 2
        	cmpb	,s							; is sector number > half numSect x track?
        	ble	2f							; no, then side = 0  is OK
  			inc	<diskSide				;
	      inc	<diskSide				; side = 2 (bit 1 set)
			subb	,s							; subtract half numSect x track
2		  	leas	1,s						; clean stack
      	std   <diskTrack   			; save calculated Track & Sector
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	Creates and fills the needed entries with relevant data
; 	input:		tabEntries = table containing (numEntry and numSector) for each free entry found
;	requires:   numEntries = number of entries to be created
;	uses:			regA, regB, regX, regU, regY
; -----------------------------------------------------------------------------------------------------------------
fillEnt	lda	#1							; ask for 1st entry
			lbsr	readEnt					; search for it and return X pointing to it
											; fill 1st entry with Flags, Name, Ext, 4 Extents, BytesLastSector (unused extents 3 zeroes)
											; Flags: bit1=0 - bit5=1 if more extents
											; if bit5=1 BytesLastsector is substituted by next DIR entry number: (numSec-3)*10+numEntry
			clra								; by default only one entry
			ldb	>table					; get table[0] - num records
			cmpb	#5							; are there more than 4 records?
			blo	3f							; no, no more entries needed go get bytesLastSector
			adda	#32						; yes, so flag more entries to come
			pshs	a							; save flags
			lda	<tabEntries+3			; get sectorNumber of next entry
			suba	#3							; deduct 3 (as 1st dir sector is #3)
			ldb	#10						; entries x sector
			mul								; regB contains numentries in previous sectors
			addb	<tabEntries+2			; add entryNumber
			puls	a							; restore flags
			bra	7f							; go write byte 25
3			ldb	>table+1					; get bytesLastSector
7			stb	24,x						; save BytesLastSector or NextEntryNumber
			sta	,x+						; save flags
			ldu	#fileN1					; point to filename
			ldb	#11						; to copy name and extension
4			lda	,u+						; get a char
			sta	,x+						; copy into entry
			decb								; decrement counter
			bne	4b							; if not done, loopback
											; fill the extents with record data
			ldb	#4							; limit is 4 for 1st entry
			ldy	#table+2					; point to 1st entry in table
5			ldu	,y++						; get 1st LSN of that record
			lda	,y+						; get numSectors of that record
			beq	9f							; if zero sectors, end of DATA
			stu	,x++						; put data
			sta	,x+						; in directory entry
			decb								; decrement counter
			bne	5b							; loopback
9			tstb								; have we put all 7 extents?
			beq	8f							; yes, skip-exit section
			clra								; get zero
A9			sta	,x+						; to clear unused extents
			sta	,x+						; put LSN zero
			sta	,x+						; put numsectors zero
			decb								; decrement counter
			bne	A9							; loopback
											; end of first entry	- beginning of 'more' entries
8			ldb	<numEntries				; get number of entries to be created
			lda	#1							; last read entry
1			decb								; decrement counter
			beq	2f							; no more entries, return
			inca								; ask for next entry
			pshs	a,b						; save entry and counter
			bsr	readEnt					; search for it and return X pointing to it
											; fill next entry with Flags, 7 Extents, 2 nulls, next Entry number (unused extents 3 zeroes)
											; Flags: bit1=1 - bit5=1 if more extents
											; Last entry has in byte24 the BytesLastSector
			clra								; by default last entry
			ldb	>table					; get table[0] - num records
			subd	#4							; deduct extents from 1st entry
			cmpb	#8							; are there more than 7 records?
			blo	N3							; no, no more entries needed go get bytesLastSector
			adda	#32						; yes, so flag more entries to come
			ldu	#tabEntries				; point to beginning of tableEntries
			pshs	a							; save flags
			lda	1,s						; get entry number
			deca								; deduct 1 (1st one has no offset)
			asla								; each record 2 bytes
			leau	a,u						; point to numEntry
			lda	1,u						; get num sector
			suba	#3							; deduct 3
			ldb	#10						; entries x sector
			mul								; regB contains numentries in previous sectors
			addb	,u							; add entry number
			puls	a							; restore flags
			bra	N7							; go write byte 25
N3			ldb	>table+1					; get bytesLastSector
N7			stb	24,x						; save BytesLastSector or NextEntryNumber
			sta	,x+						; save flags
											; fill the extents with record data - Y should be preserved as table $2f0 pointer
			ldb	#7							; limit is 7 for other entries
N5			ldu	,y++						; get 1st LSN of that record
			lda	,y+						; get numSectors of that record
			beq	N9							; if zero sectors, end of DATA
			stu	,x++						; put data
			sta	,x+						; in directory entry
			decb								; decrement counter
			bra	N5							; loopback
N9			tstb								; have we put all 7 extents?
			beq	N8							; yes, skip-exit section
			clra								; get zero
NA			sta	,x+						; to clear unused extents
			sta	,x+						; put LSN zero
			sta	,x+						; put numsectors zero
			decb								; decrement counter
			bne	NA							; loopback
N8			puls	a,b						; restore entry and counter
			bra	1b							; loopback
2			lbsr	rewDIR					; write last modified sector back to update the 'added' entries
			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	read a sector and point to the requested entry UPDATING previously modified sector if needed
;	input:   	regA	  	  = index to tabEntries
;	returns: 	regX		  = pointer to requested entry
;  requires:	tabEntries = should exists as an array of 16 bytes (8 words)
;	uses:			regB, regU
; -----------------------------------------------------------------------------------------------------------------
readEnt	pshs	a							; save parameter
			ldu	#tabEntries				; point to beginning of table
			deca								; 1st entry is at offset 0
			asla								; double
			leau	a,u						; point to requested entry in table
			ldd	,u							; regA=entry number; regB=sector
			dec	,s+						; was parameter 1 (first entry)?
			beq	3f							; yes, skip section
			cmpb	<dirSector				; is the same sector we read last time?
			beq	3f							; yes, do not save sector
											; if NOT 1st read AND new sector != previously read, UPDATE last one
			pshs	a,b						; save entry, sector
			lbsr	rewDIR					; write last modified sector back to update its 'added' entries
			puls	a,b						; restore entry, sector
3			pshs	a							; save entry number
			lbsr	openDir					; read that directory sector (regB)
			puls	a							; restore entry number
			ldx	#readBuffer				; point to beginning of buffer
1			deca								; decrement counter (we work 1-10 but 1 is ofsett 0)
			beq	2f							; if zero exit loop
			leax	25,x						; point to next entry
			bra	1b							; loopback
2			rts								; return
; -----------------------------------------------------------------------------------------------------------------
;	mark sectors as already used in the FAT
;  requires:	table $2f0 should have been filled previously
;	uses:			regA, regB, regX, regU
;	Stack:		RetAdrs, numRecords, ((bitNumber)), (FAT Byte pointer), ((used bit mask)), (((numberOfSectors, mask)))
; -----------------------------------------------------------------------------------------------------------------
tagUSect	ldu	#table					; point to table $2f0
			ldb	,u++						; get numRecords and point to 1st record
			beq	eTUS						; if no recors, exit
			pshs	b							; records counter onto stack
			clr	<fatSector				; NO previuosly read FAT sector
tUS01		ldx	,u++						; get 1st LSN of this record
			lbsr	sect2ByteBit			; should return ByteNumber in X, bitNumber in A (0-7)
			ldb	#1							; ask for FAT sector 1
			cmpx	#180						; is sector in 1st FAT sector?
			blo	tUS02						; yes, skip corrections
			leax	-180,x					; deduct 180 bytes from 1st sector
			incb								; ask for FAT sector 2
tUS02		pshs	a							; save bitNumber
			pshs	x							; save actual FAT byte pointer
			tst	<fatSector				; have a read sector?
			beq	tUS03						; no, skip section
			cmpb	<fatSector				; is sector to read the last FAT read
			beq	tUS08						; yes, do not update nor read
			pshs	b							; save new sector number
			lbsr	rewFAT					; update last FAT sector
			puls	b							; restore sector number
tUS03		stb	<fatSector				; update variable last read FAT sector
			lbsr 	openDir					; read requested FAT sector
tUS08		puls	x							; restore FAT byte pointer
			leax  readBuffer,x			; point to the calculated byte in the buffer
			orcc	#$01						; set carry, to put '1' on the mask
			ldb	#$fe						; #%11111110 mask to change bit0 to used
			stb	,-s						; put mask onto stack
			tst	1,s						; any bit to be skipped?
			beq	tUS05						; no, exit loop
tUS04		rol	,s							; move the zero to next bit to mark it as USED (bit7=1 goes into carry)
			dec	1,s						; decrement bit number
			bne	tUS04						; if not zero, loop
											; now regB has a zero at the bit that represents the 1st sector in use
tUS05		ldb	,s							; get mask
			andb	,x							; apply mask to pointed byte to mark sector as used
			stb	,x							; update FAT byte
			puls	a							; get mask
			puls	b							; get rid of bit number
			ldb	,u+						; get number of consecutive used sectors
			std	,--s						; mask and counter onto stack
			dec	1,s						; decrement numsectors counter (1st done)
			beq	noMore2					; if zero exit but must not overwrite already updated byte!
											; now stack points to mask and then is the numberOfSectors (regA-regB)
tUS06		lda	,x							; get FAT byte
tUS07		tst	,s							; is '0' mask already on bit7?
			bpl	nextByte					; yes, go for next FAT byte
			rol	,s							; move zero mask to next bit
			anda	,s							; apply mask to FAT byte
			dec	1,s						; decrement numsectors counter
			beq	noMore					; if zero exit loop
			bra	tUS07						; loopback
											; yes, work with next byte
nextByte sta	,x+						; update modified byte
			ldb	#$ff						; no masked bit
			stb	,s							; mask onto stack
			andcc	#$fe						; set Carry to zero, so that 1st time will enter bit0
			bra	tUS06						; go get next FAT byte
noMore	sta	,x							; update last modified byte
noMore2	leas	2,s						; clean stack (mask and counter)
			dec	,s							; decrement numrecords counter
			bne	tUS01						; back to get next record
			leas	1,s						; clean stack
			lbsr	rewFAT					; update last FAT sector
eTUS		rts								; return
; ---------------------------------------------------------------------------------------------------------
	if BIN_OUT == 1
			if ((*+9)%256 > 0)
					rzb	 256-((*+9)%256)	; make the BIN compatible DW4 (length multiple of 256 bytes)
			endif
	else
			if (*%256 > 0)
					rzb	 256-(*%256)	; make the ROM length multiple of 256 bytes
			endif
	endif
; ---------------------------------------------------------------------------------------------------------
